/* -*- C -*- 

   memP Heap Profiler ( http://memp.sourceforge.net/ )

   Please see COPYRIGHT AND LICENSE information at the end of this file.

   ----- 

   memPi.c -- main memP internal functions

 */

#ifndef lint
static char *svnid = "$Id: memPi.c 383 2006-12-08 18:30:15Z chcham $";
#endif

#include <string.h>
#include <strings.h>
#include <float.h>
#include <limits.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/resource.h>
#include "memPi.h"

unsigned int
memPi_callsite_stats_pc_hashkey (const void *p)
{
  int res = 0;
  int i;
  callsite_stats_t *csp = (callsite_stats_t *) p;
  MEMP_CALLSITE_STATS_COOKIE_ASSERT (csp);
  for (i = 0; i < MEMP_CALLSITE_STACK_DEPTH; i++)
  {
    res ^= (unsigned) csp->pc[i];
  }
  return 52271 ^ csp->op ^ res ^ csp->rank;
}

int
memPi_callsite_stats_pc_comparator (const void *p1, const void *p2)
{
  int i;
  callsite_stats_t *csp_1 = (callsite_stats_t *) p1;
  callsite_stats_t *csp_2 = (callsite_stats_t *) p2;
  MEMP_CALLSITE_STATS_COOKIE_ASSERT (csp_1);
  MEMP_CALLSITE_STATS_COOKIE_ASSERT (csp_2);

#define express(f) {if ((csp_1->f) > (csp_2->f)) {return 1;} if ((csp_1->f) < (csp_2->f)) {return -1;}}
  express (op);
  express (rank);

  for (i = 0;
       csp_1->pc[i] != NULL && csp_2->pc[i] != NULL
       && i < MEMP_CALLSITE_STACK_DEPTH; i++)
  {
    express (pc[i]);
  }
#undef express

  return 0;
}

unsigned int
memPi_callsite_stats_src_hashkey (const void *p)
{
  int res = 0;
  callsite_stats_t *csp = (callsite_stats_t *) p;
  MEMP_CALLSITE_STATS_COOKIE_ASSERT (csp);
  return 52271 ^ csp->op ^ res ^ csp->rank;
}

int
memPi_callsite_stats_src_comparator (const void *p1, const void *p2)
{
  callsite_stats_t *csp_1 = (callsite_stats_t *) p1;
  callsite_stats_t *csp_2 = (callsite_stats_t *) p2;
  int i;

  MEMP_CALLSITE_STATS_COOKIE_ASSERT (csp_1);
  MEMP_CALLSITE_STATS_COOKIE_ASSERT (csp_2);

#define express(f) {if ((csp_1->f) > (csp_2->f)) {return 1;} if ((csp_1->f) < (csp_2->f)) {return -1;}}
  express (op);
  express (rank);
  if (csp_1->csid == -1 || csp_2->csid == -1)
  {
    for (i = 0;
         csp_1->pc[i] != NULL && csp_2->pc[i] != NULL
         && i < MEMP_CALLSITE_STACK_DEPTH; i++)
    {
      express (pc[i]);
    }

  }
  else
    express (csid);


#undef express

  return 0;
}

unsigned int
memPi_callsite_stats_src_id_hashkey (const void *p)
{
  int res = 0;
  callsite_stats_t *csp = (callsite_stats_t *) p;
  MEMP_CALLSITE_STATS_COOKIE_ASSERT (csp);
  return 52271 ^ csp->op ^ res ^ csp->csid;
}

int
memPi_callsite_stats_src_id_comparator (const void *p1, const void *p2)
{
  callsite_stats_t *csp_1 = (callsite_stats_t *) p1;
  callsite_stats_t *csp_2 = (callsite_stats_t *) p2;
  MEMP_CALLSITE_STATS_COOKIE_ASSERT (csp_1);
  MEMP_CALLSITE_STATS_COOKIE_ASSERT (csp_2);

#define express(f) {if ((csp_1->f) > (csp_2->f)) {return 1;} if ((csp_1->f) < (csp_2->f)) {return -1;}}
  express (op);
  express (csid);
#undef express

  return 0;
}


h_t *callsite_pc_cache = NULL;

int
callsite_pc_cache_comparator (const void *p1, const void *p2)
{
  callsite_pc_cache_entry_t *cs1 = (callsite_pc_cache_entry_t *) p1;
  callsite_pc_cache_entry_t *cs2 = (callsite_pc_cache_entry_t *) p2;

  if ((long) cs1->pc > (long) cs2->pc)
  {
    return 1;
  }
  if ((long) cs1->pc < (long) cs2->pc)
  {
    return -1;
  }
  return 0;
}

unsigned int
callsite_pc_cache_hashkey (const void *p1)
{
  callsite_pc_cache_entry_t *cs1 = (callsite_pc_cache_entry_t *) p1;
  return 662917 ^ (unsigned int) cs1->pc;
}


h_t *callsite_src_id_cache = NULL;

int
callsite_src_id_cache_comparator (const void *p1, const void *p2)
{
  int i;
  callsite_src_id_cache_entry_t *csp_1 = (callsite_src_id_cache_entry_t *) p1;
  callsite_src_id_cache_entry_t *csp_2 = (callsite_src_id_cache_entry_t *) p2;

#define express(f) {if ((csp_1->f) > (csp_2->f)) {return 1;} if ((csp_1->f) < (csp_2->f)) {return -1;}}
  for (i = 0; i < MEMP_CALLSITE_STACK_DEPTH; i++)
  {
    if (csp_1->filename[i] != NULL && csp_2->filename[i] != NULL)
    {
      if (strcmp (csp_1->filename[i], csp_2->filename[i]) > 0)
      {
        return 1;
      }
      if (strcmp (csp_1->filename[i], csp_2->filename[i]) < 0)
      {
        return -1;
      }
      express (line[i]);
      if (strcmp (csp_1->functname[i], csp_2->functname[i]) > 0)
      {
        return 1;
      }
      if (strcmp (csp_1->functname[i], csp_2->functname[i]) < 0)
      {
        return -1;
      }
    }
    //else
    /*fprintf (stderr, "Comparing pc1[%d] %p with ps2[%d] %p\n", i,
       csp_1->pc[i], i, csp_2->pc[i]); */

    express (pc[i]);
    if (csp_1->pc[i] == NULL || csp_2->pc[i] == NULL
        || strcmp (csp_1->functname[i], "main") == 0
        || strcmp (csp_1->functname[i], ".main") == 0
        || strcmp (csp_1->functname[i], "MAIN__") == 0)
      return 0;
  }
#undef express
  //fprintf (stderr, "Returning 0\n");
  return 0;
}

unsigned int
callsite_src_id_cache_hashkey (const void *p1)
{
  int i, j;
  unsigned int res = 0;
  callsite_src_id_cache_entry_t *cs1 = (callsite_src_id_cache_entry_t *) p1;
  for (i = 0; i < MEMP_CALLSITE_STACK_DEPTH; i++)
  {
    if (cs1->filename[i] != NULL)
    {
      for (j = 0; cs1->filename[i][j] != '\0'; j++)
      {
        res ^= (unsigned) cs1->filename[i][j];
      }
      for (j = 0; cs1->functname[i][j] != '\0'; j++)
      {
        res ^= (unsigned) cs1->functname[i][j];
      }
    }
    res ^= cs1->line[i];
  }
  return 662917 ^ res;
}

int
memPi_insert_callsite_records (callsite_stats_t * p)
{
  callsite_stats_t *csp = NULL;

  memPi_query_src (p);          /* sets the file/line in p */

  /* Check hashtable for entry based on source info. 
     If exists, accumulate, otherwise insert. 
     This is specifically for optimizations that have multiple PCs for
     one src line. We aggregate across rank after this. 

     The collective_report reporting approach does not aggregate individual 
     process callsite information at the collector process.
   */
  memPi_msg_debug ("In %s, memPi.collective_report is %d\n", __func__,
                   memPi.collective_report);
  if (memPi.collective_report == 0)
  {
    if (NULL == h_search (memPi.global_callsite_stats, p, (void **) &csp))
    {
      int j;
      callsite_stats_t *newp = NULL;
      newp = (callsite_stats_t *) memPi_malloc (sizeof (callsite_stats_t));
      bzero (newp, sizeof (callsite_stats_t));
      newp->op = p->op;
      newp->rank = p->rank;
      for (j = 0; j < MEMP_CALLSITE_STACK_DEPTH; j++)
      {
        newp->pc[j] = p->pc[j];
        newp->filename[j] = p->filename[j];
        newp->functname[j] = p->functname[j];
        newp->lineno[j] = p->lineno[j];
      }
      newp->csid = p->csid;
      newp->count = p->count;
      newp->maxAlloc = p->maxAlloc;
      newp->minAlloc = p->minAlloc;
      newp->cumulativeAlloc = p->cumulativeAlloc;
      newp->maxFree = p->maxFree;
      newp->minFree = p->minFree;
      newp->cumulativeFree = p->cumulativeFree;
      newp->csHWM = p->csHWM;
      newp->csLWM = p->csLWM;
      newp->cookie = MEMP_CALLSITE_STATS_COOKIE;

      /* insert new record into global */
      if (h_insert (memPi.global_callsite_stats, newp) == 1)
      {
        memPi_msg_warn
          ("Duplicate insertion found in memPi.global_callsite_stats\n");
      }
    }
    else
    {
      csp->count += p->count;
      assert (csp->cumulativeTime >= 0);
      csp->maxAlloc = max (csp->maxAlloc, p->maxAlloc);
      csp->minAlloc = min (csp->minAlloc, p->minAlloc);
      csp->cumulativeAlloc += p->cumulativeAlloc;
      csp->maxFree = max (csp->maxFree, p->maxFree);
      csp->minFree = min (csp->minFree, p->minFree);
      csp->cumulativeFree += p->cumulativeFree;
      csp->csHWM = max (csp->csHWM, p->csHWM);
      csp->csLWM = min (csp->csLWM, p->csLWM);
    }
  }

  return 1;
}


callsite_stats_t *
memPi_get_callsite (unsigned int op, int rank, void **pc)
{
  callsite_stats_t key, *csp = NULL;
  int i;

  key.op = op;
  key.rank = rank;
  key.cookie = MEMP_CALLSITE_STATS_COOKIE;

  for (i = 0; i < MEMP_CALLSITE_STACK_DEPTH; i++)
  {
    key.pc[i] = pc[i];
  }

  (void) h_search (memPi.task_callsite_stats, &key, (void **) &csp);

  return csp;
}



/* 

<license>

Copyright (c) 2010, Lawrence Livermore National Security, LLC.
Produced at the Lawrence Livermore National Laboratory
Written by Christopher Chambreau <chcham@llnl.gov>.
CODE-OCEC-10-010
All rights reserved.

This file is part of memP. For details, see http://memp.sourceforge.net .
Please also read this link – Additional BSD Notice.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  o  Redistributions of source code must retain the above copyright notice, 
     this list of conditions and the disclaimer below.

  o  Redistributions in binary form must reproduce the above copyright notice, 
     this list of conditions and the disclaimer (as noted below) in the 
     documentation and/or other materials provided with the distribution.

  o  Neither the name of the LLNS/LLNL nor the names of its contributors may be
     used to endorse or promote products derived from this software without 
     specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL LAWRENCE LIVERMORE NATIONAL
SECURITY, LLC, THE U.S. DEPARTMENT OF ENERGY OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.


Additional BSD Notice

1. This notice is required to be provided under our contract with the U.S.
Department of Energy (DOE). This work was produced at Lawrence Livermore
National Laboratory under Contract No. DE-AC52-07NA27344 with the DOE.

2. Neither the United States Government nor Lawrence Livermore National
Security, LLC nor any of their employees, makes any warranty, express or
implied, or assumes any liability or responsibility for the accuracy,
completeness, or usefulness of any information, apparatus, product, or process
disclosed, or represents that its use would not infringe privately-owned rights.

3. Also, reference herein to any specific commercial products, process, or
services by trade name, trademark, manufacturer or otherwise does not
necessarily constitute or imply its endorsement, recommendation, or favoring by
the United States Government or Lawrence Livermore National Security, LLC.
The views and opinions of authors expressed herein do not necessarily state or
reflect those of the United States Government or Lawrence Livermore National
Security, LLC, and shall not be used for advertising or product endorsement
purposes.

</license>

*/


/* eof */
