/* -*- C -*- 

   memP Heap Profiler ( http://memp.sourceforge.net/ )

   Please see COPYRIGHT AND LICENSE information at the end of this file.

   ----- 

   collect_data.c -- mechanism for collecting data for reporting

 */

#ifndef lint
static char *svnid = "$Id: collect_data.c 383 2006-12-08 18:30:15Z chcham $";
#endif

#include <string.h>
#include <strings.h>
#include <float.h>
#include <limits.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/resource.h>
#include "memPi.h"

extern h_t *callsite_pc_cache;


int
memPi_open_executable (char *exe)
{
  if (exe != NULL)
    memPi.appFullName = exe;

#ifdef ENABLE_BFD
  if (memPi.appFullName != NULL)
  {
    if (open_bfd_executable (memPi.appFullName) == 0)
      memPi.do_lookup = 0;
  }
#elif defined(USE_LIBDWARF)
  if (memPi.appFullName != NULL)
  {
    if (open_dwarf_executable (memPi.appFullName) == 0)
      memPi.do_lookup = 0;
  }
#endif
#if defined(ENABLE_BFD) || defined(USE_LIBDWARF)
  else
  {
    memPi_msg_warn ("Failed to open executable\n");
    memPi.do_lookup = 0;
  }
#endif

  return memPi.do_lookup;
}

void
memPi_close_executable ()
{
#ifdef ENABLE_BFD
  close_bfd_executable ();
#elif defined(USE_LIBDWARF)
  close_dwarf_executable ();
#endif
}

int
memPi_mergeResults ()
{
  int ac = 0;
  callsite_stats_t *hwm_cs;
  int totalCount = 0;
  size_t maxCount = 0;
  int retval = 1;
  int hwm_callsite_count = 1;   /* This may be increased in the future  */
  callsite_stats_t *remoteCallsiteBuffer = NULL;

  /*  Get statistics for local HWM call site  */
  hwm_cs = memPi_get_callsite (memPi.taskHWMop, memPi.rank, memPi.taskHWMpcs);

  if (hwm_cs == NULL && memPi.taskHWMpcs[0] != NULL)
  {
    //  Insert task HWM call site info from memPi.task* variables.
    hwm_cs =
      memPi_update_callsite_stats (memPi.taskHWMop, memPi.rank,
                                   memPi.taskHWMpcs, 0,
                                   (double) memPi.taskHWMsize, (double) 0);

#if 0
    fprintf (stderr,
             "Inserted taskHWM callsite for rank %d, op %d, pc[0] 0x%p\n",
             memPi.rank, memPi.taskHWMop, memPi.taskHWMpcs[0]);
    if (hwm_cs == NULL)
      fprintf (stderr,
               "**failed to find HWM callsite for rank %d, op %d, pc[0] 0x%p!\n",
               memPi.rank, memPi.taskHWMop, memPi.taskHWMpcs[0]);
    else
      fprintf (stderr, "**added callsite for rank %d, op %d, pc[0] 0x%p!\n",
               memPi.rank, memPi.taskHWMop, memPi.taskHWMpcs[0]);
    memPi_print_stack (memPi.taskHWMpcs);
#endif
  }


  /*  Check for any records to report  */
  if (hwm_cs != NULL)
    ac = 1;
  else
    ac = 0;

  PMPI_Allreduce (&ac, &totalCount, 1, MPI_INT, MPI_SUM, memPi.comm);
  PMPI_Reduce (&ac, &maxCount, 1, MPI_INT, MPI_MAX, memPi.collectorRank,
               memPi.comm);

  if (totalCount < 1 && memPi.traceAllStacks == 1)
  {
    if (memPi.rank == memPi.collectorRank)
    {
      memPi_msg_warn
        ("Collector found no records to merge. Omitting report.\n");
    }
    return 0;
  }

  if (memPi.traceAllStacks == 1)
  {
    /*  Receive data from other tasks at collector and insert data
     *  into local hash tables.
     */
    if (memPi.rank == memPi.collectorRank)
    {
      int i;

      /*  Create call site statistic hash tables  */
      memPi.global_callsite_stats = h_open (memPi.tableSize,
                                            memPi_callsite_stats_src_hashkey,
                                            memPi_callsite_stats_src_comparator);
      memPi.global_callsite_stats_agg = h_open (memPi.tableSize,
                                                memPi_callsite_stats_src_id_hashkey,
                                                memPi_callsite_stats_src_id_comparator);


      /*  Open call site PC and source hash tables  */
      if (callsite_pc_cache == NULL)
      {
        callsite_pc_cache = h_open (memPi.tableSize,
                                    callsite_pc_cache_hashkey,
                                    callsite_pc_cache_comparator);
      }
      if (callsite_src_id_cache == NULL)
      {
        callsite_src_id_cache = h_open (memPi.tableSize,
                                        callsite_src_id_cache_hashkey,
                                        callsite_src_id_cache_comparator);
      }

      /*  Allocate space for largest number of call sites of any task  */
      remoteCallsiteBuffer =
        (callsite_stats_t *) memPi_calloc (maxCount,
                                           sizeof (callsite_stats_t));

      if (remoteCallsiteBuffer == NULL)
      {
        memPi_msg_warn ("Failed to allocate memory to collect callsite info");
        retval = 0;
      }

      if (retval == 1)
      {
        /* Insert collector callsite data into global and task-specific hash tables */
        if (hwm_cs != NULL)
        {
          memPi_insert_callsite_records (hwm_cs);
          memPi_msg_debug
            ("Collector Inserted call site for task %d, op %d, ps[0] 0x%p\n",
             hwm_cs->rank, hwm_cs->op, hwm_cs->pc[0]);
          if (memPi_debug >= MEMP_DEBUG_STACK)
            memPi_print_stack (hwm_cs->pc);
        }

        for (i = 1; i < memPi.size; i++)        /* n-1 */
        {
          MPI_Status status;
          int count;

          /* okay in any order */
          PMPI_Probe (MPI_ANY_SOURCE, memPi.tag, memPi.comm, &status);
          PMPI_Get_count (&status, MPI_BYTE, &count);
          PMPI_Recv (remoteCallsiteBuffer, count, MPI_BYTE,
                     status.MPI_SOURCE, memPi.tag, memPi.comm, &status);
          count /= sizeof (callsite_stats_t);

          /*  Count is ignored at this point, 
             as currently only 1 HWM call site is supported.
           */
          memPi_insert_callsite_records (remoteCallsiteBuffer);
        }
        memPi_free (remoteCallsiteBuffer);
      }



      /* Closing the file opened by the BFD library appears to result
         in a segfault if the file is reopened.
         memPi_close_executable ();
       */
    }
    else                        /* non-collector tasks pack/send call site statistics to collector */
    {
      int ndx;
      char *sbuf =
        (char *) memPi_malloc (hwm_callsite_count *
                               sizeof (callsite_stats_t));

      for (ndx = 0; ndx < hwm_callsite_count; ndx++)
      {
        bcopy (hwm_cs,
               &(sbuf[ndx * sizeof (callsite_stats_t)]),
               sizeof (callsite_stats_t));
      }

      PMPI_Send (sbuf, (int) (hwm_callsite_count * sizeof (callsite_stats_t)),
                 MPI_BYTE, memPi.collectorRank, memPi.tag, memPi.comm);

      memPi_free (sbuf);
    }

    if (memPi.rank == memPi.collectorRank && retval == 1)
    {
      if (memPi.collective_report == 0)
        memPi_msg_debug
          ("MEMORY : Allocated for global_callsite_stats     : %13ld\n",
           h_count (memPi.global_callsite_stats) * sizeof (callsite_stats_t));
      memPi_msg_debug
        ("MEMORY : Allocated for global_callsite_stats_agg : %13ld\n",
         h_count (memPi.global_callsite_stats_agg) *
         sizeof (callsite_stats_t));
    }
  }

  return memPi_verify_unanimous (retval);
}


void
memPi_publishResults (int report_style)
{
  FILE *fp = NULL;
  static int printCount = 0;

  if (memPi.collectorRank == memPi.rank)
  {

    /* Generate output filename, and open */
    do
    {
      printCount++;
      snprintf (memPi.oFilename, 256, "%s/%s.%d.%d.%d.memP",
                memPi.outputDir, memPi.appName, memPi.size, memPi.procID,
                printCount);
    }
    while (access (memPi.oFilename, F_OK) == 0);

    fp = fopen (memPi.oFilename, "w");

    if (fp == NULL)
    {
      memPi_msg_warn ("Could not open [%s], writing to stdout\n",
                      memPi.oFilename);
      fp = stdout;
    }
    else
    {
      memPi_msg ("\n");
      memPi_msg ("Storing memP output in [%s].\n", memPi.oFilename);
      memPi_msg ("\n");
    }
  }
  memPi_profile_print (fp, report_style);
  PMPI_Barrier (MPI_COMM_WORLD);
  if (fp != stdout && fp != NULL)
  {
    fclose (fp);
  }
}

void
print_mallinfo ()
{
  struct mallinfo mi;

  mi = mallinfo ();

  fprintf (stderr, "mallinfo : arena    : %u\n", mi.arena);
  fprintf (stderr, "mallinfo : usmblks  : %u\n", mi.usmblks);
  fprintf (stderr, "mallinfo : uordblks : %u\n", mi.uordblks);
  fprintf (stderr, "mallinfo : hblkhd   : %u\n", mi.hblkhd);
  fprintf (stderr, "mallinfo : fordblks : %u\n", mi.fordblks);
  fprintf (stderr, "mallinfo : in use   : %u\n",
           (mi.arena - mi.fordblks) + mi.hblkhd);
  fprintf (stderr, "memPi.taskHWM       : %f\n", memPi.taskHWM);
  fprintf (stderr, "memPi.unknownHeapUse: %u\n", memPi.unknownHeapUse);
  fprintf (stderr, "memPi total in use  : %f\n",
           memPi.taskHWM + (double) memPi.unknownHeapUse);

  fprintf (stderr, "\n");
}

/*
 * memPi_collect_basics() - all tasks send their basic info to the
 * collectorRank.
 */
void
memPi_collect_basics ()
{
  int i = 0, j = 0;
  double app_time = memPi.cumulativeTime;
  int cnt;
  memPi_task_info_t mti;
#define task_fields 18

  int blockcounts[task_fields] =
    { 1, 1, 1, MEMPI_HOSTNAME_LEN_MAX, 1, 1, 1, 1, 1, 1, 1,
    (int) sizeof (memPi_TIME),
    1, 1, 1, 1,
    MEMP_CALLSITE_STACK_DEPTH_MAX
  };
  MPI_Datatype types[task_fields] =
    { MPI_DOUBLE, MPI_DOUBLE, MPI_INT, MPI_CHAR, MPI_UNSIGNED_LONG,
    MPI_UNSIGNED_LONG, MPI_DOUBLE, MPI_DOUBLE, MPI_DOUBLE, MPI_DOUBLE,
    MPI_DOUBLE,
    MPI_BYTE, MPI_INT, MPI_DOUBLE, MPI_DOUBLE, MPI_DOUBLE,
    MPI_LONG
  };
  MPI_Aint displs[task_fields];
  MPI_Datatype mti_type;
  MPI_Request *recv_req_arr;

  memPi_msg_debug ("Collect Basics\n");

  cnt = 0;

  PMPI_Address (&mti.mpi_time, &displs[cnt++]);
  PMPI_Address (&mti.app_time, &displs[cnt++]);
  PMPI_Address (&mti.rank, &displs[cnt++]);
  PMPI_Address (&mti.hostname, &displs[cnt++]);
  PMPI_Address (&mti.task_hwm, &displs[cnt++]);
  PMPI_Address (&mti.task_alloc_cnt, &displs[cnt++]);
  PMPI_Address (&mti.taskHWM, &displs[cnt++]);
  PMPI_Address (&mti.taskHWMunknown, &displs[cnt++]);
  PMPI_Address (&mti.taskHWMsize, &displs[cnt++]);
  PMPI_Address (&mti.taskHWMtaskCount, &displs[cnt++]);
  PMPI_Address (&mti.taskHWMcsCount, &displs[cnt++]);
  PMPI_Address (&mti.taskHWMts, &displs[cnt++]);
  PMPI_Address (&mti.taskHWMop, &displs[cnt++]);
  PMPI_Address (&mti.taskLWM, &displs[cnt++]);
  PMPI_Address (&mti.taskHWMstackUse, &displs[cnt++]);
  PMPI_Address (&mti.taskHWMheapAvail, &displs[cnt++]);
  PMPI_Address (&mti.taskHWMpcs, &displs[cnt++]);

  for (i = (cnt - 1); i >= 0; i--)
  {
    displs[i] -= displs[0];
  }

  PMPI_Type_struct (cnt, blockcounts, displs, types, &mti_type);
  if (MPI_SUCCESS != PMPI_Type_commit (&mti_type))
    memPi_abort ("Failed PMPI_Type_commit");

  if (memPi.rank == memPi.collectorRank)
  {
    /* In the case where multiple reports are generated per run,
       only allocate memory for global_task_info once */
    if (memPi.global_task_info == NULL)
    {
      memPi.global_task_info =
        (memPi_task_info_t *) memPi_calloc ((size_t) memPi.size,
                                            sizeof (memPi_task_info_t));
      if (memPi.global_task_info == NULL)
        memPi_abort ("Failed to allocate memory for global_task_info");

      memPi_msg_debug
        ("MEMORY : Allocated for global_task_info :          %13ld\n",
         memPi.size * sizeof (memPi_task_info_t));
    }

    bzero (memPi.global_task_info, memPi.size * sizeof (memPi_task_info_t));

    recv_req_arr =
      (MPI_Request *) memPi_malloc (sizeof (MPI_Request) * memPi.size);

    if (recv_req_arr == NULL)
      memPi_abort ("Failed to allocate memory for recv_req_arr");

    for (i = 0; i < memPi.size; i++)
    {
      memPi_task_info_t *p = &memPi.global_task_info[i];
      if (i != memPi.collectorRank)
      {
        memPi_msg_debug
          ("Collect basics Remote task rank is %d, p is %p, r is %p\n",
           i, p, &(recv_req_arr[i]));
        PMPI_Irecv (p, 1, mti_type, i, memPi.tag, memPi.comm,
                    &(recv_req_arr[i]));
      }
      else
      {
        strcpy (p->hostname, memPi.hostname);
        p->app_time = app_time;
        p->rank = memPi.rank;
        recv_req_arr[i] = MPI_REQUEST_NULL;
        PMPI_Address (&mti.taskHWMcsCount, &displs[cnt++]);
        p->task_hwm = memPi.task_hwm;
        p->task_alloc_cnt = memPi.task_alloc_cnt;
        p->taskHWM = memPi.taskHWM;
        p->taskHWMunknown = memPi.taskHWMunknown;
        p->taskHWMsize = memPi.taskHWMsize;
        p->taskHWMtaskCount = memPi.taskHWMtaskCount;
        p->taskHWMcsCount = memPi.taskHWMcsCount;
        p->taskHWMts = memPi.taskHWMts;
        p->taskHWMop = memPi.taskHWMop;
        p->taskLWM = memPi.taskLWM;
        p->taskHWMstackUse = memPi.taskHWMstackUse;
        p->taskHWMheapAvail = memPi.taskHWMheapAvail;
        for (j = 0; j < MEMP_CALLSITE_STACK_DEPTH_MAX; j++)
          p->taskHWMpcs[j] = memPi.taskHWMpcs[j];
      }
    }
    PMPI_Waitall (memPi.size, recv_req_arr, MPI_STATUSES_IGNORE);

#if 0
    fprintf (stderr, "Received task 1 pc array of :\n");
    memPi_print_stack (memPi.global_task_info[1].taskHWMpcs);
    fprintf (stderr, "op %d\n", memPi.global_task_info[1].taskHWMop);
#endif

    memPi_free (recv_req_arr);
    /* task MPI time is calculated from callsites data 
       in memPi_insert_callsite_records.
     */
    for (i = 0; i < memPi.size; i++)
      memPi.global_task_info[i].mpi_time = 0.0;
  }
  else
  {
    strcpy (mti.hostname, memPi.hostname);
    mti.app_time = app_time;
    mti.rank = memPi.rank;
    mti.task_hwm = memPi.task_hwm;
    mti.task_alloc_cnt = memPi.task_alloc_cnt;
    mti.taskHWM = memPi.taskHWM;
    mti.taskHWMunknown = memPi.taskHWMunknown;
    mti.taskHWMsize = memPi.taskHWMsize;
    mti.taskHWMtaskCount = memPi.taskHWMtaskCount;
    mti.taskHWMcsCount = memPi.taskHWMcsCount;
    mti.taskHWMts = memPi.taskHWMts;
    mti.taskHWMop = memPi.taskHWMop;
    mti.taskLWM = memPi.taskLWM;
    mti.taskHWMstackUse = memPi.taskHWMstackUse;
    mti.taskHWMheapAvail = memPi.taskHWMheapAvail;
    for (j = 0; j < MEMP_CALLSITE_STACK_DEPTH_MAX; j++)
      mti.taskHWMpcs[j] = memPi.taskHWMpcs[j];


    PMPI_Send (&mti, 1, mti_type, memPi.collectorRank, memPi.tag, memPi.comm);
  }

  PMPI_Type_free (&mti_type);

  return;
}


void
memPi_generateReport (int report_style)
{
  memP_TIMER dur;
  memPi_TIME timer_start, timer_end;
  int mergeResult;

  memPi_GETTIME (&memPi.endTime);

  if (memPi.enabled)
  {
    dur = (memPi_GETTIMEDIFF (&memPi.endTime, &memPi.startTime) / 1000000.0);
    memPi.cumulativeTime += dur;
    assert (memPi.cumulativeTime >= 0);
    if (memPi.in_finalize == 0)
      memPi_GETTIME (&memPi.startTime);
  }

  if (time (&memPi.stop_timeofday) == (time_t) - 1)
  {
    memPi_msg_warn ("Could not get time of day from time()\n");
  }

  /*  Open executable for source lookup during call site hash insertion.
     Sets memPi.do_lookup to handle subsequent source lookup behavior.
     Open bfd even if not doing source lookup so that we can get 
     text/data/bss info.
   */
  (void) memPi_open_executable (NULL);

  /* We collect task data in different ways depending on the report format:
   *
   * minimal : scalable approach just provides max/mean/min
   * concise : collector allocates sizeof(double)*wsize to get stats
   * verbose : includes call site information
   *
   */

  /* Can we allocate enough memory to perform median calculation? */
  {
    char *testptr;
    if (memPi.report_style != memPi_style_text_minimal)
    {
      if ((testptr = memPi_malloc (sizeof (double) * memPi.size)) == NULL)
        memPi.report_style = memPi_style_text_minimal;
      else
        memPi_free (testptr);
    }
  }

  /* For a concise report, we use MPI collectives to generate statistics
     and identify tasks.  The verbose reports collect additional information
     that needs more coordination.
   */
  if (memPi.report_style == memPi_style_text_minimal)
  {
    memPi_publishResults (memPi_style_text_minimal);
  }
  else
  {
    /* collect results and publish */
    memPi_msg_debug0 ("starting collect_basics\n");

    memPi_GETTIME (&timer_start);
    memPi_collect_basics ();
    memPi_GETTIME (&timer_end);
    dur = (memPi_GETTIMEDIFF (&timer_end, &timer_start) / 1000000.0);

    memPi_msg_debug0 ("TIMING : collect_basics_time is %12.6f\n", dur);

    memPi_msg_debug0 ("starting mergeResults\n");

    memPi_GETTIME (&timer_start);
#if 0
    if (memPi.traceAllStacks == 1)
      mergeResult = memPi_mergeResults ();
    else
      mergeResult = 1;
#else
    mergeResult = memPi_mergeResults ();
#endif
    memPi_GETTIME (&timer_end);
    dur = (memPi_GETTIMEDIFF (&timer_end, &timer_start) / 1000000.0);

    memPi_msg_debug0 ("TIMING : merge time is          %12.6f\n", dur);
    memPi_msg_debug0 ("starting publishResults\n");

    if (mergeResult == 1)
    {
      memPi_GETTIME (&timer_start);
      if (memPi.report_style == memPi_style_both)
      {
        memPi_publishResults (memPi_style_text_concise);
        memPi_publishResults (memPi_style_verbose);
      }
      else
        memPi_publishResults (report_style);
      memPi_GETTIME (&timer_end);
      dur = (memPi_GETTIMEDIFF (&timer_end, &timer_start) / 1000000.0);
      memPi_msg_debug0 ("TIMING : publish time is        %12.6f\n", dur);
    }
  }
}


/* 

<license>

Copyright (c) 2010, Lawrence Livermore National Security, LLC.
Produced at the Lawrence Livermore National Laboratory
Written by Christopher Chambreau <chcham@llnl.gov>.
CODE-OCEC-10-010
All rights reserved.

This file is part of memP. For details, see http://memp.sourceforge.net .
Please also read this link – Additional BSD Notice.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  o  Redistributions of source code must retain the above copyright notice, 
     this list of conditions and the disclaimer below.

  o  Redistributions in binary form must reproduce the above copyright notice, 
     this list of conditions and the disclaimer (as noted below) in the 
     documentation and/or other materials provided with the distribution.

  o  Neither the name of the LLNS/LLNL nor the names of its contributors may be
     used to endorse or promote products derived from this software without 
     specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL LAWRENCE LIVERMORE NATIONAL
SECURITY, LLC, THE U.S. DEPARTMENT OF ENERGY OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.


Additional BSD Notice

1. This notice is required to be provided under our contract with the U.S.
Department of Energy (DOE). This work was produced at Lawrence Livermore
National Laboratory under Contract No. DE-AC52-07NA27344 with the DOE.

2. Neither the United States Government nor Lawrence Livermore National
Security, LLC nor any of their employees, makes any warranty, express or
implied, or assumes any liability or responsibility for the accuracy,
completeness, or usefulness of any information, apparatus, product, or process
disclosed, or represents that its use would not infringe privately-owned rights.

3. Also, reference herein to any specific commercial products, process, or
services by trade name, trademark, manufacturer or otherwise does not
necessarily constitute or imply its endorsement, recommendation, or favoring by
the United States Government or Lawrence Livermore National Security, LLC.
The views and opinions of authors expressed herein do not necessarily state or
reflect those of the United States Government or Lawrence Livermore National
Security, LLC, and shall not be used for advertising or product endorsement
purposes.

</license>

*/


/* eof */
