/* -*- C -*- 

   memP Heap Profiler ( http://memp.sourceforge.net/ )

   Please see COPYRIGHT AND LICENSE information at the end of this file.

   ----- 

   find_internal_funcs.c -- interposition mechanisms

 */

#ifndef lint
static char *svnid =
  "$Id: find_internal_funcs.c 369 2006-10-05 19:21:12Z chcham $";
#endif

#include "memPconfig.h"
#include "memPi.h"
#include "symbols.h"
#include <string.h>

#ifdef HAVE_LIBDL
#define __USE_GNU
#include <dlfcn.h>
#endif


/******************************************************************************/
/*                     ALLOCATION FUNCTIONS TO INTERPOSE                      */
/******************************************************************************/

/*  TODO: add return check  */

void
memPi_set_internal_functions ()
{
  /*  NOTE: use of dlsym results in internal calls to allocation functions.
     To address this, we use pre-defined allocation symbols for most
     allocation functions.  For posix_memalign, this did not appear
     possible, so dlsym is used to find the function to use for
     interposing.
   */

  if (memPi.malloc_ptr == NULL)
  {
    MEMP_TRACE ("Before Assign malloc ptr");
#if defined(HAVE_LIBDL) && defined(USE_DLSYM)
    memPi.malloc_ptr = (void *(*)(size_t)) dlsym (RTLD_NEXT, "malloc");
    memPi_msg_debug ("Using dlsym to find internal function for %s\n",
                     "malloc");
    if (memPi.malloc_ptr == NULL)
      memPi.malloc_ptr = MEMP_MALLOC;
#else
    extern void *MEMP_MALLOC (size_t);
    memPi.malloc_ptr = MEMP_MALLOC;
    memPi_msg_debug ("Using %s as internal function for %s\n",
                     MEMP_MALLOC_NAME, __func__);
#endif
    MEMP_TRACE ("After Assign malloc ptr");
  }

  if (memPi.realloc_ptr == NULL)
  {
    MEMP_TRACE ("Before Assign realloc ptr");
#if defined(HAVE_LIBDL) && defined(USE_DLSYM)
    memPi.realloc_ptr = (void *(*)(size_t)) dlsym (RTLD_NEXT, "realloc");
    memPi_msg_debug ("Using dlsym to find internal function for %s\n",
                     "realloc");
    if (memPi.realloc_ptr == NULL)
      memPi.realloc_ptr = MEMP_REALLOC;
#else
    extern void *MEMP_REALLOC (void *, size_t);
    memPi.realloc_ptr = MEMP_REALLOC;
    memPi_msg_debug ("Using %s as internal function for %s\n",
                     MEMP_REALLOC_NAME, "realloc");
#endif
    MEMP_TRACE ("After Assign realloc ptr");
  }

  if (memPi.calloc_ptr == NULL)
  {
    MEMP_TRACE ("Before Assign calloc ptr");
#if defined(HAVE_LIBDL) && defined(USE_DLSYM)
    /* Cannot use dlsym to get calloc as dlsym calls calloc */
    //memPi_abort ("Undefined internal calloc call.\n");
    extern void *MEMP_CALLOC (size_t, size_t);
    memPi.calloc_ptr = MEMP_CALLOC;
#else
    extern void *MEMP_CALLOC (size_t, size_t);
    memPi.calloc_ptr = MEMP_CALLOC;
    memPi_msg_debug ("Using %s as internal function for %s\n",
                     MEMP_CALLOC_NAME, "calloc");
#endif
    MEMP_TRACE ("After Assign calloc ptr");
  }

  if (memPi.memalign_ptr == NULL)
  {
    MEMP_TRACE ("Before Assign memalign ptr");
#if defined(HAVE_LIBDL) && defined(USE_DLSYM)
    memPi.memalign_ptr =
      (void *(*)(size_t, size_t)) dlsym (RTLD_NEXT, "memalign");
    memPi_msg_debug ("Using dlsym to find internal function for %s\n",
                     "memalign");
    if (memPi.memalign_ptr == NULL)
      memPi.memalign_ptr = MEMP_MEMALIGN;
#else
    extern void *MEMP_MEMALIGN (size_t, size_t);
    memPi.memalign_ptr = MEMP_MEMALIGN;
    memPi_msg_debug ("Using %s as internal function for %s\n",
                     MEMP_MEMALIGN_NAME, "memalign");
#endif
    MEMP_TRACE ("After Assign memalign ptr");
  }

  if (memPi.free_ptr == NULL)
  {
    MEMP_TRACE ("Before Assign free ptr");
#if defined(HAVE_LIBDL) && defined(USE_DLSYM)
    memPi.free_ptr = (void (*)(void *)) dlsym (RTLD_NEXT, "free");
    memPi_msg_debug ("Using dlsym to find internal function for %s\n",
                     "free");
    if (memPi.free_ptr == NULL)
      memPi.free_ptr = MEMP_FREE;
#else
    extern void MEMP_FREE (void *);
    memPi.free_ptr = MEMP_FREE;
    memPi_msg_debug ("Using %s as internal function for %s\n",
                     MEMP_FREE_NAME, "free");
#endif
    MEMP_TRACE ("After Assign free ptr");
  }

#ifdef WRAP_POSIX_MEMALIGN
  if (memPi.posix_memalign_ptr == NULL)
  {
    MEMP_TRACE ("Before Assign posix_memalign ptr");
#if defined(HAVE_LIBDL)
    memPi.posix_memalign_ptr =
      (int (*)(void **, size_t, size_t)) dlsym (RTLD_NEXT, "posix_memalign");
    memPi_msg_debug ("Using dlsym to find internal function for %s\n",
                     "posix_memalign");
    if (memPi.posix_memalign_ptr == NULL)
      memPi_abort ("Cannot use dlsym to get internal posix_memalign call.\n");

#else
#if defined(MEMP_POSIX_MEMALIGN)

    extern int MEMP_POSIX_MEMALIGN (void **, size_t, size_t);
    memPi.posix_memalign_ptr = MEMP_POSIX_MEMALIGN;
    memPi_msg_debug ("Using %s as internal function for %s\n",
                     MEMP_POSIX_MEMALIGN_NAME, "posix_memalign");

#else

    /* Cannot use internal symbol to access posix_memalign */
    memPi_abort ("Undefined internal posix_memalign call.\n");

#endif
#endif
    MEMP_TRACE ("After Assign posix_memalign ptr");
  }
#endif

  if (NULL == memPi.malloc_ptr)
    fputs ("Failed to wrap malloc!\n", stderr);
  if (NULL == memPi.realloc_ptr)
    fputs ("Failed to wrap realloc!\n", stderr);
  if (NULL == memPi.calloc_ptr)
    fputs ("Failed to wrap calloc!\n", stderr);
  if (NULL == memPi.free_ptr)
    fputs ("Failed to wrap free!\n", stderr);
  if (NULL == memPi.memalign_ptr)
    fputs ("Failed to wrap memalign!\n", stderr);
#ifdef WRAP_POSIX_MEMALIGN
  if (NULL == memPi.posix_memalign_ptr)
    fputs ("Failed to wrap posix_memalign!\n", stderr);
#endif

  if (!memPi.malloc_ptr || !memPi.realloc_ptr || !memPi.calloc_ptr
      || !memPi.free_ptr || !memPi.memalign_ptr
#ifdef WRAP_POSIX_MEMALIGN
      || !memPi.posix_memalign_ptr
#endif
    )
    MPI_Abort (MPI_COMM_WORLD, 1);
}



/* 

<license>

Copyright (c) 2010, Lawrence Livermore National Security, LLC.
Produced at the Lawrence Livermore National Laboratory
Written by Christopher Chambreau <chcham@llnl.gov>.
CODE-OCEC-10-010
All rights reserved.

This file is part of memP. For details, see http://memp.sourceforge.net .
Please also read this link – Additional BSD Notice.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  o  Redistributions of source code must retain the above copyright notice, 
     this list of conditions and the disclaimer below.

  o  Redistributions in binary form must reproduce the above copyright notice, 
     this list of conditions and the disclaimer (as noted below) in the 
     documentation and/or other materials provided with the distribution.

  o  Neither the name of the LLNS/LLNL nor the names of its contributors may be
     used to endorse or promote products derived from this software without 
     specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL LAWRENCE LIVERMORE NATIONAL
SECURITY, LLC, THE U.S. DEPARTMENT OF ENERGY OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.


Additional BSD Notice

1. This notice is required to be provided under our contract with the U.S.
Department of Energy (DOE). This work was produced at Lawrence Livermore
National Laboratory under Contract No. DE-AC52-07NA27344 with the DOE.

2. Neither the United States Government nor Lawrence Livermore National
Security, LLC nor any of their employees, makes any warranty, express or
implied, or assumes any liability or responsibility for the accuracy,
completeness, or usefulness of any information, apparatus, product, or process
disclosed, or represents that its use would not infringe privately-owned rights.

3. Also, reference herein to any specific commercial products, process, or
services by trade name, trademark, manufacturer or otherwise does not
necessarily constitute or imply its endorsement, recommendation, or favoring by
the United States Government or Lawrence Livermore National Security, LLC.
The views and opinions of authors expressed herein do not necessarily state or
reflect those of the United States Government or Lawrence Livermore National
Security, LLC, and shall not be used for advertising or product endorsement
purposes.

</license>

*/


/* eof */
