/* -*- C -*- 

   memP Heap Profiler ( http://memp.sourceforge.net/ )

   Please see COPYRIGHT AND LICENSE information at the end of this file.

   ----- 

   memPi.h -- internal memP header

   $Id: memPi.h.in 405 2007-04-26 16:15:44Z chcham $

*/

#ifndef _MEMPI_H
#define _MEMPI_H

#include <assert.h>
#include <malloc.h>
#include <math.h>
#include <setjmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>

#include "memPconfig.h"

#if !defined(CEXTRACT)
#include <mpi.h>
#include "memPi_def.h"
#endif

#include "memP-hash.h"

#include "memp_timers.h"


/* Internal C lib malloc symbols */
#define MEMP_MALLOC __libc_malloc
#define MEMP_CALLOC __libc_calloc
#define MEMP_REALLOC __libc_realloc
#define MEMP_MEMALIGN __libc_memalign
#define MEMP_FREE __libc_free
#ifdef HAVE_INTERNAL_POSIX_MEMALIGN
#define MEMP_POSIX_MEMALIGN 
#endif

#define MEMP_STR(x) MEMP_ISTR(x)
#define MEMP_ISTR(x) #x
#define MEMP_MALLOC_NAME MEMP_STR(MEMP_MALLOC)
#define MEMP_CALLOC_NAME MEMP_STR(MEMP_CALLOC)
#define MEMP_REALLOC_NAME MEMP_STR(MEMP_REALLOC)
#define MEMP_MEMALIGN_NAME MEMP_STR(MEMP_MEMALIGN)
#define MEMP_FREE_NAME MEMP_STR(MEMP_FREE)
#ifdef HAVE_INTERNAL_POSIX_MEMALIGN
#define MEMP_POSIX_MEMALIGN_STR(MEMP_POSIX_MEMALIGN)
#endif

#define MEMPI_HOSTNAME_LEN_MAX 128
#define MEMP_CALLSITE_STACK_DEPTH_MAX 16

#define MEMP_HELP_LIST "chcham@llnl.gov"

#define MEMP_CALLSITE_STACK_DEPTH (memPi.stackDepth)
#define MEMP_CALLSITE_STATS_COOKIE 518641
#define MEMP_CALLSITE_STATS_COOKIE_ASSERT(f) {assert(MEMP_CALLSITE_STATS_COOKIE==((f)->cookie));}
#define MiB (1<<20)
#define MB (1000000)

/*#define TRACE_MEMP_CALLS*/
#ifdef TRACE_MEMP_CALLS
#define MEMP_TRACE(s) fputs("*** MEMP_TRACE : ", stderr);fputs(s, stderr);fputc('\n',stderr)
#else
#define MEMP_TRACE(s) ;
#endif

typedef struct _callsite_stats
{
  unsigned op;
  int rank;
  int csid;
  long long count;
  double cumulativeTime;
  double maxAlloc;
  double minAlloc;
  double maxFree;
  double minFree;
  double cumulativeAlloc;
  double cumulativeFree;

  double csHWM;
  double csLWM;
  double csCurAlloc;

  void *pc[MEMP_CALLSITE_STACK_DEPTH_MAX];
  char *filename[MEMP_CALLSITE_STACK_DEPTH_MAX];
  char *functname[MEMP_CALLSITE_STACK_DEPTH_MAX];
  unsigned int lineno[MEMP_CALLSITE_STACK_DEPTH_MAX];
  long cookie;
}
callsite_stats_t;

typedef struct callsite_src_id_cache_entry_t
{
  int id;                       /* unique id for this src code/stack location */
  unsigned int op;              /* at the lowest level, this is a MPI op */
  char *filename[MEMP_CALLSITE_STACK_DEPTH_MAX];
  char *functname[MEMP_CALLSITE_STACK_DEPTH_MAX];
  unsigned int line[MEMP_CALLSITE_STACK_DEPTH_MAX];
  void *pc[MEMP_CALLSITE_STACK_DEPTH_MAX];
}
callsite_src_id_cache_entry_t;

extern h_t *callsite_src_id_cache;


typedef struct _memPi_task_info_t
{
  /* this section is calculated at the collector */
  double mpi_time;

  /* this section is received from each task */
  double app_time;              /* time from end of init to start of finalize */
  int rank;
  char hostname[MEMPI_HOSTNAME_LEN_MAX];

  unsigned long long task_hwm;
  unsigned long task_alloc_cnt;
  double taskHWM;
  double taskHWMunknown;
  double taskHWMsize;
  double taskHWMtaskCount;
  double taskHWMcsCount;
  memPi_TIME taskHWMts;
  unsigned int taskHWMop;
  double taskLWM;
  double taskHWMstackUse;
  double taskHWMheapAvail;
  void *taskHWMpcs[MEMP_CALLSITE_STACK_DEPTH_MAX];
}
memPi_task_info_t;

typedef struct _memPi_lookup_t
{
  int op;
  char *name;
}
memPi_lookup_t;

typedef struct callsite_cache_entry_t
{
  void *pc;
  char *filename;
  char *functname;
  unsigned int line;
}
callsite_pc_cache_entry_t;

extern memPi_lookup_t memPi_lookup[];

/*  
 *  Debug levels
 *
 *  MEMP_DEBUG_OFF   : off
 *  MEMP_DEBUG_MAJ   : Entry/Exit of major functions
 *  MEMP_DEBUG_MIN   : Entry/Exit of minor functions
 *  MEMP_DEBUG_ALLOC : Detailed allocation/free information 
 *  MEMP_DEBUG_STACK : Stack trace information
 *  MEMP_DEBUG_HASH  : Hash table details
 */
typedef enum
{ MEMP_DEBUG_OFF =
    0, MEMP_DEBUG_MAJ, MEMP_DEBUG_MIN, MEMP_DEBUG_ALLOC, MEMP_DEBUG_STACK,
  MEMP_DEBUG_HASH
} MEMP_DEBUG_LEVEL;

typedef enum
{ MEMP_REPORT_SCI_FORMAT, MEMP_REPORT_FLT_FORMAT }
MEMP_REPORT_FORMAT_TYPE;

typedef struct _alloc_stats_t
{
  long long alloc_max;
  long long alloc_cnt;
  long long alloc_tot;
  long long alloc_cur;
  long long free_cnt;
  long long free_tot;
} alloc_stats_t;

typedef struct _memPi_t
{
  int tableSize;
  int ac;
  char *av[32];
  char *toolname;
  char *appName;
  char *appFullName;
  char oFilename[256];
  int tag;
  int procID;
  int rank;
  int size;
  int collectorRank;
  MPI_Comm comm;
  char hostname[MPI_MAX_PROCESSOR_NAME];
  int hostnamelen;
  char *outputDir;
  char *envStr;
  FILE *stdout_;
  FILE *stderr_;
  memPi_TIME startTime;
  memPi_TIME endTime;

  double cumulativeTime;        /* necessary for pcontrol */
  time_t start_timeofday;
  time_t stop_timeofday;

  /* pcontrol */
  int enabled;
  int enabledCount;

  int afterInit;
  int set_nonmpi_values;

  memPi_task_info_t *global_task_info;

  h_t *task_callsite_stats;
  h_t *global_callsite_stats;
  h_t *global_callsite_stats_agg;

  memPi_lookup_t *lookup;

  int stackDepth;
  int baseNames;
  MEMP_REPORT_FORMAT_TYPE reportFormat;
  int do_lookup;
  enum memPi_report_style
  { memPi_style_verbose, memPi_style_text_concise, memPi_style_text_minimal,
    memPi_style_both, memPi_style_xml_verbose
  } report_style;
  int print_callsite_detail;
  int collective_report;
  int in_finalize;
  alloc_stats_t int_alloc;
  alloc_stats_t app_alloc;

  unsigned long long task_hwm;
  unsigned long task_alloc_cnt;
  double taskHWM;
  double taskHWMunknown;
  double taskHWMsize;
  double taskHWMtaskCount;
  double taskHWMcsCount;
  memPi_TIME taskHWMts;
  unsigned int taskHWMop;
  double taskLWM;
  void *taskHWMpcs[MEMP_CALLSITE_STACK_DEPTH_MAX];
  double taskHWMreportThreshold;
  double taskHWMstackUse;
  double taskHWMheapAvail;
  int taskHWMreportMaxCount;
  int taskHWMrank;
  int abortAtThreshold;
  int skipFrames;
  int reportHWMprintCount;
  size_t heapUseAtInit;
  size_t unknownHeapUse;
  int traceAllStacks;           /* For task report, save call site info for all allocs */
  int inWrappedFunc;
  void *topOfStack;
  size_t bss;
  size_t data;
  size_t text;

  void *(*malloc_ptr) (size_t);
  void *(*calloc_ptr) (size_t, size_t);
  void *(*realloc_ptr) (void *, size_t);
  void *(*memalign_ptr) (size_t, size_t);
  int (*posix_memalign_ptr) (void **, size_t, size_t);
  void (*free_ptr) (void *);
  h_t *alloc_ptr_hash;
  int in_init_wrappers;
}
memPi_t;

extern memPi_t memPi;

extern int memPi_vmajor;
extern int memPi_vminor;
extern int memPi_vpatch;
extern char *memPi_vdate;
extern char *memPi_vtime;
extern char *memPi_toolname;

extern int memPi_debug;
extern int memPi_do_demangle;

#ifdef HAVE_MPIR_TOPOINTER
extern void *MPIR_ToPointer (int idx);
#endif

#if !defined(UNICOS_mp)

/*  AIX  */
#if defined(AIX)

#if defined(__64BIT__)          /*  64-bit AIX  */
#define ParentFP(jb) ((void *) *(long *) jb[5])
#else /*  32-bit AIX  */
#define ParentFP(jb) ((void *) *(long *) jb[3])
#endif

/*  For both 32-bit and 64-bit AIX  */
#define FramePC(fp) ((void *) *(long *) (((long) fp) + (2 * sizeof (void *))))
#define NextFP(fp) ((void *) *(long *) fp)

/*  IA32 Linux  */
#elif BUILD_OS == linux-gnu && defined(IA32)
#define ParentFP(jb) ((void*) jb[0].__jmpbuf[3])
#define FramePC(fp) ((void*)(((void**)fp)[1]))
#define NextFP(fp) ((void*)((void**)fp)[0])

/*  X86_64 Linux  */
#elif BUILD_OS == linux-gnu && defined(X86_64)
#define ParentFP(jb) ((void*) jb[0].__jmpbuf[1])
#define FramePC(fp) ((void*)(((void**)fp)[1]))
#define NextFP(fp) ((void*)((void**)fp)[0])

/*  BG/L  */
#elif defined(ppc64)
#define ParentFP(jb) NextFP(((void*)jb[0].__jmpbuf[0]))
#define FramePC(fp) ((void*)(((char**)fp)[1]))
#define NextFP(fp) ((void*)((char**)fp)[0])

/* Catamount */
#elif defined(Catamount) && defined(X86_64)
#define ParentFP(jb) ((void*) jb[0].__jmpbuf[1])
#define FramePC(fp) ((void*)(((void**)fp)[1]))
#define NextFP(fp) ((void*)((void**)fp)[0])

/* MIPS+GCC */
/* We don't use the macros on the mips and instead
 * use gcc's __builtin_return_address(0) intrinsic
 * to get a single level of stack backtrace. The
 * value is saved in saved_ret_addr variable each time
 * setjmp is called.
 */
#elif defined(mips) && defined(__GNUC__)
#define ParentFP(jb) (0)
#define FramePC(fp) (0)
#define NextFP(fp) (0)
extern void *saved_ret_addr;


/*  undefined  */
#else
#if ! defined(HAVE_LIBUNWIND)
#warning "No stack trace mechanism defined for this platform."
#endif
#define ParentFP(jb) (0)
#define FramePC(fp) (0)
#define NextFP(fp) (0)
#endif

#else /* defined(UNICOS_mp) */

#include <intrinsics.h>
#include <stdint.h>

/*
 * Cray X1
 * Stacks grow downward in the address space.
 */


/*
 * Unlike some of the other platforms, we do not use setjmp to obtain
 * the frame pointer of the current function.  Instead, we use the
 * Cray intrinsic function _read_fp() to get the frame pointer of the 
 * current function.
 */
#define GetFP()    ((void*)(_read_fp()))


/* 
 * Given a frame pointer for a callee, the caller's frame pointer is at the 
 * callee's frame pointer address.
 */
#define NextFP(fp)      ((void*)((void**)fp)[0])


/*
 * Given a frame pointer for a callee, the return address to the caller
 * is two elements in the callee's frame.

 * TODO try masking off the high 32-bits
 */
#define FramePC(fp)     ((void*)(((uint64_t)(((void**)fp)[-2])) & 0xFFFFFFFF))


#endif /* defined(UNICOS_mp) */


#define min(x,y) ((x<y)?(x):(y))
#define max(x,y) ((x>y)?(x):(y))

#include "memPi_proto.h"

#endif

/*
  
<license>

Copyright (c) 2010, Lawrence Livermore National Security, LLC.
Produced at the Lawrence Livermore National Laboratory
Written by Christopher Chambreau <chcham@llnl.gov>.
CODE-OCEC-10-010
All rights reserved.

This file is part of memP. For details, see http://memp.sourceforge.net .
Please also read this link – Additional BSD Notice.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  o  Redistributions of source code must retain the above copyright notice, 
     this list of conditions and the disclaimer below.

  o  Redistributions in binary form must reproduce the above copyright notice, 
     this list of conditions and the disclaimer (as noted below) in the 
     documentation and/or other materials provided with the distribution.

  o  Neither the name of the LLNS/LLNL nor the names of its contributors may be
     used to endorse or promote products derived from this software without 
     specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL LAWRENCE LIVERMORE NATIONAL
SECURITY, LLC, THE U.S. DEPARTMENT OF ENERGY OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.


Additional BSD Notice

1. This notice is required to be provided under our contract with the U.S.
Department of Energy (DOE). This work was produced at Lawrence Livermore
National Laboratory under Contract No. DE-AC52-07NA27344 with the DOE.

2. Neither the United States Government nor Lawrence Livermore National
Security, LLC nor any of their employees, makes any warranty, express or
implied, or assumes any liability or responsibility for the accuracy,
completeness, or usefulness of any information, apparatus, product, or process
disclosed, or represents that its use would not infringe privately-owned rights.

3. Also, reference herein to any specific commercial products, process, or
services by trade name, trademark, manufacturer or otherwise does not
necessarily constitute or imply its endorsement, recommendation, or favoring by
the United States Government or Lawrence Livermore National Security, LLC.
The views and opinions of authors expressed herein do not necessarily state or
reflect those of the United States Government or Lawrence Livermore National
Security, LLC, and shall not be used for advertising or product endorsement
purposes.

</license>
  
*/

/* EOF */
