/* -*- C -*- 

   memP Heap Profiler ( http://memp.sourceforge.net/ )

   Please see COPYRIGHT AND LICENSE information at the end of this file.

   ----- 

   memPi_mem.c -- internal memory allocation functions

 */

#ifndef lint
static char *svnid = "$Id: memPi_mem.c 383 2006-12-08 18:30:15Z chcham $";
#endif

#include "memPi.h"

#if 1

extern void *MEMP_MALLOC (size_t);
extern void *MEMP_CALLOC (size_t, size_t);
extern void *MEMP_REALLOC (void *, size_t);
extern int MEMP_MEMALIGN (void **, size_t, size_t);
extern void MEMP_FREE (void *);

void *(*memPi_malloc_ptr) (size_t) = MEMP_MALLOC;
void *(*memPi_calloc_ptr) (size_t, size_t) = MEMP_CALLOC;
void *(*memPi_realloc_ptr) (void *, size_t) = MEMP_REALLOC;
int (*memPi_memalign_ptr) (void **, size_t, size_t) = MEMP_MEMALIGN;
void (*memPi_free_ptr) (void *) = MEMP_FREE;
#else
void *(*memPi_malloc_ptr) (size_t) = NULL;
void *(*memPi_calloc_ptr) (size_t, size_t) = NULL;
void *(*memPi_realloc_ptr) (void *, size_t) = NULL;
void (*memPi_free_ptr) (void *) = NULL;
#endif

/*  Functionality that track memP memory use has not been implemented  */
#ifdef TRACK_MEMP_MEMORY
static h_t *memPi_internal_alloc_ptr_hash = NULL;
#endif

extern unsigned int memPi_alloc_ptr_hashkey (const void *p);
extern int memPi_alloc_ptr_comparator (const void *p1, const void *p2);


void *
memPi_malloc (size_t msz)
{
  void *retval;

#ifdef TRACK_MEMP_MEMORY
  if (memPi_internal_alloc_ptr_hash == NULL)
    init_internal_alloc_hash ();
#endif

  retval = memPi_malloc_ptr (msz);

  return retval;
}

void *
memPi_calloc (size_t numb, size_t msz)
{
  void *retval;

#ifdef TRACK_MEMP_MEMORY
  if (memPi_internal_alloc_ptr_hash == NULL)
    init_internal_alloc_hash ();
#endif

  retval = memPi_calloc_ptr (numb, msz);

  return retval;
}

void *
memPi_realloc (void *ptr, size_t msz)
{
  void *retval;

#ifdef TRACK_MEMP_MEMORY
  if (memPi_internal_alloc_ptr_hash == NULL)
    init_internal_alloc_hash ();
#endif

  retval = memPi_realloc_ptr (ptr, msz);

  return retval;
}

void
memPi_free (void *ptr)
{
#ifdef TRACK_MEMP_MEMORY
  if (memPi_internal_alloc_ptr_hash == NULL)
    init_internal_alloc_hash ();
#endif

  memPi_free_ptr (ptr);
}

/* 

<license>

Copyright (c) 2010, Lawrence Livermore National Security, LLC.
Produced at the Lawrence Livermore National Laboratory
Written by Christopher Chambreau <chcham@llnl.gov>.
CODE-OCEC-10-010
All rights reserved.

This file is part of memP. For details, see http://memp.sourceforge.net .
Please also read this link – Additional BSD Notice.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  o  Redistributions of source code must retain the above copyright notice, 
     this list of conditions and the disclaimer below.

  o  Redistributions in binary form must reproduce the above copyright notice, 
     this list of conditions and the disclaimer (as noted below) in the 
     documentation and/or other materials provided with the distribution.

  o  Neither the name of the LLNS/LLNL nor the names of its contributors may be
     used to endorse or promote products derived from this software without 
     specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL LAWRENCE LIVERMORE NATIONAL
SECURITY, LLC, THE U.S. DEPARTMENT OF ENERGY OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.


Additional BSD Notice

1. This notice is required to be provided under our contract with the U.S.
Department of Energy (DOE). This work was produced at Lawrence Livermore
National Laboratory under Contract No. DE-AC52-07NA27344 with the DOE.

2. Neither the United States Government nor Lawrence Livermore National
Security, LLC nor any of their employees, makes any warranty, express or
implied, or assumes any liability or responsibility for the accuracy,
completeness, or usefulness of any information, apparatus, product, or process
disclosed, or represents that its use would not infringe privately-owned rights.

3. Also, reference herein to any specific commercial products, process, or
services by trade name, trademark, manufacturer or otherwise does not
necessarily constitute or imply its endorsement, recommendation, or favoring by
the United States Government or Lawrence Livermore National Security, LLC.
The views and opinions of authors expressed herein do not necessarily state or
reflect those of the United States Government or Lawrence Livermore National
Security, LLC, and shall not be used for advertising or product endorsement
purposes.

</license>

*/


/* eof */
