/* -*- C -*- 

   memP Heap Profiler ( http://memp.sourceforge.net/ )

   Please see COPYRIGHT AND LICENSE information at the end of this file.

   ----- 

   mem_wrappers.c -- heap allocation function wrappers

 */

#ifndef lint
static char *svnid = "$Id: mem_wrappers.c 369 2006-10-05 19:21:12Z chcham $";
#endif

#include "memPconfig.h"
#include "memPi.h"
#include "symbols.h"
#include <string.h>
#include <limits.h>
#include <float.h>

/******************************************************************************/
/*                    INITIALIZATION AND FAILURE FUNCTIONS                    */
/******************************************************************************/

static int wrappers_initialized = 0;
extern h_t *callsite_pc_cache;

void
memPi_set_nonmpi_values ()
{
  int skip_frames_default = -1;

  if (time (&memPi.start_timeofday) == (time_t) - 1)
  {
    memPi_msg_warn ("Could not get time of day from time()\n");
  }

  memPi.toolname = "memP";
  memPi.tag = 9821;
  memPi.procID = getpid ();
  memPi.stdout_ = stdout;
  memPi.stderr_ = stderr;
  memPi.lookup = memPi_lookup;

  memPi.enabled = 1;
  memPi.enabledCount = 1;
  memPi.cumulativeTime = 0.0;
  memPi.global_task_info = NULL;

  /* set some defaults values */
  memPi.collectorRank = 0;
  memPi.tableSize = 256;
  memPi.stackDepth = MEMP_CALLSITE_STACK_DEPTH_MAX;
  memPi.baseNames = 0;
  memPi.reportFormat = MEMP_REPORT_SCI_FORMAT;
  memPi.do_lookup = 1;
  memPi.report_style = memPi_style_text_concise;
  memPi.print_callsite_detail = 0;
/*#ifdef COLLECTIVE_REPORT_DEFAULT
  memPi.collective_report = 1;
#else*/
  memPi.collective_report = 0;
/*#endif*/
  memPi.app_alloc.alloc_cnt = 0;
  memPi.app_alloc.alloc_tot = 0;
  memPi.app_alloc.alloc_max = 0;
  memPi.app_alloc.alloc_cur = 0;
  memPi.app_alloc.free_cnt = 0;
  memPi.app_alloc.free_tot = 0;
  memPi.taskLWM = DBL_MAX;
  memPi.taskHWMreportThreshold = (double) MiB;
  memPi.taskHWMreportMaxCount = 0;
  memPi.abortAtThreshold = 0;
  memPi.skipFrames = skip_frames_default;
  memPi.traceAllStacks = 0;     /* set to 1 for task report */
  memPi.inWrappedFunc = 0;
  memPi.taskHWMrank = -1;
  memPi.reportHWMprintCount = 1;


#if 0                           /* Reorganizing this so that all pointers get tracked,
                                   but recursive functions (i.e. malloc->backtrace->malloc) are kept
                                   from recursing.  Statistics and pointer records are kept for all 
                                   allocations, but only some get stacktraces.  */
  /*  Get heap use at MPI_Init from mallinfo  */
#if defined(Linux)
  {
    struct mallinfo mi;
    int heap_in_use;

    mi = mallinfo ();

    /* Free space at init is block in use - free blocks + mmap'd blocks */
    heap_in_use = (mi.arena - mi.fordblks) + mi.hblkhd;

    memPi_msg_debug ("Heap use at init is %d bytes\n", heap_in_use);
    memPi.heapUseAtInit = heap_in_use;
    memPi.unknownHeapUse = heap_in_use;
  }
#else
  memPi.heapUseAtInit = 0;
  memPi.unknownHeapUse = 0;
#endif
#endif

  memPi_getenv ();

  /* if skipFrames is not (apparently) set at runtime, set it to step out of 
     the wrapper functions.
   */
  if (memPi.skipFrames == skip_frames_default)
    memPi.skipFrames = 2;

  memPi.task_callsite_stats =
    h_open (memPi.tableSize, memPi_callsite_stats_pc_hashkey,
            memPi_callsite_stats_pc_comparator);

  if (callsite_pc_cache == NULL)
  {
    callsite_pc_cache = h_open (memPi.tableSize,
                                callsite_pc_cache_hashkey,
                                callsite_pc_cache_comparator);
  }
  if (callsite_src_id_cache == NULL)
  {
    callsite_src_id_cache = h_open (memPi.tableSize,
                                    callsite_src_id_cache_hashkey,
                                    callsite_src_id_cache_comparator);
  }

  if (memPi.alloc_ptr_hash == NULL)
    memPi_open_alloc_hash ();

  memPi.set_nonmpi_values = 1;
  memPi_msg_debug("At exit of memPi_set_nonmpi_values, enabled=%d\n", memPi.enabled);
  return;
}


void
memPi_init_wrappers ()
{
  char curr_stack;

  MEMP_TRACE ("Enter memPi_init_wrappers");

  /* memPi_msg_debug ("In %s, afterInit == %d, in_inter == %d\n", __func__,
     memPi.afterInit, in_inter);
   */

  /* Set at the beginning of function to avoid possible recursion, with
     calloc for example
   */
  wrappers_initialized = 1;
  memPi.in_init_wrappers = 1;

  memPi.topOfStack = &curr_stack;

  MEMP_TRACE ("Before memPi_open_alloc_hash");
  memPi_open_alloc_hash ();     // Currently uses pre-defined internal symbols.
  MEMP_TRACE ("After memPi_open_alloc_hash");

  /*  TODO: add return check  */
  MEMP_TRACE ("Before memPi_set_internal_functions");
  memPi_set_internal_functions ();
  MEMP_TRACE ("After memPi_set_internal_functions");

#if 0
/* TODO: determine whether is information is useful at all.  */
#if defined(Linux)
  {
    struct mallinfo mi;
    size_t heap_in_use;

    mi = mallinfo ();

    /* Free space at init is block in use - free blocks + mmap'd blocks */
    /* TODO:  Is this really accurate at all? */
    heap_in_use = (mi.arena - mi.fordblks) + mi.hblkhd;
    print_mallinfo ();

    memPi_msg_debug ("Heap use at init is %u bytes\n", heap_in_use);
    memPi.heapUseAtInit = heap_in_use;
    memPi.unknownHeapUse = heap_in_use;
  }
#else
  memPi.heapUseAtInit = 0;
  memPi.unknownHeapUse = 0;
#endif
#endif

  memPi_set_nonmpi_values ();
  memPi.in_init_wrappers = 0;

  MEMP_TRACE ("Exit memPi_init_wrappers");
}

#ifdef HAVE_BLUE_GENE_KERNEL_INTERFACE
#include <stdio.h>
#include <spi/kernel_interface.h>
#endif


void
memPi_allocation_failed_msg (const char *func, size_t msz)
{
  void *pc_array[MEMP_CALLSITE_STACK_DEPTH_MAX] = { NULL };
  size_t avail;
  size_t currheaptot;


#ifdef DO_GET_AVAIL
  avail = memPi_get_heap_avail (0);
#ifdef HAVE_BLUE_GENE_KERNEL_INTERFACE
  Kernel_GetMemorySize (KERNEL_MEMSIZE_HEAP, &currheaptot);
  memPi_msg_warn
    ("Failed to allocate %lu bytes with %s, %lu available, %lu in use.\n",
     msz, func, avail, currheaptot);
#else
  memPi_msg_warn ("Failed to allocate %lu bytes with %s, %lu available.\n",
                  msz, func, avail);
#endif

#else
  memPi_msg_warn ("Failed to allocate %lu bytes with %s\n", msz, func);
#endif

  memPi_get_stack_trace (pc_array);
  memPi_print_stack (pc_array);
}



/******************************************************************************/
/*                     ALLOCATION FUNCTIONS TO INTERPOSE                      */
/******************************************************************************/

/* ----- MALLOC ------------------------------------------------ */
void *
malloc (size_t msz)
{
  unsigned int func_id = memPi_ID_malloc;
  static void *pc_array[MEMP_CALLSITE_STACK_DEPTH_MAX] = { NULL };
  callsite_stats_t *csp = NULL;
  void *retval = NULL;

  MEMP_TRACE ("Enter malloc");
  /* memPi_msg_debug ("In %s, afterInit == %d, in_inter == %d\n", __func__,
     memPi.afterInit, in_inter); */

  /* Make sure we have function pointers to the internal functions */
  if (wrappers_initialized == 0)
    memPi_init_wrappers ();

  /* make call for allocation */
  MEMP_TRACE ("Before internal malloc call");
  retval = memPi.malloc_ptr (msz);
  MEMP_TRACE ("After internal malloc call");

  if (memPi.in_finalize == 1 || memPi.enabled == 0)
    return retval;

  memPi.inWrappedFunc++;

  /* Warn if NULL result (we are assuming allocation failed) */
  //if (retval == NULL && memPi_debug > 0)
  if (retval == NULL)
  {
    // memPi_msg_warn ("%s failed for %d bytes!\n", __func__, msz);
    memPi_allocation_failed_msg (__func__, msz);
    return NULL;
  }

  //  inWrappedFunc flag ensures that recursive calls do not record callsite info.
  //  afterInit flag is used to avoid recording malloc calls before MPI_Init.
  //  MPI_Init checks MEMP args and sets up housekeeping to track allocations.
  //  The enabled flag is used to switch profiling on and off through "-o" or Pcontrol.

  if (memPi.inWrappedFunc == 1)
  {
#ifdef TESTING
    memPi_msg ("Wrapped %s for allocation of %u\n", __func__, msz);
#endif

    /*  Generate stack trace callsite info   */
    if (memPi.traceAllStacks == 1)
      memPi_get_stack_trace (pc_array);

    /*  Print the stack trace for the failed allocation, if debugging  */
    if (retval == NULL && memPi_debug > 0 && memPi.traceAllStacks == 1)
      memPi_print_stack (pc_array);

    //  Update the interposed callsite info for reporting
    if (retval != NULL)
    {
      /* If tracking call sites, update the call site stats */
      if (memPi.traceAllStacks == 1)
      {
        csp = memPi_update_callsite_stats (func_id, memPi.rank, pc_array, 0,
                                           (double) msz, (double) 0);

        /* If debugging, print stack. */
        if (memPi_debug >= MEMP_DEBUG_STACK)
          memPi_print_stack (pc_array);
      }
    }

    /* Record information associated with the pointer to the new allocation. */
    MEMP_TRACE ("Before malloc memPi_insert_alloc_ptr call");
    memPi_insert_alloc_ptr (memPi.alloc_ptr_hash, retval, msz,
                            &memPi.app_alloc, csp);
    MEMP_TRACE ("After malloc memPi_insert_alloc_ptr call");

  /* Check whether new allocation results in a new task HWM */
  MEMP_TRACE ("Before malloc memPi_update_task_stats call");
  memPi_update_task_stats (msz, 0, csp, func_id);
  MEMP_TRACE ("After malloc memPi_update_task_stats call");

  }

  memPi.inWrappedFunc--;

  MEMP_TRACE ("Exit malloc");

  return retval;
}

/* ----- CALLOC ------------------------------------------------ */

void *
calloc (size_t nmemb, size_t msz)
{
  unsigned int func_id = memPi_ID_calloc;
  void *pc_array[MEMP_CALLSITE_STACK_DEPTH_MAX] = { NULL };
  void *retval = NULL;
  callsite_stats_t *csp = NULL;

  MEMP_TRACE ("Enter calloc");

  /* memPi_msg_debug ("In %s, afterInit == %d, in_inter == %d\n", __func__,
     memPi.afterInit, in_inter);  */

  /* Make sure we have function pointers to the internal functions */
  if (wrappers_initialized == 0)
    memPi_init_wrappers ();

  if (memPi.in_init_wrappers)
  {
    /*  If we get here, it is through a call
     *  in dlsym. 
     */
    retval = memPi.calloc_ptr (nmemb, msz);

    return (retval);
  }

  /* make call for allocation */
  retval = memPi.calloc_ptr (nmemb, msz);

  if (memPi.in_finalize == 1)
    return retval;

  memPi.inWrappedFunc++;

  /* Warn if NULL result (we are assuming allocation failed) */
  if (retval == NULL)
  {
    memPi_allocation_failed_msg (__func__, (nmemb * msz));
  }

  //  inWrappedFunc flag ensures that recursive calls do not record callsite info.
  //  afterInit flag is used to avoid recording malloc calls before MPI_Init.
  //  MPI_Init checks MEMP args and sets up housekeeping to track allocations.
  //  The enabled flag is used to switch profiling on and off through "-o" or Pcontrol.

  if (memPi.inWrappedFunc == 1)
  {
#ifdef TESTING
    memPi_msg ("Wrapped %s for allocation of %u\n", __func__, nmemb * msz);
#endif
    /*  Generate stack trace callsite info   */
    if (memPi.traceAllStacks == 1)
      memPi_get_stack_trace (pc_array);

    /*  Print the stack trace for the failed allocation, if debugging  */
    if (retval == NULL && memPi_debug > 0 && memPi.traceAllStacks == 1)
      memPi_print_stack (pc_array);

    //  Update the interposed callsite info for reporting
    if (retval != NULL)
    {
      /* If tracking call sites, update the call site stats */
      if (memPi.traceAllStacks == 1)
      {
        csp = memPi_update_callsite_stats (func_id, memPi.rank, pc_array, 0,
                                           (double) (nmemb * msz),
                                           (double) 0);

        /* If debugging, print stack. */
        if (memPi_debug >= MEMP_DEBUG_STACK)
          memPi_print_stack (pc_array);
      }
    }

    /* Record information associated with the pointer to the new allocation. */
    memPi_insert_alloc_ptr (memPi.alloc_ptr_hash, retval, nmemb * msz,
                            &memPi.app_alloc, csp);

  /* Check whether new allocation results in a new task HWM */
  memPi_update_task_stats (nmemb * msz, 0, csp, func_id);
  }

  memPi.inWrappedFunc--;

  MEMP_TRACE ("Exit calloc");
  return retval;
}

/* ----- REALLOC ------------------------------------------------ */

void *
realloc (void *inptr, size_t msz)
{
  unsigned int func_id = memPi_ID_realloc;
  void *pc_array[MEMP_CALLSITE_STACK_DEPTH_MAX] = { NULL };
  void *retval = NULL;
  void *inptr_save = inptr;
  callsite_stats_t *csp = NULL;
  size_t prev_size = 0;

  MEMP_TRACE ("Enter realloc");
#if 0
  memPi_msg_debug ("In %s, afterInit == %d, in_inter == %d\n", __func__,
                   memPi.afterInit, in_inter);
#endif

  /* Make sure we have function pointers to the internal functions */
  if (wrappers_initialized == 0)
    memPi_init_wrappers ();

  /* make call for allocation */
  retval = memPi.realloc_ptr (inptr, msz);

  if (memPi.in_finalize == 1)
    return retval;

  memPi.inWrappedFunc++;

  /* Warn if NULL result (we are assuming allocation failed) */
  if (retval == NULL)
  {
    memPi_allocation_failed_msg (__func__, msz);
  }

  //  inWrappedFunc flag ensures that recursive calls do not record callsite info.
  //  afterInit flag is used to avoid recording malloc calls before MPI_Init.
  //  MPI_Init checks MEMP args and sets up housekeeping to track allocations.
  //  The enabled flag is used to switch profiling on and off through "-o" or Pcontrol.

  if (memPi.inWrappedFunc == 1)
  {
#ifdef TESTING
    memPi_msg ("Wrapped %s for allocation of %u\n", __func__, msz);
#endif
    /*  Generate stack trace callsite info   */
    if (memPi.traceAllStacks == 1)
      memPi_get_stack_trace (pc_array);

    /*  Print the stack trace for the failed allocation, if debugging  */
    if (retval == NULL && memPi_debug > 0 && memPi.traceAllStacks == 1)
      memPi_print_stack (pc_array);

    //  Update the interposed callsite info for reporting

    /*  Size of re-allocation is 0 : Behavior similar to free   */
    if (msz == 0)
    {
      msz =
        memPi_delete_alloc_ptr (memPi.alloc_ptr_hash, inptr_save,
                                &memPi.app_alloc);
      memPi_msg_debug ("before realloc-free update stats, msz is %ld\n", msz);

      if (memPi.traceAllStacks == 1)
        memPi_update_callsite_stats (func_id, memPi.rank, pc_array, 0,
                                     (double) 0, (double) msz);
    }

    /*  Input pointer is NULL : Behavior similar to malloc   */
    else if (retval != NULL && inptr_save == NULL)
    {
      if (memPi.traceAllStacks == 1)
        csp = memPi_update_callsite_stats (func_id, memPi.rank, pc_array, 0,
                                           (double) msz, (double) 0);
    }

    /*  Realloc memory to larger or smaller size   */
    else if (retval != NULL)
    {
      prev_size = memPi_delete_alloc_ptr (memPi.alloc_ptr_hash, inptr_save,
                                          &memPi.app_alloc);

      if (memPi.traceAllStacks == 1)
        csp = memPi_update_callsite_stats (func_id, memPi.rank, pc_array, 0,
                                           (double) msz, (double) prev_size);
    }

  if (msz == 0)
    memPi_update_task_stats (0, msz, csp, func_id);     //  Similar to free
  else if (retval != NULL && inptr_save == NULL)
  {
    memPi_insert_alloc_ptr (memPi.alloc_ptr_hash, retval, msz,
                            &memPi.app_alloc, csp);
    memPi_update_task_stats (msz, 0, csp, func_id);     // Similar to malloc
  }
  else if (retval != NULL && inptr_save == NULL)
  {
    memPi_insert_alloc_ptr (memPi.alloc_ptr_hash, retval, msz,
                            &memPi.app_alloc, csp);
    memPi_update_task_stats (msz, prev_size, csp, func_id);     // Change size
  }
  }

  memPi.inWrappedFunc--;

  MEMP_TRACE ("Exit realloc");
  return retval;
}

/* ----- POSIX_MEMALIGN ---------------------------------------- */

#ifdef WRAP_POSIX_MEMALIGN
int
posix_memalign (void **pinptr, size_t align, size_t msz)
{
  unsigned int func_id = memPi_ID_posix_memalign;
  void *pc_array[MEMP_CALLSITE_STACK_DEPTH_MAX] = { NULL };
  int retval = 0;
  callsite_stats_t *csp = NULL;

  MEMP_TRACE ("Enter posix_memalign");
  /* memPi_msg_debug ("In %s, afterInit == %d, in_inter == %d\n", __func__,
     memPi.afterInit, in_inter); 
   */

  if (memPi.in_init_wrappers)
  {
    /*  If we get here, there is an internal posix_memalign call
     *  in dlsym.  We should have already found the malloc internal
     *  function, so we'll just call malloc to get heap space during
     *  initialization.
     */
    *pinptr = malloc (msz);

    return (*pinptr == NULL);
  }

  /* Make sure we have function pointers to the internal functions */
  if (wrappers_initialized == 0)
    memPi_init_wrappers ();

  /* make call for allocation */
  retval = memPi.posix_memalign_ptr (pinptr, align, msz);

  if (memPi.in_finalize == 1)
    return retval;

  memPi.inWrappedFunc++;

  /* Warn if failed result */
  if (retval != 0)
  {
    memPi_allocation_failed_msg (__func__, msz);
  }

  //  inWrappedFunc flag ensures that recursive calls do not record callsite info.
  //  afterInit flag is used to avoid recording malloc calls before MPI_Init.
  //  MPI_Init checks MEMP args and sets up housekeeping to track allocations.
  //  The enabled flag is used to switch profiling on and off through "-o" or Pcontrol.
  if (memPi.inWrappedFunc == 1)
  {
#ifdef TESTING
    memPi_msg ("Wrapped %s for allocation of %u\n", __func__, msz);
#endif
    /*  Generate stack trace callsite info   */
    if (memPi.traceAllStacks == 1)
      memPi_get_stack_trace (pc_array);

    /*  Print the stack trace for the failed allocation, if debugging  */
    if (memPi.traceAllStacks == 1 && memPi_debug > 0 && retval != 0
        && memPi.afterInit == 1)
      memPi_print_stack (pc_array);

    //  Update the interposed callsite info for reporting
    if (retval == 0)
    {
      /* If tracking call sites, update the call site stats */
      if (memPi.traceAllStacks == 1)
      {
        csp = memPi_update_callsite_stats (func_id, memPi.rank, pc_array, 0,
                                           (double) msz, (double) 0);

        /* If debugging, print stack. */
        if (memPi_debug >= MEMP_DEBUG_STACK && memPi.afterInit == 1)
          memPi_print_stack (pc_array);
      }

    }
    /* Record information associated with the pointer to the new allocation. */
    memPi_insert_alloc_ptr (memPi.alloc_ptr_hash, *pinptr, msz,
                            &memPi.app_alloc, csp);

  /* Check whether new allocation results in a new task HWM */
  memPi_update_task_stats (msz, 0, csp, func_id);
  }

  memPi.inWrappedFunc--;

  MEMP_TRACE ("Exit posix_memalign");

  return retval;
}
#endif


/* ----- MEMALIGN ---------------------------------------------- */

void *
memalign (size_t boundary, size_t msz)
{
  unsigned int func_id = memPi_ID_memalign;
  void *pc_array[MEMP_CALLSITE_STACK_DEPTH_MAX] = { NULL };
  void *retval;
  callsite_stats_t *csp = NULL;

  MEMP_TRACE ("Enter memalign");

  /*
     memPi_msg_debug ("In %s, afterInit == %d, in_inter == %d\n", __func__,
     memPi.afterInit, in_inter);
   */

  /* Make sure we have function pointers to the internal functions */
  if (wrappers_initialized == 0)
    memPi_init_wrappers ();

  /* make call for allocation */
  retval = memPi.memalign_ptr (boundary, msz);

  if (memPi.in_finalize == 1)
    return retval;

  memPi.inWrappedFunc++;

  /* Warn if NULL result (we are assuming allocation failed) */
  if (retval == NULL)
  {
    memPi_allocation_failed_msg (__func__, msz);
  }

  //  inWrappedFunc flag ensures that recursive calls do not record callsite info.
  //  afterInit flag is used to avoid recording malloc calls before MPI_Init.
  //  MPI_Init checks MEMP args and sets up housekeeping to track allocations.
  //  The enabled flag is used to switch profiling on and off through "-o" or Pcontrol.

  if (memPi.inWrappedFunc == 1)
  {
#ifdef TESTING
    memPi_msg ("Wrapped %s for allocation of %u\n", __func__, msz);
#endif
    /*  Generate stack trace callsite info   */
    if (memPi.traceAllStacks == 1)
      memPi_get_stack_trace (pc_array);

    /*  Print the stack trace for the failed allocation, if debugging  */
    if (retval == NULL && memPi_debug > 0 && memPi.traceAllStacks == 1)
      memPi_print_stack (pc_array);

    //  Update the interposed callsite info for reporting
    if (retval != NULL)
    {
      /* If tracking call sites, update the call site stats */
      if (memPi.traceAllStacks == 1)
      {
        csp = memPi_update_callsite_stats (func_id, memPi.rank, pc_array, 0,
                                           (double) msz, (double) 0);

        /* If debugging, print stack. */
        if (memPi_debug >= MEMP_DEBUG_STACK)
          memPi_print_stack (pc_array);
      }
    }

    /* Record information associated with the pointer to the new allocation. */
    memPi_insert_alloc_ptr (memPi.alloc_ptr_hash, retval, msz,
                            &memPi.app_alloc, csp);

  /* Check whether new allocation results in a new task HWM */
  memPi_update_task_stats (msz, 0, csp, func_id);
  }

  memPi.inWrappedFunc--;

  MEMP_TRACE ("Exit memalign");

  return retval;
}

/* ----- VALLOC ------------------------------------------------ */

void *
valloc (size_t size)
{
  memPi_abort ("valloc not current supported.\n");

  return NULL;
}



/* ----- FREE -------------------------------------------------- */

void
free (void *inptr)
{
  unsigned int func_id = memPi_ID_free;
  size_t msz;
  void *inptr_save = inptr;

  MEMP_TRACE ("Enter free");

  /* memPi_msg_debug ("In %s, afterInit == %d, in_inter == %d\n", __func__,
     memPi.afterInit, in_inter); 
   */

  /* Make sure we have function pointers to the internal functions */
  if (wrappers_initialized == 0)
    memPi_init_wrappers ();

#ifdef TESTING
  memPi_msg ("Wrapped %s\n", __func__);
#endif

  //  inWrappedFunc flag ensures that recursive calls do not record callsite info.
  //  afterInit flag is used to avoid recording malloc calls before MPI_Init.
  //  MPI_Init checks MEMP args and sets up housekeeping to track allocations.
  //  The enabled flag is used to switch profiling on and off through "-o" or Pcontrol.

  if (memPi.in_finalize == 0
      && (msz =
          memPi_delete_alloc_ptr (memPi.alloc_ptr_hash, inptr_save,
                                  &memPi.app_alloc)) > 0)
  {
    memPi_update_task_stats (0, msz, NULL, func_id);
  }

  memPi.free_ptr (inptr);

  MEMP_TRACE ("Exit free");
  return;
}


/* 

<license>

Copyright (c) 2010, Lawrence Livermore National Security, LLC.
Produced at the Lawrence Livermore National Laboratory
Written by Christopher Chambreau <chcham@llnl.gov>.
CODE-OCEC-10-010
All rights reserved.

This file is part of memP. For details, see http://memp.sourceforge.net .
Please also read this link – Additional BSD Notice.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  o  Redistributions of source code must retain the above copyright notice, 
     this list of conditions and the disclaimer below.

  o  Redistributions in binary form must reproduce the above copyright notice, 
     this list of conditions and the disclaimer (as noted below) in the 
     documentation and/or other materials provided with the distribution.

  o  Neither the name of the LLNS/LLNL nor the names of its contributors may be
     used to endorse or promote products derived from this software without 
     specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL LAWRENCE LIVERMORE NATIONAL
SECURITY, LLC, THE U.S. DEPARTMENT OF ENERGY OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.


Additional BSD Notice

1. This notice is required to be provided under our contract with the U.S.
Department of Energy (DOE). This work was produced at Lawrence Livermore
National Laboratory under Contract No. DE-AC52-07NA27344 with the DOE.

2. Neither the United States Government nor Lawrence Livermore National
Security, LLC nor any of their employees, makes any warranty, express or
implied, or assumes any liability or responsibility for the accuracy,
completeness, or usefulness of any information, apparatus, product, or process
disclosed, or represents that its use would not infringe privately-owned rights.

3. Also, reference herein to any specific commercial products, process, or
services by trade name, trademark, manufacturer or otherwise does not
necessarily constitute or imply its endorsement, recommendation, or favoring by
the United States Government or Lawrence Livermore National Security, LLC.
The views and opinions of authors expressed herein do not necessarily state or
reflect those of the United States Government or Lawrence Livermore National
Security, LLC, and shall not be used for advertising or product endorsement
purposes.

</license>

*/


/* eof */
