/* -*- Mode: C; -*- 

   memP Heap Profiler ( http://memp.sourceforge.net/ )

   Please see COPYRIGHT AND LICENSE information at the end of this file.

   ----- 

   memp_timers.h -- timer macros

   $Id: memp_timers.h 405 2007-04-26 16:15:44Z chcham $

*/

#ifndef _MPITI_TIMERS_H
#define _MPITI_TIMERS_H

#define MSECS 1000
#define USECS 1000000
#define NSECS 1000000000

#include <unistd.h>
#include <sys/time.h>
typedef double memP_TIMER;

#if (defined(SunOS) && ! defined(USE_GETTIMEOFDAY))
#include "memp_timers/sunos_local.h"

#elif (defined(AIX) && ! defined(USE_GETTIMEOFDAY))
#include "memp_timers/aix_local.h"

#elif (defined(UNICOS_mp) && ! defined(USE_GETTIMEOFDAY))
#include "memp_timers/crayx1_hw.h"

#elif (defined(Catamount) && defined(USE_DCLOCK))
#include "memp_timers/crayxt.h"

#elif defined(USE_RTS_GET_TIMEBASE)
#include "memp_timers/bgl_local.h"

#elif (defined(Linux) && defined(USE_CLOCK_GETTIME) && defined(_POSIX_TIMERS) && (_POSIX_TIMERS > 0))
#include "memp_timers/linux_posix.h"

#elif defined(USE_WTIME)
#define memPi_TIMER double
#define memPi_TIME double
#define memPi_TIMER_NAME "PMPI_Wtime"
#define memPi_GETTIME(timeaddr) (*(timeaddr) = (PMPI_Wtime()*USECS))
#define memPi_GETUSECS(timeaddr) (*(timeaddr))
#define memPi_GETTIMEDIFF(end,start) ((*end)-(*start))
#else

/* gettimeofday returns microseconds */
#define memPi_TIMER double
#define memPi_TIME struct timeval
#define memPi_TIMER_NAME "gettimeofday"
#define memPi_ASNTIME(lhs,rhs) {bcopy(rhs, lhs, sizeof(memPi_TIMER));}
#define memPi_GETTIME(timeaddr) gettimeofday(timeaddr,NULL)
#define memPi_GETUSECS(timeaddr) (((memPi_TIMER)(timeaddr)->tv_sec)*USECS+((memPi_TIMER)(timeaddr)->tv_usec))
#define memPi_PRINTTIME(taddr) printf("Time is %ld sec and %ld usec.\n", (taddr)->tv_sec, (taddr)->tv_usec)
#define memPi_GETTIMEDIFF(end,start) ((memP_TIMER)((((memPi_TIMER)(end)->tv_sec)*USECS)+(end)->tv_usec)-((((memPi_TIMER)(start)->tv_sec)*USECS)+(start)->tv_usec))
#define memPi_PRINTTIMEDIFF(end,start) {printf("Time diff is %ld usecs.\n",memPi_GETTIMEDIFF(end,start));}

#endif

#endif


/* 

<license>

Copyright (c) 2010, Lawrence Livermore National Security, LLC.
Produced at the Lawrence Livermore National Laboratory
Written by Christopher Chambreau <chcham@llnl.gov>.
CODE-OCEC-10-010
All rights reserved.

This file is part of memP. For details, see http://memp.sourceforge.net .
Please also read this link – Additional BSD Notice.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  o  Redistributions of source code must retain the above copyright notice, 
     this list of conditions and the disclaimer below.

  o  Redistributions in binary form must reproduce the above copyright notice, 
     this list of conditions and the disclaimer (as noted below) in the 
     documentation and/or other materials provided with the distribution.

  o  Neither the name of the LLNS/LLNL nor the names of its contributors may be
     used to endorse or promote products derived from this software without 
     specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL LAWRENCE LIVERMORE NATIONAL
SECURITY, LLC, THE U.S. DEPARTMENT OF ENERGY OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.


Additional BSD Notice

1. This notice is required to be provided under our contract with the U.S.
Department of Energy (DOE). This work was produced at Lawrence Livermore
National Laboratory under Contract No. DE-AC52-07NA27344 with the DOE.

2. Neither the United States Government nor Lawrence Livermore National
Security, LLC nor any of their employees, makes any warranty, express or
implied, or assumes any liability or responsibility for the accuracy,
completeness, or usefulness of any information, apparatus, product, or process
disclosed, or represents that its use would not infringe privately-owned rights.

3. Also, reference herein to any specific commercial products, process, or
services by trade name, trademark, manufacturer or otherwise does not
necessarily constitute or imply its endorsement, recommendation, or favoring by
the United States Government or Lawrence Livermore National Security, LLC.
The views and opinions of authors expressed herein do not necessarily state or
reflect those of the United States Government or Lawrence Livermore National
Security, LLC, and shall not be used for advertising or product endorsement
purposes.

</license>

*/

/* EOF */
