/* -*- C -*- 

   memP Heap Profiler ( http://memp.sourceforge.net/ )

   Please see COPYRIGHT AND LICENSE information at the end of this file.

   ----- 

   mpi_wrappers.c -- MPI function wrappers

 */

#ifndef lint
static char *svnid = "$Id: mpi_wrappers.c 369 2006-10-05 19:21:12Z chcham $";
#endif

#include "memPconfig.h"
#include "memPi.h"
#include "symbols.h"
#include <string.h>
#include <float.h>

extern h_t *callsite_pc_cache;

/******************************************************************************/
/*                         MPI FUNCTIONS TO INTERPOSE                         */
/******************************************************************************/


/* ----- INIT -------------------------------------------------- */

static int
_MPI_Init (int *argc, char ***argv)
{
  int rc = 0;
  int enabledStatus;

  enabledStatus = memPi.enabled;
  memPi.enabled = 0;

  rc = PMPI_Init (argc, argv);

  memPi.enabled = enabledStatus;
  /* This hard-coded enabling needs to be re-examined at some point. */
  memPi.enabled = 1;
  memPi_msg_debug("At MPI_Init, enabled=%d\n", memPi.enabled);

#if defined(Linux) && ! defined(ppc64)
  memPi.appFullName = getProcExeLink ();
  //memPi_msg_debug ("appFullName is %s\n", memPi.appFullName);
  memPi_init (GetBaseAppName (memPi.appFullName));
#else
  if (argv != NULL && *argv != NULL && **argv != NULL)
  {
    memPi_init (GetBaseAppName (**argv));
    memPi.appFullName = strdup (**argv);
  }
  else
  {
    memPi_init ("Unknown");
    memPi_msg_debug ("argv is NULL\n");
  }
#endif

  return rc;
}

extern int
MPI_Init (int *argc, char ***argv)
{
  int rc = 0;

  memPi.toolname = "memP";

  rc = _MPI_Init (argc, argv);

  if (argc != NULL && argv != NULL)
    memPi_copy_given_args (&(memPi.ac), memPi.av, 32, *argc, *argv);
  else
  {
#ifdef Linux
    getProcCmdLine (&(memPi.ac), memPi.av);
#else
    memPi.ac = 0;
#endif
  }

  memPi.afterInit = 1;

  return rc;
}

extern void
F77_MPI_INIT (int *ierr)
{
  int rc = 0;
  char **tmp_argv;

  memPi.toolname = "memP";
#ifdef Linux
  getProcCmdLine (&(memPi.ac), memPi.av);
#else
  memPi_copy_args (&(memPi.ac), memPi.av, 32);
#endif

  tmp_argv = memPi.av;
  rc = _MPI_Init (&(memPi.ac), (char ***) &tmp_argv);
  *ierr = rc;

  memPi.afterInit = 1;

  return;
}


/* ----- FINALIZE -------------------------------------------------- */

static int
_MPI_Finalize ()
{
  int rc = 0;

  memPi_finalize ();
  memPi.enabled = 0;
  memPi_msg_debug ("calling PMPI_Finalize\n");
  rc = PMPI_Finalize ();
  memPi_msg_debug ("returning from PMPI_Finalize\n");

  return rc;
}

extern int
MPI_Finalize (void)
{
  int rc = 0;

  rc = _MPI_Finalize ();

  return rc;
}

extern void
F77_MPI_FINALIZE (int *ierr)
{
  int rc = 0;

  rc = _MPI_Finalize ();
  *ierr = rc;

  return;
}


/******************************************************************************/
/*                      INTERNAL INIT/FINALIZE FUNCTIONS                      */
/******************************************************************************/

void
memPi_set_mpi_values (char *appName)
{
  memPi.appName = strdup (appName);
  memPi.comm = MPI_COMM_WORLD;
  PMPI_Comm_rank (memPi.comm, &memPi.rank);
  PMPI_Comm_size (memPi.comm, &memPi.size);
  PMPI_Get_processor_name (memPi.hostname, &memPi.hostnamelen);
}


#if 0
// moved to mem_wrappers.c
void
memPi_set_nonmpi_values ()
{
  int skip_frames_default = -1;

  if (time (&memPi.start_timeofday) == (time_t) - 1)
  {
    memPi_msg_warn ("Could not get time of day from time()\n");
  }

  memPi.toolname = "memP";
  memPi.tag = 9821;
  memPi.procID = getpid ();
  memPi.stdout_ = stdout;
  memPi.stderr_ = stderr;
  memPi.lookup = memPi_lookup;

  memPi.enabled = 1;
  memPi.enabledCount = 1;
  memPi.cumulativeTime = 0.0;
  memPi.global_task_info = NULL;

  /* set some defaults values */
  memPi.collectorRank = 0;
  memPi.tableSize = 256;
  memPi.stackDepth = MEMP_CALLSITE_STACK_DEPTH_MAX;
  memPi.baseNames = 0;
  memPi.reportFormat = MEMP_REPORT_SCI_FORMAT;
  memPi.do_lookup = 1;
  memPi.report_style = memPi_style_text_concise;
  memPi.print_callsite_detail = 0;
/*#ifdef COLLECTIVE_REPORT_DEFAULT
  memPi.collective_report = 1;
#else*/
  memPi.collective_report = 0;
/*#endif*/
  memPi.app_alloc.alloc_cnt = 0;
  memPi.app_alloc.alloc_tot = 0;
  memPi.app_alloc.alloc_max = 0;
  memPi.app_alloc.alloc_cur = 0;
  memPi.app_alloc.free_cnt = 0;
  memPi.app_alloc.free_tot = 0;
  memPi.taskLWM = DBL_MAX;
  memPi.taskHWMreportThreshold = (double) MiB;
  memPi.taskHWMreportMaxCount = 0;
  memPi.abortAtThreshold = 0;
  memPi.skipFrames = skip_frames_default;
  memPi.traceAllStacks = 0;     /* set to 1 for task report */
  memPi.inWrappedFunc = 0;
  memPi.taskHWMrank = -1;
  memPi.reportHWMprintCount = 1;


#if 0                           /* Reorganizing this so that all pointers get tracked,
                                   but recursive functions (i.e. malloc->backtrace->malloc) are kept
                                   from recursing.  Statistics and pointer records are kept for all 
                                   allocations, but only some get stacktraces.  */
  /*  Get heap use at MPI_Init from mallinfo  */
#if defined(Linux)
  {
    struct mallinfo mi;
    int heap_in_use;

    mi = mallinfo ();

    /* Free space at init is block in use - free blocks + mmap'd blocks */
    heap_in_use = (mi.arena - mi.fordblks) + mi.hblkhd;

    memPi_msg_debug ("Heap use at init is %d bytes\n", heap_in_use);
    memPi.heapUseAtInit = heap_in_use;
    memPi.unknownHeapUse = heap_in_use;
  }
#else
  memPi.heapUseAtInit = 0;
  memPi.unknownHeapUse = 0;
#endif
#endif

  memPi_getenv ();

  /* if skipFrames is not (apparently) set at runtime, set it to step out of 
     the wrapper functions.
   */
  if (memPi.skipFrames == skip_frames_default)
    memPi.skipFrames = 2;

  memPi.task_callsite_stats =
    h_open (memPi.tableSize, memPi_callsite_stats_pc_hashkey,
            memPi_callsite_stats_pc_comparator);

  if (callsite_pc_cache == NULL)
  {
    callsite_pc_cache = h_open (memPi.tableSize,
                                callsite_pc_cache_hashkey,
                                callsite_pc_cache_comparator);
  }
  if (callsite_src_id_cache == NULL)
  {
    callsite_src_id_cache = h_open (memPi.tableSize,
                                    callsite_src_id_cache_hashkey,
                                    callsite_src_id_cache_comparator);
  }

  if (memPi.alloc_ptr_hash == NULL)
    memPi_open_alloc_hash ();

  return;
}
#endif


void
memPi_init (char *appName)
{
  memPi_set_mpi_values (appName);

  /* -- welcome msg only collector  */
  if (memPi.collectorRank == memPi.rank)
  {
    memPi_msg ("\n");
    memPi_msg ("%s V%d.%d.%d (Build %s/%s)\n", memPi.toolname, memPi_vmajor,
               memPi_vminor, memPi_vpatch, memPi_vdate, memPi_vtime);
    memPi_msg ("Direct questions and errors to %s\n", MEMP_HELP_LIST);
    memPi_msg ("\n");
    if ( memPi.envStr != NULL )
      memPi_msg ("Found MEMP environment variable : [%s]\n", memPi.envStr);
    memPi_msg ("\n");
  }

  memPi_msg_debug ("appName is %s\n", appName);
  memPi_msg_debug ("successful init on %d, %s\n", memPi.rank, memPi.hostname);

  if (memPi.enabled)
  {
    memPi_GETTIME (&memPi.startTime);
  }

  return;
}

void
memPi_finalize ()
{
  memPi.in_finalize = 1;
  memPi_generateReport (memPi.report_style);

  /* clean up data structures, etc */
  h_close (memPi.task_callsite_stats);

  /*  Could do a lot of housekeeping before calling PMPI_Finalize()
   *  but is it worth the additional work?
   *  For instance:
   h_gather_data (memPi.global_callsite_stats_agg, &ac, (void ***) &av);
   for (i = 0; (i < 20) && (i < ac); i++)
   {
   if (av[i]->siteData != NULL )
   memPi_free (av[i]->siteData);
   }
   */

  return;
}


/* 

<license>

Copyright (c) 2010, Lawrence Livermore National Security, LLC.
Produced at the Lawrence Livermore National Laboratory
Written by Christopher Chambreau <chcham@llnl.gov>.
CODE-OCEC-10-010
All rights reserved.

This file is part of memP. For details, see http://memp.sourceforge.net .
Please also read this link – Additional BSD Notice.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  o  Redistributions of source code must retain the above copyright notice, 
     this list of conditions and the disclaimer below.

  o  Redistributions in binary form must reproduce the above copyright notice, 
     this list of conditions and the disclaimer (as noted below) in the 
     documentation and/or other materials provided with the distribution.

  o  Neither the name of the LLNS/LLNL nor the names of its contributors may be
     used to endorse or promote products derived from this software without 
     specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL LAWRENCE LIVERMORE NATIONAL
SECURITY, LLC, THE U.S. DEPARTMENT OF ENERGY OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.


Additional BSD Notice

1. This notice is required to be provided under our contract with the U.S.
Department of Energy (DOE). This work was produced at Lawrence Livermore
National Laboratory under Contract No. DE-AC52-07NA27344 with the DOE.

2. Neither the United States Government nor Lawrence Livermore National
Security, LLC nor any of their employees, makes any warranty, express or
implied, or assumes any liability or responsibility for the accuracy,
completeness, or usefulness of any information, apparatus, product, or process
disclosed, or represents that its use would not infringe privately-owned rights.

3. Also, reference herein to any specific commercial products, process, or
services by trade name, trademark, manufacturer or otherwise does not
necessarily constitute or imply its endorsement, recommendation, or favoring by
the United States Government or Lawrence Livermore National Security, LLC.
The views and opinions of authors expressed herein do not necessarily state or
reflect those of the United States Government or Lawrence Livermore National
Security, LLC, and shall not be used for advertising or product endorsement
purposes.

</license>

*/


/* eof */
