/* -*- C -*- 

   memP Heap Profiler ( http://memp.sourceforge.net/ )

   Please see COPYRIGHT AND LICENSE information at the end of this file.

   ----- 

   pc_lookup.c -- functions that interface to bfd for source code lookup

 */

#ifndef lint
static char *svnid = "$Id: pc_lookup.c 376 2006-11-06 19:16:21Z chcham $";
#endif

#if defined(DEMANGLE_GNU)
#include <ansidecl.h>
#endif

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "memPi.h"
#include "memPconfig.h"

#ifdef ENABLE_BFD
#ifndef CEXTRACT
#include "bfd.h"
#endif
static asymbol **syms;
static bfd_vma pc;
static const char *filename;
static const char *functionname;
static unsigned int line;
static bfd *abfd = NULL;
extern h_t *callsite_pc_cache;
int callsite_src_id_counter = 1;

/*  BFD boolean and bfd_boolean types have changed through versions.
    It looks like bfd_boolean will be preferred.                     */
#ifdef HAVE_BFD_BOOLEAN
typedef boolean bfd_boolean;
#endif
#ifndef FALSE
#define FALSE 0
#endif
static bfd_boolean found;

/* take a callstats record (the pc) and determine src file, line, if
   possible and assign a callsite id.
 */
int
memPi_query_pc (void *pc, char **filename, char **functname,
                unsigned int *lineno)
{
  int rc = 0;
  callsite_pc_cache_entry_t key;
  callsite_pc_cache_entry_t *csp;
  char addr_buf[24];

  key.pc = pc;
  /* do we have a cache entry for this pc? If so, use entry */
  if (h_search (callsite_pc_cache, &key, (void **) &csp) == NULL)
  {
    /* no cache entry: create, lookup, and insert */
    csp =
      (callsite_pc_cache_entry_t *)
      memPi_malloc (sizeof (callsite_pc_cache_entry_t));
    csp->pc = pc;
#if defined(ENABLE_BFD) || defined(USE_LIBDWARF)
    if (memP_find_src_loc (pc, filename, lineno, functname) == 0)
    {
      if (*filename == NULL || strcmp (*filename, "??") == 0)
        *filename = "[unknown]";

      if (*functname == NULL)
        *functname = "[unknown]";

      memPi_msg_debug
        ("Successful Source lookup for [%s]: %s, %d, %s\n",
         memP_format_address (pc, addr_buf), *filename, *lineno, *functname);

      csp->filename = strdup (*filename);
      csp->functname = strdup (*functname);
      csp->line = *lineno;
    }
    else
    {
      if (pc != NULL)
      {
        memPi_msg_debug ("Unsuccessful Source lookup for [%s]\n",
                         memP_format_address (pc, addr_buf));
      }
      csp->filename = strdup ("[unknown]");
      csp->functname = strdup ("[unknown]");
      csp->line = 0;
    }
#else /* ! ENABLE_BFD || USE_LIBDWARF */
    csp->filename = strdup ("[unknown]");
    csp->functname = strdup ("[unknown]");
    csp->line = 0;
#endif
    if (h_insert (callsite_pc_cache, csp) == 1)
      memPi_msg_warn ("Duplicate insertion found in callsite_pc_cache\n");
  }

  *filename = csp->filename;
  *functname = csp->functname;
  *lineno = csp->line;

  if (*lineno == 0)
    rc = 1;                     /* use this value to indicate a failed lookup */

  return rc;
}


int
memPi_query_src (callsite_stats_t * p)
{
  int i;
  callsite_src_id_cache_entry_t key;
  callsite_src_id_cache_entry_t *csp;
  int before_main = 1;

  assert (p != NULL && p->pc != NULL);

  /* Because multiple pcs can map to the same source line, we must
     check that mapping here. If we got unknown, then we assign
     different ids */
  bzero (&key, sizeof (callsite_src_id_cache_entry_t));


  for (i = 0; (i < MEMP_CALLSITE_STACK_DEPTH) && (p->pc[i] != NULL); i++)
  {
    if (before_main == 0)
      p->pc[i] = NULL;

    if (memPi.do_lookup == 1)
      (void) memPi_query_pc (p->pc[i], &(p->filename[i]), &(p->functname[i]),
                             &(p->lineno[i]));
    else
    {
      p->filename[i] = strdup ("[unknown]");
      p->functname[i] = strdup ("[unknown]");
      p->lineno[i] = 0;
    }

    key.filename[i] = p->filename[i];
    key.functname[i] = p->functname[i];
    key.line[i] = p->lineno[i];
    key.pc[i] = p->pc[i];

    if (p->functname[i] != NULL && (strcmp (p->functname[i], "main") == 0
                                    || strcmp (p->functname[i],
                                               ".main") == 0
                                    || strcmp (p->functname[i],
                                               "MAIN__") == 0))
      before_main = 0;
  }

  /* lookup/generate an ID based on the callstack, not just the callsite pc */
  if (h_search (callsite_src_id_cache, &key, (void **) &csp) == NULL)
  {
    /* create a new entry, and assign an id based on callstack */
    csp =
      (callsite_src_id_cache_entry_t *)
      memPi_malloc (sizeof (callsite_src_id_cache_entry_t));

    if (csp == NULL)
      memPi_abort ("Failed to allocate callsite entry with HWM %.0f\n",
                   memPi.taskHWM);

    bzero (csp, sizeof (callsite_src_id_cache_entry_t));

    for (i = 0; (i < MEMP_CALLSITE_STACK_DEPTH) && (p->pc[i] != NULL); i++)
    {
      csp->filename[i] = strdup (key.filename[i]);
      csp->functname[i] = strdup (key.functname[i]);
      csp->line[i] = key.line[i];
      csp->pc[i] = p->pc[i];

#if 0
      /*  Do not bother printing stack frames above main   */
      /* Adding NULL PCs above main should keep from generating 
       * multiple call sites for what appears to be the same 
       * call stack.
       */
      if (csp->functname[i] && (strcmp (csp->functname[i], "main") == 0
                                || strcmp (csp->functname[i],
                                           ".main") == 0
                                || strcmp (csp->functname[i], "MAIN__") == 0))
        do_cs_copy = 0;
#endif
    }
    csp->id = callsite_src_id_counter++;
    csp->op = p->op;
    if (h_insert (callsite_src_id_cache, csp) == 1)
      memPi_msg_warn ("Duplicate insertion found in callsite_src_id_cache\n");
  }

  /* assign ID to this record */
  p->csid = csp->id;
  return p->csid;
}

#ifdef DEMANGLE_IBM

#include <demangle.h>
#include <string.h>

char *
memPdemangle (const char *mangledSym)
{
  Name *ret;
  char *rest;
  unsigned long opts = 0x1;

  ret = demangle ((char *) mangledSym, &rest, opts);

  if (ret != NULL)
    return strdup (text (ret));
  else
    return NULL;
}

#elif defined(DEMANGLE_Compaq)

#include <demangle.h>
#include <string.h>

char *
memPdemangle (const char *mangledSym)
{
  char out[1024];

  MLD_demangle_string (mangledSym, out, 1024,
                       MLD_SHOW_INFO | MLD_SHOW_DEMANGLED_NAME);

  return strdup (out);
}


#elif defined(DEMANGLE_GNU)

#ifdef HAVE_DEMANGLE_H
#include "demangle.h"
#else
#define DMGL_PARAMS    (1 << 0)
#define DMGL_ANSI      (1 << 1)
extern char *cplus_demangle (const char *mangled, int options);
#endif

char *
memPdemangle (const char *mangledSym)
{
  return cplus_demangle (mangledSym, DMGL_ANSI | DMGL_PARAMS);
}

#endif


void
find_address_in_section (abfd, section, data)
     bfd *abfd;
     asection *section;
     PTR data;
{
  bfd_vma vma;
  bfd_size_type size;
  bfd_vma local_pc = pc;
  char addr_buf1[24], addr_buf2[24], addr_buf3[24];

  assert (abfd != NULL);
  if (found)
    return;

#ifdef OSF1
  local_pc = pc | 0x100000000;
#elif defined(UNICOS_mp)
  /* use only least significant 32-bits */
  local_pc = pc & 0xFFFFFFFF;
#elif defined(AIX)
  local_pc = pc;
  if (memPi.obj_mode == 32)
    local_pc -= 0x10000000;
  else
    local_pc &= 0x00000000FFFFFFFF;
  local_pc += memPi.text_start;
  memPi_msg_debug ("pc is %s, text_start is %s, local_pc is %s\n",
                   memP_format_address ((void *) pc, addr_buf1),
                   memP_format_address ((void *) memPi.text_start, addr_buf2),
                   memP_format_address ((void *) local_pc, addr_buf3));
#else
  local_pc = pc /*& (~0x10000000) */ ;
#endif

  if ((bfd_get_section_flags (abfd, section) & SEC_ALLOC) == 0)
  {
    memPi_msg_debug ("failed bfd_get_section_flags\n");
    return;
  }
  vma = bfd_get_section_vma (abfd, section);

  if (local_pc < vma)
  {
    memPi_msg_debug
      ("failed bfd_get_section_vma: local_pc=%s  vma=%s\n",
       memP_format_address ((void *) (unsigned long) local_pc, addr_buf1),
       memP_format_address ((void *) (unsigned long) vma, addr_buf2));
    return;
  }

/* The following define addresses binutils modifications between
 * version 2.15 and 2.15.96
 */
#ifdef HAVE_BFD_GET_SECTION_SIZE
  size = bfd_get_section_size (section);
#else
  if (section->reloc_done)
    size = bfd_get_section_size_after_reloc (section);
  else
    size = bfd_get_section_size_before_reloc (section);
#endif

  if (local_pc >= vma + size)
  {
    memPi_msg_debug ("PC not in section: pc=%s vma=%s-%s\n",
                     memP_format_address ((void *) (unsigned long) local_pc,
                                          addr_buf1),
                     memP_format_address ((void *) (unsigned long) vma,
                                          addr_buf2),
                     memP_format_address ((void *) (unsigned long) (vma +
                                                                    size),
                                          addr_buf3));
    return;
  }


  found = bfd_find_nearest_line (abfd, section, syms, local_pc - vma,
                                 &filename, &functionname, &line);

  if (!found)
  {
    memPi_msg_debug ("bfd_find_nearest_line failed for : pc=%s vma=%s-%s\n",
                     memP_format_address ((void *) (unsigned long) local_pc,
                                          addr_buf1),
                     memP_format_address ((void *) (unsigned long) vma,
                                          addr_buf2),
                     memP_format_address ((void *) (unsigned long) (vma +
                                                                    size),
                                          addr_buf3));
  }

  memPi_msg_debug ("bfd_find_nearest_line for : pc=%s vma=%s-%s\n",
                   memP_format_address ((void *) (unsigned long) local_pc,
                                        addr_buf1),
                   memP_format_address ((void *) (unsigned long) vma,
                                        addr_buf2),
                   memP_format_address ((void *) (unsigned long) (vma + size),
                                        addr_buf3));

  memPi_msg_debug ("                 returned : %s:%s:%u\n",
                   filename, functionname, line);
}


int
memP_find_src_loc (void *i_addr_hex, char **o_file_str,
                   unsigned int *o_lineno, char **o_funct_str)
{
  char buf[128];
  char addr_buf[24];

  if (i_addr_hex == NULL)
  {
    memPi_msg_debug
      ("memP_find_src_loc returning failure as i_addr_hex == NULL\n");
    return 1;
  }

  if (abfd == NULL)
  {
    memPi_msg_debug ("memP_find_src_loc returning failure as abfd == NULL\n");
    return 1;
  }

  snprintf (buf, 128, "%s", memP_format_address (i_addr_hex, addr_buf));
  pc = bfd_scan_vma (buf, NULL, 16);

  /* jsv hack - trim high bit off of address */
  /*  pc &= !0x10000000; */
  found = FALSE;

  bfd_map_over_sections (abfd, find_address_in_section, (PTR) NULL);

  if (!found)
  {
    memPi_msg_debug ("returning not found in memP_find_src_loc\n");
    return 1;
  }
  else
  {
    /* determine the function name */
    if (functionname == NULL || *functionname == '\0')
    {
      *o_funct_str = strdup ("[unknown]");
    }
    else
    {
      char *res = NULL;

#if defined(DEMANGLE_IBM) || defined(DEMANGLE_Compaq) || defined(DEMANGLE_GNU)
      res = memPdemangle (functionname);
#endif
      if (res == NULL)
      {
/*	      *o_funct_str = strdup (functionname);*/
        *o_funct_str = memPi_escape_xml_chars (functionname);
      }
      else
      {
              /**o_funct_str = res;*/
        *o_funct_str = memPi_escape_xml_chars (res);
      }

#if defined(DEMANGLE_IBM) || defined(DEMANGLE_Compaq) || defined(DEMANGLE_GNU)
      memPi_msg_debug ("attempted demangle %s->%s\n", functionname,
                       *o_funct_str);
#endif
    }

    /* set the filename and line no */
    if (memPi.baseNames == 0 && filename != NULL)
    {
      char *h;
      h = strrchr (filename, '/');
      if (h != NULL)
        filename = h + 1;
    }

    *o_lineno = line;
    *o_file_str = strdup (filename ? filename : "[unknown]");

    memPi_msg_debug ("BFD: %s -> %s:%u:%s\n", buf, *o_file_str, *o_lineno,
                     *o_funct_str);
  }
  return 0;
}

void
get_section_sizes (abfd, section, data)
     bfd *abfd;
     asection *section;
     PTR data;
{
  flagword flags;
  long size;

  flags = bfd_get_section_flags (abfd, section);

  if ((flags & SEC_ALLOC) == 0)
    return;

  size = (long) bfd_get_section_size (section);

  if ((flags & SEC_CODE) != 0 || (flags & SEC_READONLY) != 0)
    memPi.text += size;
  else if ((flags & SEC_HAS_CONTENTS) != 0)
    memPi.data += size;
  else
    memPi.bss += size;
}

int
open_bfd_executable (char *filename)
{
  char *target = NULL;
  char **matching = NULL;
  long storage;
  long symcount;
  unsigned int size;

  memPi_msg_debug_level (MEMP_DEBUG_MAJ,
                         "Entered open_bfd_executable for %s\n", filename);

  if (abfd != NULL)
    return 1;

  if (filename == NULL)
  {
    memPi_msg_warn ("Executable filename is NULL!\n");
    memPi_msg_warn
      ("If this is a Fortran application, you may be using the incorrect memP library.\n");
    return 0;
  }

#ifdef AIX

  /*  Kludge to get XCOFF text_start value to feed an approriate address
     value to bfd for looking up source info
   */
  memPi.text_start = memPi_get_text_start (filename);
#endif

  bfd_init ();
  /* set_default_bfd_target (); */
  memPi_msg_debug ("opening filename %s\n", filename);
  abfd = bfd_openr (filename, target);
  if (abfd == NULL)
  {
    memPi_msg_warn ("could not open filename %s\n", filename);
    return 0;
  }
  if (bfd_check_format (abfd, bfd_archive))
  {
    memPi_msg_warn ("can not get addresses from archive");
    bfd_close (abfd);
    return 0;
  }
  if (!bfd_check_format_matches (abfd, bfd_object, &matching))
  {
    char *curr_match;
    if (matching != NULL)
    {
      for (curr_match = matching[0]; curr_match != NULL; curr_match++)
        memPi_msg_debug ("found matching type %s\n", curr_match);
      free (matching);
    }
    memPi_msg_warn ("matching failed");
    bfd_close (abfd);
    return 0;
  }

#if 1
  if ((bfd_get_file_flags (abfd) & HAS_SYMS) == 0)
  {
    memPi_msg_warn ("No symbols in the executable\n");
    bfd_close (abfd);
    return 0;
  }
#endif
  /* TODO: move this to the begining of the process so that the user
     knows before the application begins */
  storage = bfd_get_symtab_upper_bound (abfd);
  if (storage < 0)
  {
    memPi_msg_warn ("storage < 0");
    bfd_close (abfd);
    return 0;
  }

  symcount = bfd_read_minisymbols (abfd, FALSE, (void *) &syms, &size);
  if (symcount == 0)
    symcount =
      bfd_read_minisymbols (abfd, TRUE /* dynamic */ , (void *) &syms, &size);

  if (symcount < 0)
  {
    memPi_msg_warn ("symcount < 0");
    bfd_close (abfd);
    return 0;
  }
  else
  {
    memPi_msg_debug ("\n");
    memPi_msg_debug ("found %d symbols in file [%s]\n", symcount, filename);
  }

  bfd_map_over_sections (abfd, get_section_sizes, (PTR) NULL);

  return 1;
}

void
close_bfd_executable ()
{
  if (memPi.do_lookup == 1)
    bfd_close (abfd);
}

#elif !defined(USE_LIBDWARF)

int
memP_find_src_loc (void *i_addr_hex, char **o_file_str, unsigned int *o_lineno,
                   char **o_funct_str)
{
  return 1;                     /*  failure  */
}


/*  Source lookup by other means not supported at this time.  */
int
memPi_query_src (callsite_stats_t * p)
{
  return 0;
}


#endif /* ENABLE_BFD */

#ifdef AIX
/*  Get the text_start value from the auxiliary header
    to provide BFD with pc values that match source information.
*/
#define __XCOFF32__
#define __XCOFF64__
#include <fcntl.h>
#include <filehdr.h>
#include <aouthdr.h>
#include <scnhdr.h>

unsigned long long
memPi_get_text_start (char *filename)
{
  int fh;
  short magic;
  FILHDR FileHeader32;
  FILHDR_64 FileHeader64;
  AOUTHDR AoutHeader32;
  AOUTHDR_64 AoutHeader64;
  SCNHDR SectHeader32;
  SCNHDR_64 SectHeader64;

  unsigned long long text_start = 0;
  int count = 0;

  fh = open (filename, O_RDONLY);
  if (fh == -1)
    return 0;

  read (fh, &magic, 2);
  memPi_msg_debug ("magic is 0x%x\n", magic);
  lseek (fh, 0, 0);

  if (magic == 0x01DF)          /* 32-bit  */
  {
    memPi.obj_mode = 32;
    read (fh, &FileHeader32, sizeof (FILHDR));
    memPi_msg_debug ("aout size is %d\n", FileHeader32.f_opthdr);
    read (fh, &AoutHeader32, FileHeader32.f_opthdr);
    memPi_msg_debug ("text start is 0x%0x\n", AoutHeader32.o_text_start);

    while (count++ < FileHeader32.f_nscns)
    {
      read (fh, &SectHeader32, sizeof (SCNHDR));
      memPi_msg_debug ("found header name %s\n", SectHeader32.s_name);
      memPi_msg_debug ("found header raw ptr 0x%0x\n", SectHeader32.s_scnptr);

      if (SectHeader32.s_flags & STYP_TEXT)
        text_start = AoutHeader32.o_text_start - SectHeader32.s_scnptr;
    }
  }
  else if (magic == 0x01EF || magic == 0x01F7)  /*  64-bit  */
  {
    memPi.obj_mode = 64;
    read (fh, &FileHeader64, sizeof (FILHDR_64));
    memPi_msg_debug ("aout size is %d\n", FileHeader64.f_opthdr);
    read (fh, &AoutHeader64, FileHeader64.f_opthdr);
    memPi_msg_debug ("text start is 0x%0llx\n", AoutHeader64.o_text_start);

    while (count++ < FileHeader64.f_nscns)
    {
      read (fh, &SectHeader64, sizeof (SCNHDR_64));
      memPi_msg_debug ("found header name %s\n", SectHeader64.s_name);
      memPi_msg_debug ("found header raw ptr 0x%0llx\n",
                       SectHeader64.s_scnptr);

      if (SectHeader64.s_flags & STYP_TEXT)
        text_start = AoutHeader64.o_text_start - SectHeader64.s_scnptr;
    }
  }
  else
  {
    memPi_msg_debug ("invalid magic number.\n");
    return 0;
  }

  memPi_msg_debug ("text_start is 0x%0llx\n", text_start);
  close (fh);

  return text_start;
}

#endif



/* 

<license>

Copyright (c) 2010, Lawrence Livermore National Security, LLC.
Produced at the Lawrence Livermore National Laboratory
Written by Christopher Chambreau <chcham@llnl.gov>.
CODE-OCEC-10-010
All rights reserved.

This file is part of memP. For details, see http://memp.sourceforge.net .
Please also read this link – Additional BSD Notice.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  o  Redistributions of source code must retain the above copyright notice, 
     this list of conditions and the disclaimer below.

  o  Redistributions in binary form must reproduce the above copyright notice, 
     this list of conditions and the disclaimer (as noted below) in the 
     documentation and/or other materials provided with the distribution.

  o  Neither the name of the LLNS/LLNL nor the names of its contributors may be
     used to endorse or promote products derived from this software without 
     specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL LAWRENCE LIVERMORE NATIONAL
SECURITY, LLC, THE U.S. DEPARTMENT OF ENERGY OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.


Additional BSD Notice

1. This notice is required to be provided under our contract with the U.S.
Department of Energy (DOE). This work was produced at Lawrence Livermore
National Laboratory under Contract No. DE-AC52-07NA27344 with the DOE.

2. Neither the United States Government nor Lawrence Livermore National
Security, LLC nor any of their employees, makes any warranty, express or
implied, or assumes any liability or responsibility for the accuracy,
completeness, or usefulness of any information, apparatus, product, or process
disclosed, or represents that its use would not infringe privately-owned rights.

3. Also, reference herein to any specific commercial products, process, or
services by trade name, trademark, manufacturer or otherwise does not
necessarily constitute or imply its endorsement, recommendation, or favoring by
the United States Government or Lawrence Livermore National Security, LLC.
The views and opinions of authors expressed herein do not necessarily state or
reflect those of the United States Government or Lawrence Livermore National
Security, LLC, and shall not be used for advertising or product endorsement
purposes.

</license>

*/

/* eof */
