/* -*- C -*- 

   memP Heap Profiler ( http://memp.sourceforge.net/ )

   Please see COPYRIGHT AND LICENSE information at the end of this file.

   ----- 

   pcontrol.c -- implement pcontrol for MEMP

 */

#ifndef lint
static char *svnid = "$Id: pcontrol.c 369 2006-10-05 19:21:12Z chcham $";
#endif

#include <float.h>
#include "memPi.h"
#include "symbols.h"

void
memPi_reset_callsite_data ()
{
  int ac, ndx;
  callsite_stats_t **av;
  callsite_stats_t *csp = NULL;

  /* gather local task data */
  h_gather_data (memPi.task_callsite_stats, &ac, (void ***) &av);

  for (ndx = 0; ndx < ac; ndx++)
  {
    csp = av[ndx];

    csp->maxFree = 0;
    csp->minFree = DBL_MAX;
    csp->maxAlloc = 0;
    csp->minAlloc = DBL_MAX;

    csp->count = 0;
    csp->cumulativeTime = 0;
    csp->cumulativeAlloc = 0;
    csp->cumulativeFree = 0;
  }

  if (time (&memPi.start_timeofday) == (time_t) - 1)
  {
    memPi_msg_warn ("Could not get time of day from time()\n");
  }

  memPi_GETTIME (&memPi.startTime);
  memPi.cumulativeTime = 0;

  memPi_free (av);
}


static int
memPi_MPI_Pcontrol (const int flag)
{
  memP_TIMER dur;
  memPi_msg_debug ("MPI_Pcontrol encountered: flag = %d\n", flag);

  if (flag == 0)
  {
    if (!memPi.enabled)
      memPi_msg_warn
        ("MPI_Pcontrol trying to disable MEMP while it is already disabled.\n");

    memPi_GETTIME (&memPi.endTime);
    dur = (memPi_GETTIMEDIFF (&memPi.endTime, &memPi.startTime) / 1000000.0);
    memPi.cumulativeTime += dur;
    assert (memPi.cumulativeTime >= 0);
    memPi.enabled = 0;
  }
  else if (flag == 2)
  {
    memPi_reset_callsite_data ();
  }
  else if (flag == 3)
  {
    memPi_generateReport (memPi_style_verbose);
    memPi_GETTIME (&memPi.startTime);
  }
  else if (flag == 4)
  {
    memPi_generateReport (memPi_style_text_concise);
    memPi_GETTIME (&memPi.startTime);
  }
  else
  {
    if (memPi.enabled)
      memPi_msg_warn
        ("MPI_Pcontrol trying to enable MEMP while it is already enabled.\n");

    memPi.enabled = 1;
    memPi.enabledCount++;
    memPi_GETTIME (&memPi.startTime);
  }

  return MPI_SUCCESS;
}

int
MPI_Pcontrol (const int flag, ...)
{
  return memPi_MPI_Pcontrol (flag);
}

int
F77_MPI_PCONTROL (int *flag, ...)
{
  return memPi_MPI_Pcontrol (*flag);
}



/* 

<license>

Copyright (c) 2010, Lawrence Livermore National Security, LLC.
Produced at the Lawrence Livermore National Laboratory
Written by Christopher Chambreau <chcham@llnl.gov>.
CODE-OCEC-10-010
All rights reserved.

This file is part of memP. For details, see http://memp.sourceforge.net .
Please also read this link – Additional BSD Notice.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  o  Redistributions of source code must retain the above copyright notice, 
     this list of conditions and the disclaimer below.

  o  Redistributions in binary form must reproduce the above copyright notice, 
     this list of conditions and the disclaimer (as noted below) in the 
     documentation and/or other materials provided with the distribution.

  o  Neither the name of the LLNS/LLNL nor the names of its contributors may be
     used to endorse or promote products derived from this software without 
     specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL LAWRENCE LIVERMORE NATIONAL
SECURITY, LLC, THE U.S. DEPARTMENT OF ENERGY OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.


Additional BSD Notice

1. This notice is required to be provided under our contract with the U.S.
Department of Energy (DOE). This work was produced at Lawrence Livermore
National Laboratory under Contract No. DE-AC52-07NA27344 with the DOE.

2. Neither the United States Government nor Lawrence Livermore National
Security, LLC nor any of their employees, makes any warranty, express or
implied, or assumes any liability or responsibility for the accuracy,
completeness, or usefulness of any information, apparatus, product, or process
disclosed, or represents that its use would not infringe privately-owned rights.

3. Also, reference herein to any specific commercial products, process, or
services by trade name, trademark, manufacturer or otherwise does not
necessarily constitute or imply its endorsement, recommendation, or favoring by
the United States Government or Lawrence Livermore National Security, LLC.
The views and opinions of authors expressed herein do not necessarily state or
reflect those of the United States Government or Lawrence Livermore National
Security, LLC, and shall not be used for advertising or product endorsement
purposes.

</license>

*/


/* eof */
