/* -*- C -*- 

   memP Heap Profiler ( http://memp.sourceforge.net/ )

   Please see COPYRIGHT AND LICENSE information at the end of this file.

   ----- 

   report-xml.c -- Toolgear XML reporting functions

*/

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <values.h>
#include <fcntl.h>

#ifndef lint
static char *svnid = "$Id: report-xml.c 383 2006-12-08 18:30:15Z chcham $";
#endif

#include "memPi.h"
#define memP_func_name(idx,op) \
          (idx) != 0 ? "" : ( \
          strncmp(memPi.lookup[(op) - memPi_BASE].name, "MPI", 3) == 0 ?  \
            &(memPi.lookup[(op) - memPi_BASE].name[4]) : \
            memPi.lookup[(op) - memPi_BASE].name)
static int xml_indent = 0;



void
print_xml_section_comment (FILE * fp, char *fmt, ...)
{
  va_list args;
  va_start (args, fmt);
  fprintf (fp, "\n\n<!--  ");
  vfprintf (fp, fmt, args);
  va_end (args);
  fprintf (fp, "  -->\n\n");
}


void
print_xml_indent_line (FILE * fp, char *fmt, ...)
{
  va_list args;
  char *ccr;
  int id_mod = 0;
  const int id_width = 2;

  va_start (args, fmt);

  for (ccr = fmt; ccr != NULL && *ccr != '\0'; ccr++)
  {
    if (*ccr == '<')
    {
      if (*(ccr + 1) == '/')
      {
        id_mod -= id_width;
      }
      else if (*(ccr + 1) != '!')
        id_mod += id_width;
    }
  }

  if (id_mod < 0)
    xml_indent += id_mod;

  fprintf (fp, "%*s", xml_indent, "");
  vfprintf (fp, fmt, args);
  va_end (args);
  if (id_mod > 0)
    xml_indent += id_mod;
}


void
print_xml_report_header (FILE * fh, char *fn)
{
  print_xml_section_comment (fh, "REPORT HEADER");
  print_xml_indent_line (fh, "<tool_gear>\n\n");
  print_xml_indent_line (fh, "<format>1</format><version>2.02</version>\n");
  print_xml_indent_line (fh, "<tool_title>memPView</tool_title>\n");
  print_xml_indent_line (fh, "<status>Reading in %s</status>\n", fn);
}


void
print_xml_messsage_folder (FILE * fh, char *tag, char *title)
{
  print_xml_section_comment (fh, "MESSAGE FOLDER : %s", title);

  print_xml_indent_line (fh, "<message_folder>\n");
  print_xml_indent_line (fh, "<tag>%s</tag>\n", tag);
  print_xml_indent_line (fh, "<title>%s</title>\n", title);
  print_xml_indent_line (fh, "<if_empty>show</if_empty>\n");
  print_xml_indent_line (fh, "</message_folder>\n");
}

void
format_xml_callsite (char *buf, char *file, int line, char *func)
{
  sprintf (buf, "<site><file>%30s</file><line>%8d</line><desc>%30s</desc>\n",
           file, line, func);
}


char *
memPi_escape_xml_chars (const char *xml_buf)
{
  char *obuf;
  struct escape_char_t
  {
    char c;
    char *esc_str;
  } xml_esc_table[5] =
  {
    {
    '&', "&amp;"},
    {
    '\'', "&apos;"},
    {
    '"', "&quot;"},
    {
    '<', "&lt;"},
    {
    '>', "&gt;"}
  };
  int i, j;
  const char *ip;
  char *op;
  size_t new_size = 0;

/*  fprintf (stderr, "In memPi_escape_xml_chars for %s\n", xml_buf);*/
  for (ip = xml_buf, new_size = 0; *ip != '\0'; ip++)
  {
    for (j = 0; j < 5; j++)
      if (*ip == xml_esc_table[j].c)
      {
        new_size += strlen (xml_esc_table[j].esc_str);
      }
      else
      {
        new_size++;
      }
  }

  new_size++;
/*  fprintf (stderr, "In memPi_escape_xml_chars original size %d new_size %d\n",
	   strlen (xml_buf), new_size);*/

  obuf = memPi_malloc (new_size);
  memset (obuf, '\0', new_size);

  if (obuf != NULL)
  {

    for (ip = xml_buf, op = obuf, i = 0; ip != NULL && *ip != '\0'; i++, ip++)
    {
      int esc_found = 0;

      for (j = 0; j < 5; j++)
      {
/*	    fprintf (stderr, "In memPi_escape_xml_chars checking %c for %c\n",
		     *ip, xml_esc_table[j].c);*/
        if (*ip == xml_esc_table[j].c)
        {
/*		fprintf (stderr, "Found match! should add %s to %s\n", xml_esc_table[j].esc_str, obuf);*/

          strcat (obuf, xml_esc_table[j].esc_str);
          op += strlen (xml_esc_table[j].esc_str);
          esc_found = 1;
        }
      }
      if (esc_found == 0)
      {
        *op = *ip;
        op++;
      }
    }
    *op = '\0';
  }
  else
  {
    memPi_msg_warn ("Failed to alloc xml obuf!\n");
  }

/*  fprintf (stderr, "In memPi_escape_xml_chars returning %s\n", obuf);*/
  return obuf;
}


void
print_xml_annot (FILE * fh, char *title, callsite_src_id_cache_entry_t * cs)
{
  char *abuf = NULL;
  char addr_buf[4096];
  int j, stack_continue_flag;
  size_t MEMP_MAX_STR_LEN = 4096;
  char lb[12];

  abuf = memPi_malloc (MEMP_MAX_STR_LEN);

  print_xml_indent_line (fh, "<annot>\n");

  if ( cs == NULL )
  {
    memPi_msg_warn("In print_xml_annot, received NULL cache entry.\n");
    return;
  }

  for (j = 0, stack_continue_flag = 1;
       (j < MEMP_CALLSITE_STACK_DEPTH) && (cs->filename[j] != NULL) &&
       stack_continue_flag == 1; j++)
  {
    abuf[0] = '\0';
    print_xml_indent_line (fh, "<site>\n");
    strcat (abuf, "<file>");
    if (cs->line[j] == 0 &&
        (strcmp (cs->filename[j], "[unknown]") == 0 ||
         strcmp (cs->functname[j], "[unknown]") == 0))
    {
      strcat (abuf, memP_format_address (cs->pc[j], addr_buf));
      strcat (abuf, "</file><line>");
      strcat (abuf, (char *) memP_format_address (cs->pc[j], addr_buf));
      strcat (abuf, "</line><desc>");
      strcat (abuf, cs->functname[j]);
      strcat (abuf, "</desc>\n");
    }
    else
    {
      strcat (abuf, cs->filename[j]);
      strcat (abuf, "</file><line>");
      snprintf (lb, 12, "%u", cs->line[j]);
      strcat (abuf, lb);
      strcat (abuf, "</line><desc>");
      strcat (abuf, cs->functname[j]);
      strcat (abuf, "</desc>\n");
    }
    print_xml_indent_line (fh, abuf);
    print_xml_indent_line (fh, "</site>\n");

    /*  Do not bother printing stack frames above main   */
    if (strcmp (cs->functname[j], "main") == 0
        || strcmp (cs->functname[j], ".main") == 0
        || strcmp (cs->functname[j], "MAIN__") == 0)
      stack_continue_flag = 0;
  }

  print_xml_indent_line (fh, "</annot>\n");

  memPi_free (abuf);
}


void *
format_stack_trace (callsite_src_id_cache_entry_t * cs)
{
  char *abuf = NULL;
  char *addr_buf = NULL;
  char *line_buf = NULL;
  int j, stack_continue_flag;
  size_t MEMP_MAX_STR_LEN = 4096;

  abuf = memPi_malloc (MEMP_MAX_STR_LEN);
  addr_buf = memPi_malloc (MEMP_MAX_STR_LEN);
  line_buf = memPi_malloc (MEMP_MAX_STR_LEN);
  abuf[0] = '\0';

  for (j = 0, stack_continue_flag = 1;
       (j < MEMP_CALLSITE_STACK_DEPTH) && (cs->filename[j] != NULL) &&
       stack_continue_flag == 1; j++)
  {
    line_buf[0] = '\0';
    memP_format_address (cs->pc[j], addr_buf);
    if (cs->line[j] == 0 &&
        (strcmp (cs->filename[j], "[unknown]") == 0 ||
         strcmp (cs->functname[j], "[unknown]") == 0))
    {
      snprintf (line_buf, MEMP_MAX_STR_LEN, "    %-30s %11s   %s\n",
                "", addr_buf, cs->functname[j]);
      strcat (abuf, line_buf);
    }
    else
    {
      snprintf (line_buf, MEMP_MAX_STR_LEN,
                "    %-30s %11u   %35s   %s\n", cs->filename[j],
                cs->line[j], cs->functname[j], addr_buf);
      strcat (abuf, line_buf);
    }

    /*  Do not bother printing stack frames above main   */
    if (strcmp (cs->functname[j], "main") == 0
        || strcmp (cs->functname[j], ".main") == 0
        || strcmp (cs->functname[j], "MAIN__") == 0)
      stack_continue_flag = 0;
  }

  free (addr_buf);
  free (line_buf);

  return abuf;
}


void
print_xml_messsage (FILE * fh, char *folder, char *heading, char *body,
                    char *annot_title, callsite_src_id_cache_entry_t * cs)
{
  print_xml_section_comment (fh, "MESSAGE");
  print_xml_indent_line (fh, "<message>\n");
  print_xml_indent_line (fh, "<folder>%s</folder>\n", folder);
  print_xml_indent_line (fh, "<heading>%s</heading>\n", heading);
  if (body != NULL)
  {
    print_xml_indent_line (fh, "<body>%s\n", body);
    print_xml_indent_line (fh, "</body>\n");
  }
  if (cs != NULL)
    print_xml_annot (fh, annot_title, cs);
  print_xml_indent_line (fh, "</message>\n");
}


void
print_xml_message_fmt (FILE * fp, char *folder, char *fmt, ...)
{
  va_list args;
  va_start (args, fmt);

  print_xml_section_comment (fp, "MESSAGE");
  print_xml_indent_line (fp, "<message>\n");
  print_xml_indent_line (fp, "<folder>%s</folder>\n", folder);
  print_xml_indent_line (fp, "<heading>");
  vfprintf (fp, fmt, args);
  print_xml_indent_line (fp, "</heading>\n");
  print_xml_indent_line (fp, "</message>\n");
  va_end (args);
}


void
print_xml_report_footer (FILE * fh, char *message)
{
  print_xml_section_comment (fh, "REPORT FOOTER");
  print_xml_indent_line (fh, "<status>%s</status>\n", message);
  print_xml_indent_line (fh, "\n");
  print_xml_indent_line (fh, "</tool_gear>\n");
}


void
memPi_print_xml_task_assignment (FILE * fp)
{
  int i;
  char buf[256];

  print_xml_messsage_folder (fp, "task_assignment", "MPI Task Assignment");
  for (i = 0; i < memPi.size; i++)
  {
    snprintf (buf, 256, "%6d %s",
              memPi.global_task_info[i].rank,
              memPi_escape_xml_chars (memPi.global_task_info[i].hostname));

    print_xml_messsage (fp, "task_assignment", buf, NULL, NULL, NULL);
  }
}

extern int callsite_src_id_cache_sort_by_id (const void *a, const void *b);
void
memPi_print_xml_callsites (FILE * fp)
{
  int i, ac;
  char buf[256];
  callsite_src_id_cache_entry_t **av;
  int fileLenMax = 18;
  int funcLenMax = 24;

  h_gather_data (callsite_src_id_cache, &ac, (void ***) &av);
  snprintf (buf, 256, "Allocation Callsites: %d", ac);

  qsort (av, (size_t) ac, sizeof (void *), callsite_src_id_cache_sort_by_id);
  print_xml_messsage_folder (fp, "callsites", "Allocation Callsites");

  /* Find longest file and function names for formatting */
  for (i = 0; i < ac; i++)
  {
    int j;
    size_t currlen;

    for (j = 0;
         (j < MEMP_CALLSITE_STACK_DEPTH) && (av[i]->filename[j] != NULL); j++)
    {
      currlen = strlen (av[i]->filename[j]);
      fileLenMax = currlen > fileLenMax ? currlen : fileLenMax;
      currlen = strlen (av[i]->functname[j]);
      funcLenMax = currlen > funcLenMax ? currlen : funcLenMax;
    }
  }

#if 0
  fprintf (fp, "%3s %3s %-*s %5s %-*s %s\n",
           "ID", "Lev", fileLenMax, "File/Address", "Line", funcLenMax,
           "Parent_Funct", "Call");
#endif

  for (i = 0; i < ac; i++)
  {
    /*
     * For each callsite, construct:
     * 
     *   header : <base call site info (func[#])>
     *   body   : <stack trace>
     *   annot  : <stack trace>
     */

    if (av[i]->op != memPi_ID_free && av[i]->filename[0] != NULL)
    {
      char buf[4096];
      char *bt = NULL;

      /*fprintf (stderr, "checking av[%d]->functname %s\n", i,
         av[i]->functname[0]); */
      snprintf (buf, 4096, "%6d  %8s  %s [%d]  (%s : %u)", av[i]->id,
                memP_func_name (0, av[i]->op), av[i]->functname[0],
                av[i]->id, av[i]->filename[0], av[i]->line[0]);
      bt = format_stack_trace (av[i]);

      print_xml_messsage (fp, "callsites", buf, bt, "Stack Trace", av[i]);
      free (bt);
    }
#if 0
    for (j = 0, stack_continue_flag = 1;
         (j < MEMP_CALLSITE_STACK_DEPTH) && (av[i]->filename[j] != NULL) &&
         stack_continue_flag == 1; j++)
    {
      if (av[i]->line[j] == 0 &&
          (strcmp (av[i]->filename[j], "[unknown]") == 0 ||
           strcmp (av[i]->functname[j], "[unknown]") == 0))
      {
        fprintf (fp, "%3d %3d %-*s %-*s %s\n",
                 av[i]->id,
                 j,
                 fileLenMax + 6,
                 memP_format_address (av[i]->pc[j], addr_buf),
                 funcLenMax,
                 av[i]->functname[j], memP_func_name (j, av[i]->op));
      }
      else
      {
        fprintf (fp, "%3d %3d %-*s %5d %-*s %s\n",
                 av[i]->id,
                 j,
                 fileLenMax,
                 av[i]->filename[j], av[i]->line[j],
                 funcLenMax,
                 av[i]->functname[j], memP_func_name (j, av[i]->op));
      }
      /*  Do not bother printing stack frames above main   */
      if (strcmp (av[i]->functname[j], "main") == 0
          || strcmp (av[i]->functname[j], ".main") == 0
          || strcmp (av[i]->functname[j], "MAIN__") == 0)
        stack_continue_flag = 0;
    }
#endif
  }
  memPi_free (av);
}


static int
callsite_sort_by_name_id_rank (const void *a, const void *b)
{
  int rc;
  callsite_stats_t *csp_1 = *(callsite_stats_t **) a;
  callsite_stats_t *csp_2 = *(callsite_stats_t **) b;

  MEMP_CALLSITE_STATS_COOKIE_ASSERT (csp_1);
  MEMP_CALLSITE_STATS_COOKIE_ASSERT (csp_2);

  if ((rc = strcmp (memPi.lookup[csp_1->op - memPi_BASE].name,
                    memPi.lookup[csp_2->op - memPi_BASE].name)) != 0)
  {
    return rc;
  }

#define express(f) {if ((csp_1->f) > (csp_2->f)) {return 1;} if ((csp_1->f) < (csp_2->f)) {return -1;}}
  express (csid);
  express (rank);
#undef express

  return 0;
}


static callsite_stats_t *
memPi_get_task_callsite_record (int rank, unsigned int op,
                                void *pc[MEMP_CALLSITE_STACK_DEPTH])
{
  callsite_stats_t *csp = NULL, key;
  int i;

  /*memPi_query_src (p); *//* sets the file/line in p */
  bzero (&key, sizeof (callsite_stats_t));
  key.op = op;
  key.csid = -1;
  key.rank = rank;
  key.cookie = MEMP_CALLSITE_STATS_COOKIE;
  for (i = 0; i < MEMP_CALLSITE_STACK_DEPTH; i++)
  {
    key.pc[i] = pc[i];
  }
  memPi_query_src (&key);

#if 0

  h_gather_data (memPi.global_callsite_stats, &ac, (void ***) &av);
  for (i = 0; i < ac; i++)
  {
    fprintf (stderr, "op : %d, rank : %d, pc[0] %p\n", av[i]->op,
             av[i]->rank, av[i]->pc[0]);
  }
#endif

  /*fprintf (stderr, "searching for rank %d, op %d, pc[0] %p\n", key.rank,
     key.op, key.pc[0]); */
  if ( NULL == memPi.global_callsite_stats )
    return NULL;

  if (NULL == h_search (memPi.global_callsite_stats, &key, (void **) &csp))
  {
    memPi_msg_warn ("Failed to find Task HWM call site!\n");
    csp = NULL;
  }

  return csp;
}


static int
task_info_sort_by_hwm (const void *p1, const void *p2)
{
  memPi_task_info_t *tip_1 = *(memPi_task_info_t **) p1;
  memPi_task_info_t *tip_2 = *(memPi_task_info_t **) p2;

  /*fprintf (stderr, "Comparing %f:%d and %f:%d\n", tip_1->taskHWM, tip_1->rank,
     tip_2->taskHWM, tip_2->rank); */

  return (int) (tip_2->taskHWM - tip_1->taskHWM);
}


void
memPi_print_xml_task_hwm (FILE * fp)
{
  int i;
  char header_buf[2048];
  char body_buf[2048];
  char temp_buf[2048];
  memP_TIMER hwm_time;

  callsite_stats_t *cs = NULL;
  callsite_src_id_cache_entry_t **av;
  int ac;
  char *bt = NULL;
  int indent = 2;
  int body_width = 60;
  memPi_task_info_t **tips = NULL;
  int dlw = 25;
  double sum = 0.0, mean = 0.0, stddev = 0.0;
  int *bucket_rank_list = NULL;

  header_buf[0] = '\0';
  body_buf[0] = '\0';

  /* Get callsite source information that can be accessed by call site id */
  h_gather_data (callsite_src_id_cache, &ac, (void ***) &av);
  qsort (av, ac, sizeof (void *), callsite_src_id_cache_sort_by_id);

  /* Create an array to access task info sorted by descending HWM 
     TODO: Why is this not done in place?  */
  tips =
    (memPi_task_info_t **) memPi_malloc (sizeof (memPi_task_info_t *) *
                                         memPi.size);
  for (i = 0; i < memPi.size; i++)
  {
    tips[i] = &(memPi.global_task_info[i]);
  }
  qsort (tips, memPi.size, sizeof (memPi_task_info_t *),
         task_info_sort_by_hwm);

  /* HWM high level XML folder  */
  print_xml_messsage_folder (fp, "task_hwm",
                             "Task High Water Mark (in decreasing HWM order)");

  //  If the user only wants a report from 1 task...
  if (memPi.taskHWMrank != -1)
  {
    tips[i] = &(memPi.global_task_info[memPi.taskHWMrank]);
    memPi_msg_debug ("Printing HWM info for task %d\n", tips[i]->rank);

    /*  TODO: Continue Group checking here!
       if ( i < (memPi.size-1) && tips[i]->HWM == tips[i+1]->HWM )
       bucket_rank_list = (int*) memPi_malloc (sizeof (int *) * memPi.size);
     */

    cs = NULL;

    //  Find call site based on stack trace
    if (tips[i]->taskHWM > 0)
    {
      cs =
        memPi_get_task_callsite_record (tips[i]->rank,
                                        tips[i]->taskHWMop,
                                        tips[i]->taskHWMpcs);
    }

    sum += tips[i]->taskHWM;

    if (cs != NULL)
    {
      //  Create Header XML Entry
      snprintf (header_buf, 2048,
                "Task %5d   HWM %.0f [%.0f:%.0f unk]   %s of %.0fB   Call site HWM %.0f in %s[%d]",
                tips[i]->rank,
                tips[i]->taskHWM, tips[i]->taskHWM - tips[i]->taskHWMunknown,
                tips[i]->taskHWMunknown, memP_func_name (0,
                                                         tips[i]->taskHWMop),
                tips[i]->taskHWMsize, cs->csHWM,
                av[cs->csid - 1]->functname[0], cs->csid);

      //  Create Body XML Entry
      snprintf (body_buf, 2048,
                "High Water Mark (HWM) information for %s[%d]  (%s:%u)\n",
                av[cs->csid - 1]->functname[0], cs->csid,
                av[cs->csid - 1]->filename[0], av[cs->csid - 1]->line[0]);

      snprintf (temp_buf, 2048,
                "%*s%-*s%.0f\n", indent, "", body_width,
                "Task HWM at this callsite: ", tips[i]->taskHWM);
      strcat (body_buf, temp_buf);

      snprintf (temp_buf, 2048,
                "%*s%-*s%.0f (of %lu)\n", indent, "", body_width,
                "Task HWM reached at task allocation count ",
                tips[i]->taskHWMtaskCount, tips[i]->task_alloc_cnt);
      strcat (body_buf, temp_buf);

      snprintf (temp_buf, 2048,
                "%*s%-*s%.0f (of %lld)\n", indent, "", body_width,
                "Task HWM reached at this callsite's allocation count",
                tips[i]->taskHWMcsCount, cs->count);
      strcat (body_buf, temp_buf);

      //memPi_PRINTTIME (&memPi.startTime);
      //memPi_PRINTTIME (&tips[i]->taskHWMts);

      hwm_time =
        (memPi_GETTIMEDIFF
         (&tips[i]->taskHWMts, &memPi.startTime) / 1000000.0);

      snprintf (temp_buf, 2048,
                "%*s%-*s%.0f (of %.0f sec)\n", indent, "", body_width,
                "Task HWM reached at runtime ", hwm_time, tips[i]->app_time);
      strcat (body_buf, temp_buf);

      snprintf (temp_buf, 2048,
                "%*s%-*s%.0f\n", indent, "", body_width,
                "Callsite HWM (may not be at Task HWM): ", cs->csHWM);
      strcat (body_buf, temp_buf);

      snprintf (temp_buf, 2048,
                "%*s%-*s%.0f\n", indent, "", body_width,
                "Task Low Water Mark at this callsite: ", cs->csLWM);
      strcat (body_buf, temp_buf);

      bt = format_stack_trace (av[cs->csid - 1]);

      strcat (body_buf, bt);

      print_xml_messsage (fp, "task_hwm", header_buf, body_buf,
                          "Stack Trace", av[cs->csid - 1]);
    }
    else
    {
      fprintf (stderr, "Callsite lookup failed!   Why?\n");
      //  This is the low-overhead, stacktrace-free report version
      //  Create Header XML Entry
      snprintf (header_buf, 2048, "Task %5d", tips[i]->rank);
      snprintf (header_buf, 2048,
                "Task %5d   HWM %.0f [%.0f:%.0f unk]   %s of %.0fB",
                tips[i]->rank,
                tips[i]->taskHWM, tips[i]->taskHWM - tips[i]->taskHWMunknown,
                tips[i]->taskHWMunknown, memP_func_name (0,
                                                         tips[i]->taskHWMop),
                tips[i]->taskHWMsize);

#if 0
      //  Create Body XML Entry
      snprintf (body_buf, 2048,
                "High Water Mark (HWM) information for %s[%d]  (%s:%u)\n",
                av[cs->csid - 1]->functname[0], cs->csid,
                av[cs->csid - 1]->filename[0], av[cs->csid - 1]->line[0]);

      snprintf (temp_buf, 2048,
                "%*s%-*s%.0f\n", indent, "", body_width,
                "Task HWM at this callsite: ", tips[i]->taskHWM);
      strcat (body_buf, temp_buf);

      snprintf (temp_buf, 2048,
                "%*s%-*s%.0f (of %lu)\n", indent, "", body_width,
                "Task HWM reached at task allocation count ",
                tips[i]->taskHWMtaskCount, tips[i]->task_alloc_cnt);
      strcat (body_buf, temp_buf);

      snprintf (temp_buf, 2048,
                "%*s%-*s%.0f (of %lld)\n", indent, "", body_width,
                "Task HWM reached at this callsite's allocation count",
                tips[i]->taskHWMcsCount, cs->count);
      strcat (body_buf, temp_buf);

      //memPi_PRINTTIME (&memPi.startTime);
      //memPi_PRINTTIME (&tips[i]->taskHWMts);

      hwm_time =
        (memPi_GETTIMEDIFF
         (&tips[i]->taskHWMts, &memPi.startTime) / 1000000.0);

      snprintf (temp_buf, 2048,
                "%*s%-*s%.0f (of %.0f sec)\n", indent, "", body_width,
                "Task HWM reached at runtime ", hwm_time, tips[i]->app_time);
      strcat (body_buf, temp_buf);

      snprintf (temp_buf, 2048,
                "%*s%-*s%.0f\n", indent, "", body_width,
                "Callsite HWM (may not be at Task HWM): ", cs->csHWM);
      strcat (body_buf, temp_buf);

      snprintf (temp_buf, 2048,
                "%*s%-*s%.0f\n", indent, "", body_width,
                "Task Low Water Mark at this callsite: ", cs->csLWM);
      strcat (body_buf, temp_buf);

      bt = format_stack_trace (av[cs->csid - 1]);

      strcat (body_buf, bt);
#endif

      print_xml_messsage (fp, "task_hwm", header_buf, NULL, NULL, NULL);
    }
    //free(bt);
  }
  else                          // Report on all tasks
  {
    for (i = 0; i < memPi.size; i++)
    {
      memPi_msg_debug ("Printing HWM info for task %d\n", tips[i]->rank);

      /*  TODO: Continue Group checking here!
         if ( i < (memPi.size-1) && tips[i]->HWM == tips[i+1]->HWM )
       */

      cs = NULL;

      //  Find call site based on stack trace
      if (tips[i]->taskHWM > 0)
      {
        memPi_msg_debug ("task %d: get_task_callsite_record.\n", tips[i]->rank);
        cs =
          memPi_get_task_callsite_record (tips[i]->rank,
                                          tips[i]->taskHWMop,
                                          tips[i]->taskHWMpcs);
        memPi_msg_debug ("task %d: after get_task_callsite_record.\n", tips[i]->rank);

        if (cs == NULL)
        {
          memPi_msg_debug
            ("failed to find HWM callsite for rank %d, op %d, pc[0] 0x%p!\n",
             tips[i]->rank, tips[i]->taskHWMop, tips[i]->taskHWMpcs[0]);
          if (memPi_debug >= MEMP_DEBUG_STACK)
            memPi_print_stack (tips[i]->taskHWMpcs);
        }
      }

      sum += tips[i]->taskHWM;

      if (cs != NULL)
      {

        //  Create Header XML Entry
        snprintf (header_buf, 2048,
                  "Task %5d   HWM %.0f [%.0f:%.0f unk]   %s of %.0fB   Call site HWM %.0f in %s[%d]",
                  tips[i]->rank,
                  tips[i]->taskHWM,
                  tips[i]->taskHWM - tips[i]->taskHWMunknown,
                  tips[i]->taskHWMunknown, memP_func_name (0,
                                                           tips
                                                           [i]->taskHWMop),
                  tips[i]->taskHWMsize, cs->csHWM,
                  av[cs->csid - 1]->functname[0], cs->csid);

        //  Create Body XML Entry
        snprintf (body_buf, 2048,
                  "High Water Mark (HWM) information for %s[%d]  (%s:%u)\n",
                  av[cs->csid - 1]->functname[0], cs->csid,
                  av[cs->csid - 1]->filename[0], av[cs->csid - 1]->line[0]);

        snprintf (temp_buf, 2048,
                  "%*s%-*s%.0f\n", indent, "", body_width,
                  "Task HWM at this callsite: ", tips[i]->taskHWM);
        strcat (body_buf, temp_buf);

        snprintf (temp_buf, 2048,
                  "%*s%-*s%.0f (of %lu)\n", indent, "", body_width,
                  "Task HWM reached at task allocation count ",
                  tips[i]->taskHWMtaskCount, tips[i]->task_alloc_cnt);
        strcat (body_buf, temp_buf);

        snprintf (temp_buf, 2048,
                  "%*s%-*s%.0f (of %lld)\n", indent, "", body_width,
                  "Task HWM reached at this callsite's allocation count",
                  tips[i]->taskHWMcsCount, cs->count);
        strcat (body_buf, temp_buf);

        //memPi_PRINTTIME (&memPi.startTime);
        //memPi_PRINTTIME (&tips[i]->taskHWMts);

        hwm_time =
          (memPi_GETTIMEDIFF
           (&tips[i]->taskHWMts, &memPi.startTime) / 1000000.0);

        snprintf (temp_buf, 2048,
                  "%*s%-*s%.0f (of %.0f sec)\n", indent, "", body_width,
                  "Task HWM reached at runtime ", hwm_time,
                  tips[i]->app_time);
        strcat (body_buf, temp_buf);

        snprintf (temp_buf, 2048,
                  "%*s%-*s%.0f\n", indent, "", body_width,
                  "Callsite HWM (may not be at Task HWM): ", cs->csHWM);
        strcat (body_buf, temp_buf);

        snprintf (temp_buf, 2048,
                  "%*s%-*s%.0f\n", indent, "", body_width,
                  "Task Low Water Mark at this callsite: ", cs->csLWM);
        strcat (body_buf, temp_buf);

        bt = format_stack_trace (av[cs->csid - 1]);

        strcat (body_buf, bt);

        print_xml_messsage (fp, "task_hwm", header_buf, body_buf,
                            "Stack Trace", av[cs->csid - 1]);
      }
      else
      {
        //  This is the low-overhead, stacktrace-free report version
        //  Create Header XML Entry
        snprintf (header_buf, 2048, "Task %5d", tips[i]->rank);
        snprintf (header_buf, 2048,
                  "Task %5d   HWM %.0f [%.0f:%.0f unk]   %s of %.0fB",
                  tips[i]->rank,
                  tips[i]->taskHWM,
                  tips[i]->taskHWM - tips[i]->taskHWMunknown,
                  tips[i]->taskHWMunknown, memP_func_name (0,
                                                           tips
                                                           [i]->taskHWMop),
                  tips[i]->taskHWMsize);

#if 0
        //  Create Body XML Entry
        snprintf (body_buf, 2048,
                  "High Water Mark (HWM) information for %s[%d]  (%s:%u)\n",
                  av[cs->csid - 1]->functname[0], cs->csid,
                  av[cs->csid - 1]->filename[0], av[cs->csid - 1]->line[0]);

        snprintf (temp_buf, 2048,
                  "%*s%-*s%.0f\n", indent, "", body_width,
                  "Task HWM at this callsite: ", tips[i]->taskHWM);
        strcat (body_buf, temp_buf);

        snprintf (temp_buf, 2048,
                  "%*s%-*s%.0f (of %lu)\n", indent, "", body_width,
                  "Task HWM reached at task allocation count ",
                  tips[i]->taskHWMtaskCount, tips[i]->task_alloc_cnt);
        strcat (body_buf, temp_buf);

        snprintf (temp_buf, 2048,
                  "%*s%-*s%.0f (of %lld)\n", indent, "", body_width,
                  "Task HWM reached at this callsite's allocation count",
                  tips[i]->taskHWMcsCount, cs->count);
        strcat (body_buf, temp_buf);

        //memPi_PRINTTIME (&memPi.startTime);
        //memPi_PRINTTIME (&tips[i]->taskHWMts);

        hwm_time =
          (memPi_GETTIMEDIFF
           (&tips[i]->taskHWMts, &memPi.startTime) / 1000000.0);

        snprintf (temp_buf, 2048,
                  "%*s%-*s%.0f (of %.0f sec)\n", indent, "", body_width,
                  "Task HWM reached at runtime ", hwm_time,
                  tips[i]->app_time);
        strcat (body_buf, temp_buf);

        snprintf (temp_buf, 2048,
                  "%*s%-*s%.0f\n", indent, "", body_width,
                  "Callsite HWM (may not be at Task HWM): ", cs->csHWM);
        strcat (body_buf, temp_buf);

        snprintf (temp_buf, 2048,
                  "%*s%-*s%.0f\n", indent, "", body_width,
                  "Task Low Water Mark at this callsite: ", cs->csLWM);
        strcat (body_buf, temp_buf);

        bt = format_stack_trace (av[cs->csid - 1]);

        strcat (body_buf, bt);
#endif

        print_xml_messsage (fp, "task_hwm", header_buf, NULL, NULL, NULL);
      }
      //free(bt);
    }

  }
  memPi_free (bucket_rank_list);


  /*  Use sorted HWM array to generate Statistics folder  */

  print_xml_messsage_folder (fp, "hwm_stats",
                             "Task High Water Mark Statistics");

  print_xml_message_fmt (fp, "hwm_stats", "%-*s : %.0f", dlw, "Max",
                         tips[0]->taskHWM);

  if (memPi.size % 2 == 0)
  {
    double median_mean =
      (tips[memPi.size / 2]->taskHWM +
       tips[(memPi.size / 2) - 1]->taskHWM) / 2;
    print_xml_message_fmt (fp, "hwm_stats", "%-*s : %.0f", dlw, "Median",
                           median_mean);
  }
  else
    print_xml_message_fmt (fp, "hwm_stats", "%-*s : %.0f", dlw, "Median",
                           tips[(memPi.size - 1) / 2]->taskHWM);

  mean = sum / memPi.size;
  print_xml_message_fmt (fp, "hwm_stats", "%-*s : %.0f", dlw, "Mean", mean);
  print_xml_message_fmt (fp, "hwm_stats", "%-*s : %.0f", dlw, "Min",
                         tips[memPi.size - 1]->taskHWM);

  /*  Calculate stddev  */
  for (i = 0; i < memPi.size; i++)
    stddev += (tips[i]->taskHWM - mean) * (tips[i]->taskHWM - mean);

  stddev = sqrt (stddev / memPi.size);
  print_xml_message_fmt (fp, "hwm_stats", "%-*s : %.0f", dlw, "Stddev",
                         stddev);
  print_xml_message_fmt (fp, "hwm_stats", "%-*s : %f", dlw,
                         "Coefficient of variation", stddev / mean);

  memPi_free (tips);
}


static int
callsite_sort_by_size_in_use (const void *a, const void *b)
{
  callsite_stats_t **a1 = (callsite_stats_t **) a;
  callsite_stats_t **b1 = (callsite_stats_t **) b;

  /* NOTE: want descending sort, so compares are reveresed */
  if ((*a1)->csCurAlloc < (*b1)->csCurAlloc)
  {
    return 1;
  }
  if ((*a1)->csCurAlloc > (*b1)->csCurAlloc)
  {
    return -1;
  }
  return 0;
}


void
memPi_print_xml_task_mem_in_use (FILE * fp)
{
  int i;
  char header_buf[2048];
  char temp_buf[2048];

  callsite_stats_t **av = NULL;
  int ac;
  callsite_src_id_cache_entry_t **cssv;
  int cssc;
  int cid;
  int found = 0;
  void *tpc;
  int tdx;


  h_gather_data (memPi.task_callsite_stats, &ac, (void ***) &av);

  qsort (av, ac, sizeof (void *), callsite_sort_by_size_in_use);

  snprintf (temp_buf, 2048, "Task Memory In Use (Task %d)", memPi.rank);
  print_xml_messsage_folder (fp, "task_in_use", temp_buf);

  memPi_open_executable (NULL);

  if (memPi.do_lookup)
  {
    for (i = 0; i < ac; i++)
    {
      if (av[i]->csCurAlloc > 0)
      {
        if (av[i]->functname[0] == NULL)
          cid = memPi_query_src (av[i]);
      }
    }
  }

  /* Closing the file opened by the BFD library apprears to result
     in a segfault if the file is reopened.
     memPi_close_executable ();
   */

  /* Get callsite source information that can be accessed by call site id */
  h_gather_data (callsite_src_id_cache, &cssc, (void ***) &cssv);
  qsort (cssv, cssc, sizeof (void *), callsite_src_id_cache_sort_by_id);

  for (i = 0; i < ac; i++)
  {
    //memPi_msg_debug ("Printing HWM info for task %d\n", tips[i]->rank);

    //  Create Header XML Entry
    if (av[i]->csCurAlloc > 0)
    {
/*	  snprintf (header_buf, 2048,
		    "Task %5d   in use %.0f   %s of %.0fB   Call site HWM %.0f in %s",
		    av[i]->rank,
		    av[i]->csCurAlloc, memP_func_name (0, av[i]->op),
		    memPi.global_task_info[memPi.rank].taskHWM, av[i]->csHWM,
		    cssv[av[i]->csid - 1]->functname[0]);
*/

      if (av[i]->lineno[0] == 0)
      {
        found = 0;
        tdx = 1;
        tpc = av[i]->pc[tdx];

        while ( tpc != NULL && found == 0)
        {
          if (av[i]->lineno[tdx] != 0)
          {
            snprintf (header_buf, 2048,
                "CSID %5d  Task %5d   in use %12.6f  of %12.6f MB at %s (%s:%u)",
                i, av[i]->rank, av[i]->csCurAlloc / MB,
                memPi.taskHWM / MB, av[i]->functname[tdx],
                av[i]->filename[tdx], av[i]->lineno[tdx]);
            found = 1;
          }

          tdx++;
          tpc = av[i]->pc[tdx];
        }

        if ( found == 0 )
        {
          snprintf (header_buf, 2048,
              "CSID %5d  Task %5d   in use %12.6f  of %12.6f MB at %p",
              i, av[i]->rank, av[i]->csCurAlloc / MB,
              memPi.taskHWM / MB, av[i]->pc[0]);
        }
      }
      else
      {
        snprintf (header_buf, 2048,
                  "CSID %5d  Task %5d   in use %12.6f  of %12.6f MB at %s (%s:%u)",
                  i, av[i]->rank, av[i]->csCurAlloc / MB,
                  memPi.taskHWM / MB, av[i]->functname[0],
                  av[i]->filename[0], av[i]->lineno[0]);
      }

      if ( av[i]->csid > 0 )
      {
        print_xml_messsage (fp, "task_in_use", header_buf, NULL,
            "Stack Trace", cssv[av[i]->csid - 1]);
      }
    }

#if 0
    //  Create Body XML Entry
    snprintf (body_buf, 2048,
              "High Water Mark (HWM) information for %s[%d]  (%s:%d)\n",
              av[cs->csid - 1]->functname[0], cs->csid,
              av[cs->csid - 1]->filename[0], av[cs->csid - 1]->line[0]);

    snprintf (temp_buf, 2048,
              "%*s%-*s%.0f\n", indent, "", body_width,
              "Task HWM at this callsite: ", tips[i]->taskHWM);
    strcat (body_buf, temp_buf);

    snprintf (temp_buf, 2048,
              "%*s%-*s%.0f (of %lld)\n", indent, "", body_width,
              "Task HWM reached at task allocation count ",
              tips[i]->taskHWMtaskCount, tips[i]->task_alloc_cnt);
    strcat (body_buf, temp_buf);

    snprintf (temp_buf, 2048,
              "%*s%-*s%.0f (of %lld)\n", indent, "", body_width,
              "Task HWM reached at this callsite's allocation count",
              tips[i]->taskHWMcsCount, cs->count);
    strcat (body_buf, temp_buf);

    //memPi_PRINTTIME (&memPi.startTime);
    //memPi_PRINTTIME (&tips[i]->taskHWMts);

    hwm_time =
      (memPi_GETTIMEDIFF (&tips[i]->taskHWMts, &memPi.startTime) / 1000000.0);

    snprintf (temp_buf, 2048,
              "%*s%-*s%.0f (of %.0f sec)\n", indent, "", body_width,
              "Task HWM reached at runtime ", hwm_time, tips[i]->app_time);
    strcat (body_buf, temp_buf);

    snprintf (temp_buf, 2048,
              "%*s%-*s%.0f\n", indent, "", body_width,
              "Callsite HWM (may not be at Task HWM): ", cs->csHWM);
    strcat (body_buf, temp_buf);

    snprintf (temp_buf, 2048,
              "%*s%-*s%.0f\n", indent, "", body_width,
              "Task Low Water Mark at this callsite: ", cs->csLWM);
    strcat (body_buf, temp_buf);

    bt = format_stack_trace (av[cs->csid - 1]);

    strcat (body_buf, bt);

    print_xml_messsage (fp, "task_hwm", header_buf, body_buf,
                        "Stack Trace", av[cs->csid - 1]);
#endif

  }

  // memPi_free (tips);
}

void
memPi_print_xml_job_details (FILE * fp)
{
  int i;
  char buf[2048];
  int dlw = 28;

  print_xml_messsage_folder (fp, "job_details", "Job Details");

  strcpy (buf, memPi.av[0]);
  for (i = 1; i < memPi.ac; i++)
  {
    strcat (buf, " ");
    strcat (buf, memPi.av[i]);
  }
  print_xml_message_fmt (fp, "job_details", "%-*s : %s", dlw, "Command", buf);

  {
    char nowstr[128];
    const struct tm *nowstruct;
    char *fmtstr = "%Y %m %d %T";

    nowstruct = localtime (&memPi.start_timeofday);
    if (strftime (nowstr, 128, fmtstr, nowstruct) == (size_t) 0)
      memPi_msg_warn ("Could not get string from strftime()\n");
    print_xml_message_fmt (fp, "job_details", "%-*s : %s", dlw,
                           "Job start time", nowstr);
    /*    print_intro_line (fp, "Start time", "%s", nowstr); */

    nowstruct = localtime (&memPi.stop_timeofday);
    if (strftime (nowstr, 128, fmtstr, nowstruct) == (size_t) 0)
      memPi_msg_warn ("Could not get string from strftime()\n");
    /*    print_intro_line (fp, "Stop time", "%s", nowstr); */
    print_xml_message_fmt (fp, "job_details", "%-*s : %s", dlw,
                           "Job stop time", nowstr);
  }

  print_xml_message_fmt (fp, "job_details", "%-*s : %s", dlw, "Tool name",
                         memPi.toolname);

  print_xml_message_fmt (fp, "job_details", "%-*s : %d.%d.%d", dlw, "Version",
                         memPi_vmajor, memPi_vminor, memPi_vpatch);
  print_xml_message_fmt (fp, "job_details", "%-*s : %s, %s", dlw,
                         "Build date", memPi_vdate, memPi_vtime);


#if 0
  print_intro_line (fp, "Version", "%d.%d.%d", memPi_vmajor, memPi_vminor,
                    memPi_vpatch);
  print_intro_line (fp, "MEMP Build date", "%s, %s", memPi_vdate,
                    memPi_vtime);
#endif


  print_xml_message_fmt (fp, "job_details", "%-*s : %s", dlw, "MEMP env var",
                         memPi.envStr == NULL ? "[null]" : memPi.envStr);

  print_xml_message_fmt (fp, "job_details", "%-*s : %s", dlw,
                         "Final Output Dir", memPi.outputDir);

  print_xml_message_fmt (fp, "job_details", "%-*s : %d", dlw,
                         "Heap Use at Init", memPi.heapUseAtInit);

  print_xml_message_fmt (fp, "job_details", "%-*s : %d", dlw,
                         "Unknown Heap Use at Finalize",
                         memPi.unknownHeapUse);

#if 0
  print_intro_line (fp, "Timer Used", "%s", memPi_TIMER_NAME);
  print_intro_line (fp, "MEMP env var", "%s",
                    memPi.envStr == NULL ? "[null]" : memPi.envStr);
  print_intro_line (fp, "Collector Rank", "%d", memPi.collectorRank);
  print_intro_line (fp, "Collector PID", "%d", memPi.procID);

  print_intro_line (fp, "Final Output Dir", "%s", memPi.outputDir);
  print_intro_line (fp, "Report generation", "%s",
                    memPi.collective_report ==
                    0 ? "Single collector task" : "Collective");
#endif
}


void
memPi_profile_print_task_xml_verbose ()
{
  FILE *fp = NULL;
  static int printCount = 0;
  char tb[1024];

  /* Generate output filename, and open */
  do
  {
    printCount++;
    snprintf (memPi.oFilename, 256, "%s/%s.rank-%d.%d.%d.%.0f.memP",
              memPi.outputDir, memPi.appName, memPi.rank, printCount,
              memPi.procID, memPi.taskHWM);
  }
  while (access (memPi.oFilename, F_OK) == 0);

  fp = fopen (memPi.oFilename, "w");
  snprintf (tb, 1024,
            "Task Memory In-Use Report : Rank %d, Index %d, HWM %.0f",
            memPi.rank, printCount, memPi.taskHWM);
  print_xml_report_header (fp, tb);
  memPi_print_xml_task_mem_in_use (fp);
  memPi_print_xml_job_details (fp);
  print_xml_report_footer (fp, tb);
  fclose (fp);
}


void
memPi_profile_print_xml_verbose (FILE * fp)
{
  if (memPi.collectorRank == memPi.rank)
  {
    print_xml_report_header (fp, memPi.oFilename);

    memPi_print_xml_task_hwm (fp);
    if (memPi.print_callsite_detail == 1)
      memPi_print_xml_callsites (fp);
    memPi_print_xml_job_details (fp);
    memPi_print_xml_task_assignment (fp);
    //memPi_print_xml_task_mem_in_use (fp);

    // memPi_print_verbose_task_info (fp);
  }

  if (memPi.collectorRank == memPi.rank)
  {
    print_xml_report_footer (fp, memPi.oFilename);
  }
}

#ifdef XMLTESTING

int
main (int argc, char **argv)
{
  FILE *fh;
  char *tf = "test.xml";
  callsite_src_id_cache_entry_t csid;
  memPi.stackDepth = 8;

  fh = fopen (tf, "w+");
  print_xml_report_header (fh, tf);
  print_xml_messsage_folder (fh, "test_folder", "Folder for Testing");

  csid.filename[0] = "memPi.c";
  csid.functname[0] = "alloc";
  csid.line[0] = 47;

  print_xml_messsage (fh, "test_folder", "message header", "message body",
                      "Annotation Title", &csid);
  print_xml_report_footer (fh);
  fclose (fh);
}
#endif

/* 

<license>

Copyright (c) 2010, Lawrence Livermore National Security, LLC.
Produced at the Lawrence Livermore National Laboratory
Written by Christopher Chambreau <chcham@llnl.gov>.
CODE-OCEC-10-010
All rights reserved.

This file is part of memP. For details, see http://memp.sourceforge.net .
Please also read this link – Additional BSD Notice.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  o  Redistributions of source code must retain the above copyright notice, 
     this list of conditions and the disclaimer below.

  o  Redistributions in binary form must reproduce the above copyright notice, 
     this list of conditions and the disclaimer (as noted below) in the 
     documentation and/or other materials provided with the distribution.

  o  Neither the name of the LLNS/LLNL nor the names of its contributors may be
     used to endorse or promote products derived from this software without 
     specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL LAWRENCE LIVERMORE NATIONAL
SECURITY, LLC, THE U.S. DEPARTMENT OF ENERGY OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.


Additional BSD Notice

1. This notice is required to be provided under our contract with the U.S.
Department of Energy (DOE). This work was produced at Lawrence Livermore
National Laboratory under Contract No. DE-AC52-07NA27344 with the DOE.

2. Neither the United States Government nor Lawrence Livermore National
Security, LLC nor any of their employees, makes any warranty, express or
implied, or assumes any liability or responsibility for the accuracy,
completeness, or usefulness of any information, apparatus, product, or process
disclosed, or represents that its use would not infringe privately-owned rights.

3. Also, reference herein to any specific commercial products, process, or
services by trade name, trademark, manufacturer or otherwise does not
necessarily constitute or imply its endorsement, recommendation, or favoring by
the United States Government or Lawrence Livermore National Security, LLC.
The views and opinions of authors expressed herein do not necessarily state or
reflect those of the United States Government or Lawrence Livermore National
Security, LLC, and shall not be used for advertising or product endorsement
purposes.

</license>

*/


/* eof */
