/* -*- C -*- 

   memP Heap Profiler ( http://memp.sourceforge.net/ )

   Please see COPYRIGHT AND LICENSE information at the end of this file.

   ----- 

   report.c -- reporting functions

   NOTE: Many of these functions are mpiP-specific and may not be used
   by memP at this time.

*/

#ifndef lint
static char *svnid = "$Id: report.c 383 2006-12-08 18:30:15Z chcham $";
#endif

#include <math.h>
#include <float.h>
#include <search.h>
#include <setjmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#include "memPi.h"


static void
print_section_heading (FILE * fp, char *str)
{
  int slen;
  int maxlen = 79;
  assert (fp);
  assert (str);
  assert (maxlen > 0);

  slen = 0;
  for (; slen < maxlen; slen++)
  {
    fputc ('-', fp);
  }
  fprintf (fp, "\n");
  slen = strlen (str);
  fprintf (fp, "@--- %s ", str);
  slen += 6;
  for (; slen < maxlen; slen++)
  {
    fputc ('-', fp);
  }
  fprintf (fp, "\n");
  slen = 0;
  for (; slen < maxlen; slen++)
  {
    fputc ('-', fp);
  }
  fprintf (fp, "\n");
}

int
callsite_src_id_cache_sort_by_id (const void *a, const void *b)
{
  callsite_src_id_cache_entry_t **a1 = (callsite_src_id_cache_entry_t **) a;
  callsite_src_id_cache_entry_t **b1 = (callsite_src_id_cache_entry_t **) b;

  if ((*a1)->id > (*b1)->id)
  {
    return 1;
  }
  if ((*a1)->id < (*b1)->id)
  {
    return -1;
  }
  return 0;
}

#include <stdarg.h>

/* proto -- print_intro_line(filep, name, "fmt to print value", args for fmt ) */
static void
print_intro_line (FILE * fp, char *name, char *fmt, ...)
{
  va_list args;
  char *cp = "@";

  va_start (args, fmt);

  fprintf (fp, "%-2s%-25s: ", cp, name);
  vfprintf (fp, fmt, args);
  fprintf (fp, "\n");

  va_end (args);
}


#if 0
double
calc_COV (double *data, int dataSize)
{
  int idx;
  double tot, avrg, var, std;

  tot = 0.0;
  avrg = 0.0;
  var = 0.0;
  std = 0.0;

  if (dataSize <= 1)
    return 0;

  for (idx = 0; idx < dataSize; idx++)
    tot += data[idx];

  avrg = tot / dataSize;

  for (idx = 0; idx < dataSize; idx++)
    var += (data[idx] - avrg) * (data[idx] - avrg);

  if (avrg > 0 && dataSize > 1)
  {
    var /= dataSize - 1;
    std = sqrt (var);
    return std / avrg;
  }
  else
    return 0;
}

void
print_alloc_stats (FILE * fp)
{
  int i;
  alloc_stats_t *gas;
  char *hfmt = "%6s %7s %12s %7s %11s %15s %10s\n";
  char *dfmt = "%6d %7lld %12.6g %7lld %11.6g %15lld %10.5g\n";
  char buf[128];
  size_t min_max_alloc = SIZE_MAX;
  size_t max_max_alloc = 0;
  size_t sum_max_alloc = 0;
  size_t med_max_alloc = 0;
  int min_max_alloc_rank = 0;
  int max_max_alloc_rank = 0;
  int sum_max_alloc_rank = 0;
  int med_max_alloc_rank = 0;
  char *peak_metric_label = "Peak Allocation";

#if 0
  if (memPi.collectorRank == memPi.rank)
  {
    fprintf (fp, "Total allocations    : %ld\n", memPi.app_alloc.alloc_cnt);
    fprintf (fp, "Total allocated size : %lld\n", memPi.app_alloc.alloc_tot);
    fprintf (fp, "Total frees          : %ld\n", memPi.app_alloc.free_cnt);
    fprintf (fp, "Total free size      : %lld\n", memPi.app_alloc.free_tot);
    fprintf (fp, "\n");
  }
#endif

  if (memPi.collectorRank == memPi.rank)
    gas = memPi_malloc (sizeof (alloc_stats_t) * memPi.size);

  PMPI_Gather (&memPi.app_alloc, sizeof (alloc_stats_t), MPI_BYTE, gas,
               sizeof (alloc_stats_t), MPI_BYTE, memPi.collectorRank,
               MPI_COMM_WORLD);


  if (memPi.collectorRank == memPi.rank)
  {
    sprintf (buf, "Task Allocation Statistics");
    print_section_heading (fp, buf);

    for (i = 0; i < memPi.size; i++)
    {
      if (gas[i].alloc_max < min_max_alloc)
      {
        min_max_alloc = gas[i].alloc_max;
        min_max_alloc_rank = i;
      }
      if (gas[i].alloc_max > max_max_alloc)
      {
        max_max_alloc = gas[i].alloc_max;
        max_max_alloc_rank = i;
      }
      sum_max_alloc += gas[i].alloc_max;
    }
    fprintf (fp, "  Least   %s : %18lld (Rank %d)\n", peak_metric_label,
             min_max_alloc, min_max_alloc_rank);
    fprintf (fp, "  Mean    %s : %18lld\n", peak_metric_label,
             sum_max_alloc / memPi.size);
    fprintf (fp, "  Largest %s : %18lld (Rank %d)\n", peak_metric_label,
             max_max_alloc, max_max_alloc_rank);
    fprintf (fp, "\n");
    /*  Calculate Median   */

#if 0
    size_t min_max_alloc = 0;
    size_t max_max_alloc = 0;
    size_t sum_max_alloc = 0;
    size_t med_max_alloc = 0;
    int min_max_alloc_rank = 0;
    int max_max_alloc_rank = 0;
    int sum_max_alloc_rank = 0;
    int med_max_alloc_rank = 0;
#endif

    fprintf (fp, hfmt, "Rank", "Alloc#", "Alloc Sum", "Free#", "Free Sum",
             "Alloc-Free", "Peak Alloc");

    for (i = 0; i < memPi.size; i++)
    {
      if (memPi.collectorRank == memPi.rank)
        fprintf (fp, dfmt,
                 i, gas[i].alloc_cnt, (double) gas[i].alloc_tot,
                 gas[i].free_cnt, (double) gas[i].free_tot,
                 gas[i].alloc_tot - gas[i].free_tot,
                 (double) gas[i].alloc_max);
    }

    if (memPi.collectorRank == memPi.rank)
      memPi_free (gas);
  }

}
#endif

void
memPi_print_report_header (FILE * fp)
{
  int i;

  fprintf (fp, "@ %s\n", memPi.toolname);
  fprintf (fp, "@ Command : ");
  for (i = 0; i < memPi.ac; i++)
  {
    fprintf (fp, "%s ", memPi.av[i]);
  }
  fprintf (fp, "\n");
  print_intro_line (fp, "Version", "%d.%d.%d", memPi_vmajor, memPi_vminor,
                    memPi_vpatch);
  print_intro_line (fp, "MEMP Build date", "%s, %s", memPi_vdate,
                    memPi_vtime);
  {
    char nowstr[128];
    const struct tm *nowstruct;
    char *fmtstr = "%Y %m %d %T";

    nowstruct = localtime (&memPi.start_timeofday);
    if (strftime (nowstr, 128, fmtstr, nowstruct) == (size_t) 0)
      memPi_msg_warn ("Could not get string from strftime()\n");
    print_intro_line (fp, "Start time", "%s", nowstr);

    nowstruct = localtime (&memPi.stop_timeofday);
    if (strftime (nowstr, 128, fmtstr, nowstruct) == (size_t) 0)
      memPi_msg_warn ("Could not get string from strftime()\n");
    print_intro_line (fp, "Stop time", "%s", nowstr);
  }

  print_intro_line (fp, "MEMP env var", "%s",
                    memPi.envStr == NULL ? "[null]" : memPi.envStr);
  print_intro_line (fp, "Collector Rank", "%d", memPi.collectorRank);
  print_intro_line (fp, "Collector PID", "%d", memPi.procID);
  print_intro_line (fp, "Job Size", "%d", memPi.size);

  print_intro_line (fp, "Final Output Dir", "%s", memPi.outputDir);
  print_intro_line (fp, "Report generation", "%s",
                    memPi.collective_report ==
                    0 ? "Single collector task" : "Collective");
  print_intro_line (fp, "bss  section size", "%u", memPi.bss);
  print_intro_line (fp, "text section size", "%u", memPi.text);
  print_intro_line (fp, "data section size", "%u", memPi.data);
  fprintf (fp, "\n");
}

#if 0
void
memPi_print_task_assignment (FILE * fp)
{
  int i;
  for (i = 0; i < memPi.size; i++)
  {
    print_intro_line (fp, "MPI Task Assignment", "%d %s",
                      memPi.global_task_info[i].rank,
                      memPi.global_task_info[i].hostname);
  }
}


void
memPi_print_verbose_task_info (FILE * fp)
{
  int i;
  /* -- calc the app runtime of each task */
  memPi.global_app_time = 0.0;

  for (i = 0; i < memPi.size; i++)
  {
    memPi_msg_debug ("app runtime for task %d is %g\n", i,
                     memPi.global_task_info[i].app_time);
    memPi.global_app_time += memPi.global_task_info[i].app_time;
    /*memPi.global_task_info[i].mpi_time = 0; */
  }


#if 0
  /* -- print the mpi/app times */
  print_section_heading (fp, "MPI Time (seconds)");
  fprintf (fp, "%-4s %10s %10s    %5s\n", "Task", "AppTime", "MPITime",
           "MPI%");
  for (i = 0; i < memPi.size; i++)
  {
    double ratio;

    if (memPi.global_task_info[i].app_time > 0)
    {
      ratio = (100.0 * memPi.global_task_info[i].mpi_time / 1e6) /
        memPi.global_task_info[i].app_time;
    }
    else
      ratio = 0;

    fprintf (fp,
             memP_Report_Formats[MEMP_MPI_TIME_FMT][memPi.reportFormat],
             i, memPi.global_task_info[i].app_time,
             memPi.global_task_info[i].mpi_time / 1e6, ratio);
  }
  fprintf (fp,
           memP_Report_Formats[MEMP_MPI_TIME_SUMMARY_FMT][memPi.reportFormat],
           memPi.global_app_time, memPi.global_mpi_time / 1e6,
           memPi.global_app_time >
           0 ? (100.0 * memPi.global_mpi_time / 1e6) /
           memPi.global_app_time : 0);
#endif
}

void
memPi_print_concise_task_info (FILE * fp)
{
  double min_app_time = DBL_MAX, min_mpi_time = DBL_MAX;
  double max_app_time = 0.0, max_mpi_time = 0.0;
  double tot_app_time = 0.0, tot_mpi_time = 0.0;
  double mean_app_time = 0.0, mean_mpi_time = 0.0;
  double var_app_time = 0.0, var_mpi_time = 0.0;

  int min_app_rank, min_mpi_rank, max_app_rank, max_mpi_rank;
  int colw = 10;
  int i;

  for (i = 0; i < memPi.size; i++)
  {
    if (memPi.global_task_info[i].app_time < min_app_time)
    {
      min_app_time = memPi.global_task_info[i].app_time;
      min_app_rank = i;
    }

    if (memPi.global_task_info[i].app_time > max_app_time)
    {
      max_app_time = memPi.global_task_info[i].app_time;
      max_app_rank = i;
    }

    tot_app_time += memPi.global_task_info[i].app_time;

    if (memPi.global_task_info[i].mpi_time < min_mpi_time)
    {
      min_mpi_time = memPi.global_task_info[i].mpi_time;
      min_mpi_rank = i;
    }

    if (memPi.global_task_info[i].mpi_time > max_mpi_time)
    {
      max_mpi_time = memPi.global_task_info[i].mpi_time;
      max_mpi_rank = i;
    }

    tot_mpi_time += memPi.global_task_info[i].mpi_time;
    memPi.global_app_time += memPi.global_task_info[i].app_time;
  }

  mean_app_time = tot_app_time / memPi.size;
  mean_mpi_time = tot_mpi_time / memPi.size;

  for (i = 0; i < memPi.size; i++)
  {
    var_app_time +=
      pow (mean_app_time - memPi.global_task_info[i].app_time, 2);
    var_mpi_time +=
      pow (mean_mpi_time - memPi.global_task_info[i].mpi_time, 2);
  }
  var_app_time /= memPi.size - 1;
  var_mpi_time /= memPi.size - 1;

  print_section_heading (fp, "Task Time Statistics (seconds)");
  fprintf (fp, "%*s %*s %*s %*s %*s %*s\n",
           colw, " ",
           colw, "AppTime",
           colw, "MPITime", colw, "MPI%", colw, "App Task", colw, "MPI Task");
  fprintf (fp, "%-*s %*f %*f %*s %*d %*d\n",
           colw, "Max",
           colw, max_app_time,
           colw, max_mpi_time / USECS,
           colw, "", colw, max_app_rank, colw, max_mpi_rank);
  fprintf (fp, "%-*s %*f %*f\n",
           colw, "Mean", colw, mean_app_time, colw, mean_mpi_time / USECS);
  fprintf (fp, "%-*s %*f %*f %*s %*d %*d\n",
           colw, "Min",
           colw, min_app_time,
           colw, min_mpi_time / USECS,
           colw, "", colw, min_app_rank, colw, min_mpi_rank);
  fprintf (fp, "%-*s %*f %*f\n",
           colw, "Stddev",
           colw, sqrt (var_app_time), colw, sqrt (var_mpi_time) / USECS);
  fprintf (fp, "%-*s %*f %*f %*.2f\n",
           colw, "Aggregate",
           colw, tot_app_time,
           colw, tot_mpi_time / USECS,
           colw, tot_mpi_time / USECS / tot_app_time * 100);
}


void
memPi_print_callsites (FILE * fp)
{
  int i, ac;
  char buf[256];
  callsite_src_id_cache_entry_t **av;
  int fileLenMax = 18;
  int funcLenMax = 24;
  int stack_continue_flag;
  char addr_buf[24];

  h_gather_data (callsite_src_id_cache, &ac, (void ***) &av);
  sprintf (buf, "Callsites: %d", ac);
  qsort (av, ac, sizeof (void *), callsite_src_id_cache_sort_by_id);
  print_section_heading (fp, buf);

  /* Find longest file and function names for formatting */
  for (i = 0; i < ac; i++)
  {
    int j, currlen;
    for (j = 0;
         (j < MEMP_CALLSITE_STACK_DEPTH) && (av[i]->filename[j] != NULL); j++)
    {
      currlen = strlen (av[i]->filename[j]);
      fileLenMax = currlen > fileLenMax ? currlen : fileLenMax;
      currlen = strlen (av[i]->functname[j]);
      funcLenMax = currlen > funcLenMax ? currlen : funcLenMax;
    }
  }

  fprintf (fp, "%3s %3s %-*s %5s %-*s %s\n",
           "ID", "Lev", fileLenMax, "File/Address", "Line", funcLenMax,
           "Parent_Funct", "Call");

  for (i = 0; i < ac; i++)
  {
    int j;
    for (j = 0, stack_continue_flag = 1;
         (j < MEMP_CALLSITE_STACK_DEPTH) && (av[i]->filename[j] != NULL) &&
         stack_continue_flag == 1; j++)
    {
      if (av[i]->line[j] == 0 &&
          (strcmp (av[i]->filename[j], "[unknown]") == 0 ||
           strcmp (av[i]->functname[j], "[unknown]") == 0))
      {
        fprintf (fp, "%3d %3d %-*s %-*s %s\n",
                 av[i]->id,
                 j,
                 fileLenMax + 6,
                 memP_format_address (av[i]->pc[j], addr_buf),
                 funcLenMax,
                 av[i]->functname[j], memP_func_name (j, av[i]->op));
      }
      else
      {
        fprintf (fp, "%3d %3d %-*s %5d %-*s %s\n",
                 av[i]->id,
                 j,
                 fileLenMax,
                 av[i]->filename[j], av[i]->line[j],
                 funcLenMax,
                 av[i]->functname[j], memP_func_name (j, av[i]->op));
      }
      /*  Do not bother printing stack frames above main   */
      if (strcmp (av[i]->functname[j], "main") == 0
          || strcmp (av[i]->functname[j], ".main") == 0
          || strcmp (av[i]->functname[j], "MAIN__") == 0)
        stack_continue_flag = 0;
    }
  }
  memPi_free (av);
}

void
memPi_print_top_time_sites (FILE * fp)
{
  int i, ac;
  callsite_stats_t **av;
  double timeCOV;
  int header_printed = 0;

  h_gather_data (memPi.global_callsite_stats_agg, &ac, (void ***) &av);

  /* -- now that we have all the statistics in a queue, which is
   * pointers to the data, we can sort it however we need to.
   */
  qsort (av, ac, sizeof (void *), callsite_sort_by_cumulative_time);

  for (i = 0; (i < 20) && (i < ac); i++)
  {
    if (av[i]->cumulativeTime > 0)
    {
      if (!header_printed)
      {
        print_section_heading (fp,
                               "Aggregate Time (top twenty, descending, milliseconds)");

        if (memPi.calcCOV)
        {
          fprintf (fp, "%-20s %4s %12s%6s  %6s  %6s\n", "Call",
                   "Site", "Time  ", "App%", "MPI%", "COV");
        }
        else
        {
          fprintf (fp, "%-20s %4s %12s%6s  %6s\n", "Call", "Site",
                   "Time  ", "App%", "MPI%");
        }

        header_printed = 1;
      }
      if (memPi.calcCOV)
      {
        timeCOV = calc_COV (av[i]->siteData, av[i]->siteDataIdx);

        fprintf (fp,
                 memP_Report_Formats[MEMP_AGGREGATE_COV_TIME_FMT]
                 [memPi.reportFormat], memP_func_name (0, av[i]->op),
                 av[i]->csid, av[i]->cumulativeTime / 1000.0,
                 100.0 * av[i]->cumulativeTime / (memPi.global_app_time *
                                                  1e6),
                 memPi.global_mpi_time >
                 0 ? (100.0 * av[i]->cumulativeTime /
                      memPi.global_mpi_time) : 0, timeCOV);
      }
      else
      {
        fprintf (fp,
                 memP_Report_Formats[MEMP_AGGREGATE_TIME_FMT]
                 [memPi.reportFormat], memP_func_name (0, av[i]->op),
                 av[i]->csid, av[i]->cumulativeTime / 1000.0,
                 memPi.global_app_time >
                 0 ? 100.0 * av[i]->cumulativeTime / (memPi.global_app_time *
                                                      1e6) : 0,
                 memPi.global_mpi_time >
                 0 ? 100.0 * av[i]->cumulativeTime /
                 memPi.global_mpi_time : 0);
      }
    }
  }

  memPi_free (av);
}

void
memPi_print_top_sent_sites (FILE * fp)
{
  int i, ac;
  callsite_stats_t **av;

  if (memPi.global_mpi_size > 0)
  {
    h_gather_data (memPi.global_callsite_stats_agg, &ac, (void ***) &av);

    /* -- now that we have all the statistics in a queue, which is
     * pointers to the data, we can sort it however we need to.
     */
    qsort (av, ac, sizeof (void *), callsite_sort_by_cumulative_size);

#if 0
    print_section_heading (fp,
                           "Aggregate Sent Message Size (top twenty, descending, bytes)");
#endif
    print_section_heading (fp,
                           "Aggregate Allocated Size (top twenty, descending, bytes)");
    fprintf (fp, "%-20s %4s %10s %10s %10s %6s\n", "Call", "Site",
             "Count", "Total", "Avrg", "Alloc%");

    for (i = 0; (i < 20) && (i < ac); i++)
    {
      if (av[i]->cumulativeAlloc > 0)
      {
        fprintf (fp,
                 memP_Report_Formats[MEMP_AGGREGATE_MESS_FMT]
                 [memPi.reportFormat], memP_func_name (0, av[i]->op),
                 av[i]->csid, av[i]->count, av[i]->cumulativeAlloc,
                 av[i]->cumulativeAlloc / av[i]->count,
                 av[i]->cumulativeAlloc * 100 / memPi.global_mpi_size);
      }
    }
    if (memPi.messageCountThreshold >= 0)
    {
      fprintf (fp,
               "\nTotal send/collective operation calls >= %d bytes : %lld of %lld operations\n",
               memPi.messageCountThreshold,
               memPi.global_mpi_msize_threshold_count,
               memPi.global_mpi_sent_count);
    }

    memPi_free (av);
  }
}

void
memPi_print_top_io_sites (FILE * fp)
{
  int i, ac;
  callsite_stats_t **av;

  /*  This function is only called by the collector, which is the only
   *  process that should have a valid global_mpi_io value.
   *  */
  if (memPi.global_mpi_io > 0)
  {
    h_gather_data (memPi.global_callsite_stats_agg, &ac, (void ***) &av);

    /* -- now that we have all the statistics in a queue, which is
     * pointers to the data, we can sort it however we need to.
     */
    qsort (av, ac, sizeof (void *), callsite_sort_by_cumulative_io);

    print_section_heading (fp,
                           "Aggregate Free Size (top twenty, descending, bytes)");
    fprintf (fp, "%-20s %4s %10s %10s %10s %6s\n", "Call", "Site",
             "Count", "Total", "Avrg", "Free%");

    for (i = 0; (i < 20) && (i < ac); i++)
    {
      if (av[i]->cumulativeFree > 0)
      {
        fprintf (fp,
                 memP_Report_Formats[MEMP_AGGREGATE_IO_FMT]
                 [memPi.reportFormat], memP_func_name (0, av[i]->op),
                 av[i]->csid, av[i]->count, av[i]->cumulativeFree,
                 av[i]->cumulativeFree / av[i]->count,
                 av[i]->cumulativeFree * 100 / memPi.global_mpi_io);
      }
    }

    memPi_free (av);
  }
}

void
memPi_print_all_callsite_time_info (FILE * fp)
{
  int i, ac;
  char buf[256];
  callsite_stats_t **av;

  h_gather_data (memPi.global_callsite_stats, &ac, (void ***) &av);

  /* -- now that we have all the statistics in a queue, which is
   * pointers to the data, we can sort it however we need to.
   */
  qsort (av, ac, sizeof (void *), callsite_sort_by_name_id_rank);

  sprintf (buf, "Callsite Time statistics (all, milliseconds): %d", ac);
  print_section_heading (fp, buf);
  fprintf (fp, "%-17s %4s %4s %6s %8s %8s %8s %6s %6s\n", "Name", "Site",
           "Rank", "Count", "Max", "Mean", "Min", "App%", "MPI%");

  {
    long long sCount = 0;
    double sMin = DBL_MAX;
    double sMax = 0;
    double sCumulative = 0;

    for (i = 0; i < ac; i++)
    {
      if (i != 0 && (av[i]->csid != av[i - 1]->csid))
      {
        if (sCumulative > 0)
        {
          fprintf (fp,
                   memP_Report_Formats[MEMP_CALLSITE_TIME_SUMMARY_FMT]
                   [memPi.reportFormat],
                   memP_func_name (0, av[i - 1]->op),
                   av[i - 1]->csid, "*", sCount, sMax / 1000.0,
                   sCumulative / (sCount * 1000.0), sMin / 1000.0,
                   memPi.global_app_time >
                   0 ? 100.0 * sCumulative / (memPi.global_app_time *
                                              1e6) : 0,
                   memPi.global_mpi_time >
                   0 ? 100.0 * sCumulative / memPi.global_mpi_time : 0);
          fprintf (fp, "\n");
        }
        sCount = 0;
        sMax = 0;
        sMin = DBL_MAX;
        sCumulative = 0;
      }

      sCount += av[i]->count;
      sCumulative += av[i]->cumulativeTime;
      sMax = max (av[i]->maxDur, sMax);
      sMin = min (av[i]->minDur, sMin);

      if (memPi.global_task_info[av[i]->rank].mpi_time != 0 &&
          (100.0 * av[i]->cumulativeTime /
           memPi.global_task_info[av[i]->rank].mpi_time)
          >= memPi.reportPrintThreshold)
      {
        fprintf (fp,
                 memP_Report_Formats[MEMP_CALLSITE_TIME_RANK_FMT]
                 [memPi.reportFormat], memP_func_name (0, av[i]->op),
                 av[i]->csid, av[i]->rank, av[i]->count,
                 av[i]->maxDur / 1000.0,
                 av[i]->cumulativeTime / (av[i]->count * 1000.0),
                 av[i]->minDur / 1000.0,
                 100.0 * av[i]->cumulativeTime /
                 (memPi.global_task_info[av[i]->rank].app_time * 1e6),
                 100.0 * av[i]->cumulativeTime /
                 memPi.global_task_info[av[i]->rank].mpi_time);
      }
    }
    if (sCumulative > 0)
    {
      fprintf (fp,
               memP_Report_Formats[MEMP_CALLSITE_TIME_SUMMARY_FMT]
               [memPi.reportFormat], memP_func_name (0, av[i]->op),
               av[i - 1]->csid, "*", sCount, sMax / 1000.0,
               sCumulative / (sCount * 1000.0), sMin / 1000.0,
               memPi.global_app_time >
               0 ? 100.0 * sCumulative / (memPi.global_app_time * 1e6) : 0,
               memPi.global_mpi_time >
               0 ? 100.0 * sCumulative / memPi.global_mpi_time : 0);
    }
  }

  memPi_free (av);
}

int
callsite_stats_sort_by_cumulative (memPi_callsite_summary_t * cs1,
                                   memPi_callsite_summary_t * cs2)
{
  if (cs1->cumulative > cs2->cumulative)
  {
    return -1;
  }
  if (cs1->cumulative < cs2->cumulative)
  {
    return 1;
  }
  return 0;
}

void
memPi_print_concise_callsite_time_info (FILE * fp)
{
  int i, ac, csidx = 0;
  char buf[256];
  callsite_stats_t **av;
  memPi_callsite_summary_t *callsite_stats = NULL;

  h_gather_data (memPi.global_callsite_stats, &ac, (void ***) &av);

  /* -- now that we have all the statistics in a queue, which is
   * pointers to the data, we can sort it however we need to.
   */
  qsort (av, ac, sizeof (void *), callsite_sort_by_name_id_rank);
  callsite_stats =
    (memPi_callsite_summary_t *)
    memPi_malloc (sizeof (memPi_callsite_summary_t) *
                  callsite_src_id_cache->size);

  if (callsite_stats == NULL)
  {
    memPi_msg_warn
      ("Failed to allocate space for callsite time summary reporting\n");
    memPi_free (av);
    return;
  }

  {
    long long sCount = 0;
    double sMin = DBL_MAX;
    double sMax = 0;
    double sCumulative = 0;
    int max_rnk, min_rnk;

    for (i = 0; i < ac; i++)
    {
      if (i != 0 && (av[i]->csid != av[i - 1]->csid))
      {
        callsite_stats[csidx].name = memP_func_name (0, av[i - 1]->op);
        callsite_stats[csidx].site = av[i - 1]->csid;
        callsite_stats[csidx].count = sCount;
        callsite_stats[csidx].max = sMax;
        callsite_stats[csidx].min = sMin;
        callsite_stats[csidx].cumulative = sCumulative;
        callsite_stats[csidx].max_rnk = max_rnk;
        callsite_stats[csidx].min_rnk = min_rnk;

        sCount = 0;
        sMax = 0;
        sMin = DBL_MAX;
        sCumulative = 0;
        csidx++;
      }

      sCount++;
      sCumulative += av[i]->cumulativeTime;
      if (av[i]->cumulativeTime > sMax)
      {
        sMax = av[i]->cumulativeTime;
        max_rnk = av[i]->rank;
      }
      if (av[i]->cumulativeTime < sMin)
      {
        sMin = av[i]->cumulativeTime;
        min_rnk = av[i]->rank;
      }
    }
    callsite_stats[csidx].name =
      memP_func_name (0, av[i - 1]->op),
      callsite_stats[csidx].site = av[i - 1]->csid;
    callsite_stats[csidx].count = sCount;
    callsite_stats[csidx].max = sMax;
    callsite_stats[csidx].min = sMin;
    callsite_stats[csidx].cumulative = sCumulative;
    callsite_stats[csidx].max_rnk = max_rnk;
    callsite_stats[csidx].min_rnk = min_rnk;

  }

  memPi_free (av);
  sprintf (buf, "Callsite Time statistics (all callsites, milliseconds): %d",
           csidx + 1);
  print_section_heading (fp, buf);
  fprintf (fp, "%-17s %4s %7s %9s %9s %9s %6s %6s\n", "Name", "Site",
           "Tasks", "Max", "Mean", "Min", "MaxRnk", "MinRnk");

  qsort (callsite_stats, csidx + 1, sizeof (memPi_callsite_summary_t),
         (int (*)(const void *, const void *))
         callsite_stats_sort_by_cumulative);
  for (i = 0; i <= csidx; i++)
  {
    fprintf (fp,
             memP_Report_Formats[MEMP_CALLSITE_TIME_CONCISE_FMT]
             [memPi.reportFormat], callsite_stats[i].name,
             callsite_stats[i].site, callsite_stats[i].count,
             callsite_stats[i].max / 1000.0,
             callsite_stats[i].cumulative / (callsite_stats[i].count * 1000),
             callsite_stats[i].min / 1000.0, callsite_stats[i].max_rnk,
             callsite_stats[i].min_rnk);
  }
  memPi_free (callsite_stats);
}

void
memPi_print_callsite_sent_info (FILE * fp)
{
  if (memPi.report_style == memPi_style_verbose)
    memPi_print_all_callsite_sent_info (fp);
  else if (memPi.report_style == memPi_style_concise)
    memPi_print_concise_callsite_sent_info (fp);
}


void
memPi_print_all_callsite_sent_info (FILE * fp)
  /*  Print Sent Data Section  */
{
  int i, ac;
  char buf[256];
  callsite_stats_t **av;

  if (memPi.global_mpi_size > 0)
  {
    h_gather_data (memPi.global_callsite_stats, &ac, (void ***) &av);

    /* -- now that we have all the statistics in a queue, which is
     * pointers to the data, we can sort it however we need to.
     */
    qsort (av, ac, sizeof (void *), callsite_sort_by_name_id_rank);

    sprintf (buf, "Callsite Allocation statistics (all, allocated bytes)");
    print_section_heading (fp, buf);
    fprintf (fp, "%-17s %4s %4s %7s %9s %9s %9s %9s\n", "Name", "Site",
             "Rank", "Count", "Max", "Mean", "Min", "Sum");

    {
      long long sCount = 0;
      double sMin = DBL_MAX;
      double sMax = 0;
      double sCumulative = 0;
      int lastcsid = 0;

      for (i = 0; i < ac; i++)
      {
        if (i != 0 && sCumulative > 0 && (av[i]->csid != av[i - 1]->csid))
        {
          fprintf (fp,
                   memP_Report_Formats[MEMP_CALLSITE_MESS_SUMMARY_FMT]
                   [memPi.reportFormat],
                   memP_func_name (0, av[i - 1]->op),
                   av[i - 1]->csid, "*", sCount, sMax,
                   sCumulative / sCount, sMin, sCumulative);

          sCount = 0;
          sMax = 0;
          sMin = DBL_MAX;
          sCumulative = 0;
        }

        if (av[i]->cumulativeAlloc > 0)
        {
          sCount += av[i]->count;
          sCumulative += av[i]->cumulativeAlloc;
          sMax = max (av[i]->maxAlloc, sMax);
          sMin = min (av[i]->minAlloc, sMin);

          if (lastcsid != 0 && lastcsid != av[i]->csid)
            fprintf (fp, "\n");

          fprintf (fp,
                   memP_Report_Formats[MEMP_CALLSITE_MESS_RANK_FMT]
                   [memPi.reportFormat],
                   memP_func_name (0, av[i]->op),
                   av[i]->csid, av[i]->rank, av[i]->count,
                   av[i]->maxAlloc,
                   av[i]->cumulativeAlloc / av[i]->count,
                   av[i]->minAlloc, av[i]->cumulativeAlloc);

          lastcsid = av[i]->csid;
        }
      }

      if (sCumulative > 0)
      {
        fprintf (fp,
                 memP_Report_Formats[MEMP_CALLSITE_MESS_SUMMARY_FMT]
                 [memPi.reportFormat],
                 memP_func_name (0, av[i - 1]->op),
                 av[i - 1]->csid, "*", sCount, sMax,
                 sCumulative / sCount, sMin, sCumulative);
      }
    }

    memPi_free (av);
  }
}


void
memPi_print_concise_callsite_sent_info (FILE * fp)
{
  int i, ac, csidx = 0;
  char buf[256];
  callsite_stats_t **av;
  memPi_callsite_summary_t *callsite_stats = NULL;

  h_gather_data (memPi.global_callsite_stats, &ac, (void ***) &av);

  /* -- now that we have all the statistics in a queue, which is
   * pointers to the data, we can sort it however we need to.
   */
  qsort (av, ac, sizeof (void *), callsite_sort_by_name_id_rank);
  callsite_stats =
    (memPi_callsite_summary_t *)
    memPi_malloc (sizeof (memPi_callsite_summary_t) *
                  callsite_src_id_cache->size);

  if (callsite_stats == NULL)
  {
    memPi_msg_warn
      ("Failed to allocate space for callsite volume summary reporting\n");
    memPi_free (av);
    return;
  }

  {
    long long sCount = 0;
    double sMin = DBL_MAX;
    double sMax = 0;
    double sCumulative = 0;
    int max_rnk = -1, min_rnk = -1;

    for (i = 0, csidx = 0; i < ac; i++)
    {
      if (i != 0 && (av[i]->csid != av[i - 1]->csid))
      {
        if (sCumulative > 0)
        {
          callsite_stats[csidx].name = memP_func_name (0, av[i - 1]->op);
          callsite_stats[csidx].site = av[i - 1]->csid;
          callsite_stats[csidx].count = sCount;
          callsite_stats[csidx].max = sMax;
          callsite_stats[csidx].min = sMin;
          callsite_stats[csidx].cumulative = sCumulative;
          callsite_stats[csidx].max_rnk = max_rnk;
          callsite_stats[csidx].min_rnk = min_rnk;
          csidx++;
        }

        sCount = 0;
        sMax = 0;
        sMin = DBL_MAX;
        sCumulative = 0;
        max_rnk = -1;
        min_rnk = -1;
      }

      sCount++;
      sCumulative += av[i]->cumulativeAlloc;

      if (av[i]->cumulativeAlloc > sMax)
      {
        sMax = av[i]->cumulativeAlloc;
        max_rnk = av[i]->rank;
      }
      if (av[i]->cumulativeAlloc < sMin)
      {
        sMin = av[i]->cumulativeAlloc;
        min_rnk = av[i]->rank;
      }
    }

    if (sCumulative > 0)
    {
      callsite_stats[csidx].name = memP_func_name (0, av[i - 1]->op);
      callsite_stats[csidx].site = av[i - 1]->csid;
      callsite_stats[csidx].count = sCount;
      callsite_stats[csidx].max = sMax;
      callsite_stats[csidx].min = sMin;
      callsite_stats[csidx].cumulative = sCumulative;
      callsite_stats[csidx].max_rnk = max_rnk;
      callsite_stats[csidx].min_rnk = min_rnk;
    }
    else
      csidx--;

  }

  memPi_free (av);
  if (csidx > 0)
  {
    sprintf (buf,
             "Callsite Allocation statistics (all callsites, allocated bytes): %d",
             csidx + 1);
    print_section_heading (fp, buf);
    fprintf (fp, "%-17s %4s %7s %9s %9s %9s %6s %6s\n", "Name", "Site",
             "Tasks", "Max", "Mean", "Min", "MaxRnk", "MinRnk");

    qsort (callsite_stats, csidx + 1, sizeof (memPi_callsite_summary_t),
           (int (*)(const void *, const void *))
           callsite_stats_sort_by_cumulative);
    for (i = 0; i <= csidx; i++)
    {
      fprintf (fp,
               memP_Report_Formats[MEMP_CALLSITE_MESS_CONCISE_FMT]
               [memPi.reportFormat], callsite_stats[i].name,
               callsite_stats[i].site, callsite_stats[i].count,
               callsite_stats[i].max,
               callsite_stats[i].cumulative / callsite_stats[i].count,
               callsite_stats[i].min, callsite_stats[i].max_rnk,
               callsite_stats[i].min_rnk);
    }
  }
  memPi_free (callsite_stats);
}

void
memPi_print_all_callsite_io_info (FILE * fp)
{
  int i, ac;
  char buf[256];
  callsite_stats_t **av;

  /*  This function is only called by the collector, which is the only
   *  process that should have a valid global_mpi_io value.
   *  */
  if (memPi.global_mpi_io > 0)
  {
    h_gather_data (memPi.global_callsite_stats, &ac, (void ***) &av);

    /* -- now that we have all the statistics in a queue, which is
     * pointers to the data, we can sort it however we need to.
     */
    qsort (av, ac, sizeof (void *), callsite_sort_by_name_id_rank);

    sprintf (buf, "Callsite Free statistics (all, bytes)");
    print_section_heading (fp, buf);
    fprintf (fp, "%-17s %4s %4s %7s %9s %9s %9s %9s\n", "Name", "Site",
             "Rank", "Count", "Max", "Mean", "Min", "Sum");

    {
      long long sCount = 0;
      double sMin = DBL_MAX;
      double sMax = 0;
      double sCumulative = 0;
      int lastcsid = 0;

      for (i = 0; i < ac; i++)
      {
        if (i != 0 && sCumulative > 0 && (av[i]->csid != av[i - 1]->csid))
        {
          fprintf (fp,
                   memP_Report_Formats[MEMP_CALLSITE_IO_SUMMARY_FMT]
                   [memPi.reportFormat],
                   memP_func_name (0, av[i - 1]->op),
                   av[i - 1]->csid, "*", sCount, sMax,
                   sCumulative / sCount, sMin, sCumulative);

          sCount = 0;
          sMax = 0;
          sMin = DBL_MAX;
          sCumulative = 0;
        }

        if (av[i]->cumulativeFree > 0)
        {
          sCount += av[i]->count;
          sCumulative += av[i]->cumulativeFree;
          sMax = max (av[i]->maxFree, sMax);
          sMin = min (av[i]->minFree, sMin);

          if (lastcsid != 0 && lastcsid != av[i]->csid)
            fprintf (fp, "\n");

          fprintf (fp,
                   memP_Report_Formats[MEMP_CALLSITE_IO_RANK_FMT]
                   [memPi.reportFormat],
                   memP_func_name (0, av[i]->op),
                   av[i]->csid, av[i]->rank, av[i]->count,
                   av[i]->maxFree, av[i]->cumulativeFree / av[i]->count,
                   av[i]->minFree, av[i]->cumulativeFree);

          lastcsid = av[i]->csid;
        }
      }

      if (sCumulative > 0)
      {
        fprintf (fp,
                 memP_Report_Formats[MEMP_CALLSITE_IO_SUMMARY_FMT]
                 [memPi.reportFormat],
                 memP_func_name (0, av[i - 1]->op),
                 av[i - 1]->csid, "*", sCount, sMax,
                 sCumulative / sCount, sMin, sCumulative);
      }
    }

    memPi_free (av);
  }
}


void
memPi_print_concise_callsite_io_info (FILE * fp)
{
  int i, ac, csidx = 0;
  char buf[256];
  callsite_stats_t **av;
  memPi_callsite_summary_t *callsite_stats = NULL;

  /*  This function is only called by the collector, which is the only
   *  process that should have a valid global_mpi_io value.
   *  */
  if (memPi.global_mpi_io > 0)
  {
    h_gather_data (memPi.global_callsite_stats, &ac, (void ***) &av);

    /* -- now that we have all the statistics in a queue, which is
     * pointers to the data, we can sort it however we need to.
     */
    qsort (av, ac, sizeof (void *), callsite_sort_by_name_id_rank);
    callsite_stats =
      (memPi_callsite_summary_t *)
      memPi_malloc (sizeof (memPi_callsite_summary_t) *
                    callsite_src_id_cache->size);

    if (callsite_stats == NULL)
    {
      memPi_msg_warn
        ("Failed to allocate space for callsite volume summary reporting\n");
      memPi_free (av);
      return;
    }

    {
      long long sCount = 0;
      double sMin = DBL_MAX;
      double sMax = 0;
      double sCumulative = 0;
      int max_rnk = -1, min_rnk = -1;

      for (i = 0, csidx = 0; i < ac; i++)
      {
        if (i != 0 && (av[i]->csid != av[i - 1]->csid))
        {
          if (sCumulative > 0)
          {
            callsite_stats[csidx].name = memP_func_name (0, av[i - 1]->op);
            callsite_stats[csidx].site = av[i - 1]->csid;
            callsite_stats[csidx].count = sCount;
            callsite_stats[csidx].max = sMax;
            callsite_stats[csidx].min = sMin;
            callsite_stats[csidx].cumulative = sCumulative;
            callsite_stats[csidx].max_rnk = max_rnk;
            callsite_stats[csidx].min_rnk = min_rnk;
            csidx++;
          }

          sCount = 0;
          sMax = 0;
          sMin = DBL_MAX;
          sCumulative = 0;
          max_rnk = -1;
          min_rnk = -1;
        }

        sCount++;
        sCumulative += av[i]->cumulativeFree;

        if (av[i]->cumulativeFree > sMax)
        {
          sMax = av[i]->cumulativeFree;
          max_rnk = av[i]->rank;
        }
        if (av[i]->cumulativeFree < sMin)
        {
          sMin = av[i]->cumulativeFree;
          min_rnk = av[i]->rank;
        }
      }

      if (sCumulative > 0)
      {
        callsite_stats[csidx].name = memP_func_name (0, av[i - 1]->op);
        callsite_stats[csidx].site = av[i - 1]->csid;
        callsite_stats[csidx].count = sCount;
        callsite_stats[csidx].max = sMax;
        callsite_stats[csidx].min = sMin;
        callsite_stats[csidx].cumulative = sCumulative;
        callsite_stats[csidx].max_rnk = max_rnk;
        callsite_stats[csidx].min_rnk = min_rnk;
      }
      else
        csidx--;

    }

    memPi_free (av);

    if (csidx > 0)
    {
      snprintf (buf, 256,
                "Callsite Free statistics (all callsites, bytes): %d",
                csidx + 1);
      print_section_heading (fp, buf);
      fprintf (fp, "%-17s %4s %7s %9s %9s %9s %6s %6s\n", "Name", "Site",
               "Tasks", "Max", "Mean", "Min", "MaxRnk", "MinRnk");

      qsort (callsite_stats, csidx + 1, sizeof (memPi_callsite_summary_t),
             (int (*)(const void *, const void *))
             callsite_stats_sort_by_cumulative);
      for (i = 0; i <= csidx; i++)
      {
        fprintf (fp,
                 memP_Report_Formats[MEMP_CALLSITE_MESS_CONCISE_FMT]
                 [memPi.reportFormat], callsite_stats[i].name,
                 callsite_stats[i].site, callsite_stats[i].count,
                 callsite_stats[i].max,
                 callsite_stats[i].cumulative / callsite_stats[i].count,
                 callsite_stats[i].min, callsite_stats[i].max_rnk,
                 callsite_stats[i].min_rnk);
      }
    }
    memPi_free (callsite_stats);
  }
}

void
memPi_coll_print_all_callsite_time_info (FILE * fp)
{
  int i, j, ac;
  char buf[256];
  callsite_stats_t **av;
  callsite_stats_t *task_stats, *task_lookup, *task_data;
  callsite_stats_t cs_buf;
  long long sCount = 0;
  double sMin = DBL_MAX;
  double sMax = 0;
  double sCumulative = 0;
  int malloc_check = 1;

#ifndef HAVE_MPI_IO
  return;
#endif

  /* Gather global callsite information at collectorRank and print header */
  if (memPi.rank == memPi.collectorRank)
  {
    h_gather_data (memPi.global_callsite_stats_agg, &ac, (void ***) &av);
    qsort (av, ac, sizeof (void *), callsite_sort_by_name_id_rank);

    task_data = memPi_malloc (sizeof (callsite_stats_t) * memPi.size);
    if (task_data == NULL)
    {
      memPi_msg_warn ("Failed to allocate space for task time data\n");
      malloc_check = 0;
      memPi_free (av);
    }
    else
    {
      sprintf (buf, "Callsite Time statistics (all, milliseconds): %lld",
               memPi.global_time_callsite_count);
      print_section_heading (fp, buf);
      fprintf (fp, "%-17s %4s %4s %6s %8s %8s %8s %6s %6s\n", "Name",
               "Site", "Rank", "Count", "Max", "Mean", "Min", "App%", "MPI%");
    }
  }

  /*  Check whether collector malloc succeeded.   */
  PMPI_Bcast (&malloc_check, 1, MPI_INT, memPi.collectorRank, memPi.comm);
  if (malloc_check == 0)
    return;

  PMPI_Bcast (&ac, 1, MPI_INT, memPi.collectorRank, memPi.comm);

  /* For each global callsite:
   *   Broadcast callsite info to remote tasks
   *   Lookup local callsite info in local hashes
   *   Gather all task info at collectorRank
   */
  for (i = 0; i < ac; i++)
  {
    if (memPi.rank == memPi.collectorRank)
      task_stats = av[i];
    else
      task_stats = &cs_buf;

    PMPI_Bcast (task_stats, sizeof (callsite_stats_t),
                MPI_CHAR, memPi.collectorRank, memPi.comm);

    task_stats->rank = memPi.rank;

    if (h_search
        (memPi.task_callsite_stats, task_stats,
         (void **) &task_lookup) == NULL)
    {
      task_lookup = &cs_buf;
      task_lookup->count = 0;
      task_lookup->cumulativeTime = 0;
      task_lookup->cumulativeTimeSquared = 0;
      task_lookup->maxDur = 0;
      task_lookup->minDur = 0;
      task_lookup->cumulativeAlloc = 0;
      task_lookup->cumulativeFree = 0;
      task_lookup->maxAlloc = 0;
      task_lookup->minAlloc = 0;
      task_lookup->maxFree = 0;
      task_lookup->minFree = 0;
      task_lookup->arbitraryMessageCount = 0;
      task_lookup->rank = memPi.rank;
    }

    PMPI_Gather (task_lookup, sizeof (callsite_stats_t),
                 MPI_CHAR, task_data,
                 sizeof (callsite_stats_t), MPI_CHAR,
                 memPi.collectorRank, memPi.comm);

    if (memPi.rank == memPi.collectorRank)
    {
      sCount = 0;
      sMax = 0;
      sMin = DBL_MAX;
      sCumulative = 0;

      for (j = 0; j < memPi.size; j++)
      {
        sCount += task_data[j].count;
        sCumulative += task_data[j].cumulativeTime;
        sMax = max (task_data[j].maxDur, sMax);
        sMin = min (task_data[j].minDur, sMin);

        if (task_data[j].count > 0 &&
            (100.0 * task_data[j].cumulativeTime /
             memPi.global_task_info[task_data[j].rank].mpi_time)
            >= memPi.reportPrintThreshold)
        {
          fprintf (fp,
                   memP_Report_Formats[MEMP_CALLSITE_TIME_RANK_FMT]
                   [memPi.reportFormat],
                   memP_func_name (0, task_stats->op),
                   av[i]->csid, task_data[j].rank,
                   task_data[j].count, task_data[j].maxDur / 1000.0,
                   task_data[j].cumulativeTime / (task_data[j].count *
                                                  1000.0),
                   task_data[j].minDur / 1000.0,
                   100.0 * task_data[j].cumulativeTime /
                   (memPi.global_task_info[task_data[j].rank].app_time * 1e6),
                   100.0 * task_data[j].cumulativeTime /
                   memPi.global_task_info[task_data[j].rank].mpi_time);
        }
      }
      if (sCount > 0)
      {
        fprintf (fp,
                 memP_Report_Formats[MEMP_CALLSITE_TIME_SUMMARY_FMT]
                 [memPi.reportFormat],
                 memP_func_name (0, task_data[j - 1].op),
                 av[i]->csid, "*", sCount, sMax / 1000.0,
                 sCumulative / (sCount * 1000.0), sMin / 1000.0,
                 memPi.global_app_time >
                 0 ? 100.0 * sCumulative / (memPi.global_app_time *
                                            1e6) : 0,
                 memPi.global_mpi_time >
                 0 ? 100.0 * sCumulative / memPi.global_mpi_time : 0);
        fprintf (fp, "\n");
      }
    }
  }


  if (memPi.rank == memPi.collectorRank)
  {
    memPi_free (av);
    memPi_free (task_data);
  }
}


void
memPi_coll_print_concise_callsite_time_info (FILE * fp)
{
  int i, ac;
  char buf[256];
  callsite_stats_t **av;
  callsite_stats_t *task_stats, *task_lookup;
  callsite_stats_t cs_buf;
  double tot_time;
  long long task_flag, tot_tasks;
  struct
  {
    double val;
    int rank;
  } min_time, max_time, local_min_time, local_max_time;

  if (memPi.rank == memPi.collectorRank)
  {
    /*  Sort aggregate callsite stats by descending cumulative time   */
    h_gather_data (memPi.global_callsite_stats_agg, &ac, (void ***) &av);
    qsort (av, ac, sizeof (void *), callsite_sort_by_cumulative_time);

    snprintf (buf, 256,
              "Callsite Time statistics (all callsites, milliseconds): %d",
              ac);
    print_section_heading (fp, buf);
    fprintf (fp, "%-17s %4s %7s %9s %9s %9s %6s %6s\n", "Name", "Site",
             "Tasks", "Max", "Mean", "Min", "MaxRnk", "MinRnk");
  }

  PMPI_Bcast (&ac, 1, MPI_INT, memPi.collectorRank, memPi.comm);

  /* For each global callsite:
   *   Broadcast callsite info to remote tasks
   *   Lookup local callsite info in local hashes
   *   Use Reduce to get task statistics at collectorRank
   */
  for (i = 0; i < ac; i++)
  {
    if (memPi.rank == memPi.collectorRank)
      task_stats = av[i];
    else
      task_stats = &cs_buf;

    /*  Broadcast current call site to all tasks   */
    PMPI_Bcast (task_stats, sizeof (callsite_stats_t),
                MPI_CHAR, memPi.collectorRank, memPi.comm);


    /*  Search for task local entry for the current call site   */
    task_stats->rank = memPi.rank;
    if (h_search
        (memPi.task_callsite_stats, task_stats,
         (void **) &task_lookup) == NULL)
    {
      task_lookup = &cs_buf;
      task_lookup->count = 0;
      task_lookup->cumulativeTime = 0;
      task_lookup->cumulativeTimeSquared = 0;
      task_lookup->maxDur = 0;
      task_lookup->minDur = DBL_MAX;
      task_lookup->cumulativeAlloc = 0;
      task_lookup->cumulativeFree = 0;
      task_lookup->maxAlloc = 0;
      task_lookup->minAlloc = DBL_MAX;
      task_lookup->maxFree = 0;
      task_lookup->minFree = DBL_MAX;
      task_lookup->arbitraryMessageCount = 0;
      task_lookup->rank = memPi.rank;
    }
    tot_tasks = 0;
    task_flag = task_lookup->count > 0 ? 1 : 0;

    /*  Get minimum aggregate time and rank for this call site   */
    if (task_lookup->cumulativeTime > 0)
      local_min_time.val = task_lookup->cumulativeTime;
    else
      local_min_time.val = DBL_MAX;
    local_min_time.rank = memPi.rank;
    PMPI_Reduce (&local_min_time, &min_time, 1, MPI_DOUBLE_INT, MPI_MINLOC,
                 memPi.collectorRank, memPi.comm);

    /*  Get maximum aggregate time and rank for this call site   */
    local_max_time.val = task_lookup->cumulativeTime;
    local_max_time.rank = memPi.rank;
    PMPI_Reduce (&local_max_time, &max_time, 1, MPI_DOUBLE_INT, MPI_MAXLOC,
                 memPi.collectorRank, memPi.comm);


    /*  Get sum of aggregate time for all tasks for this call site   */
    PMPI_Reduce (&(task_lookup->cumulativeTime), &tot_time, 1, MPI_DOUBLE,
                 MPI_SUM, memPi.collectorRank, memPi.comm);


    /*  Get the number of tasks with non-zero values for this call site   */
    PMPI_Reduce (&task_flag, &tot_tasks, 1, MPI_LONG_LONG, MPI_SUM,
                 memPi.collectorRank, memPi.comm);

    /*  Print summary statistics for this call site  */
    if (memPi.rank == memPi.collectorRank)
    {
      fprintf (fp,
               memP_Report_Formats[MEMP_CALLSITE_TIME_CONCISE_FMT]
               [memPi.reportFormat], memP_func_name (0, av[i]->op),
               av[i]->csid, tot_tasks, max_time.val / 1000.0,
               tot_time / (tot_tasks * 1000), min_time.val / 1000.0,
               max_time.rank, min_time.rank);
    }
  }

  if (memPi.rank == memPi.collectorRank)
    memPi_free (av);
}


void
memPi_coll_print_concise_callsite_sent_info (FILE * fp)
{
  int ci, i, ac, callsite_count;
  char buf[256];
  callsite_stats_t **av;
  callsite_stats_t *task_stats, *task_lookup;
  callsite_stats_t cs_buf;
  double tot_sent;
  long long task_flag, tot_tasks;
  struct
  {
    double val;
    int rank;
  } min_sent, max_sent, local_min_sent, local_max_sent;

  if (memPi.rank == memPi.collectorRank)
  {
    /*  Sort aggregate callsite stats by descending cumulative sent   */
    h_gather_data (memPi.global_callsite_stats_agg, &ac, (void ***) &av);
    qsort (av, ac, sizeof (void *), callsite_sort_by_cumulative_size);
    for (i = 0, callsite_count = 0; i < ac; i++)
    {
      if (av[i]->cumulativeAlloc > 0)
        callsite_count++;
    }
    if (callsite_count > 0)
    {

      snprintf (buf, 256,
                "Callsite Allocation statistics (all callsites, bytes sent): %d",
                callsite_count);
      print_section_heading (fp, buf);
      fprintf (fp, "%-17s %4s %7s %9s %9s %9s %6s %6s\n", "Name", "Site",
               "Tasks", "Max", "Mean", "Min", "MaxRnk", "MinRnk");
    }
  }

  PMPI_Bcast (&callsite_count, 1, MPI_INT, memPi.collectorRank, memPi.comm);

  /* For each global callsite with cumulativeAlloc > 0:
   *   Broadcast callsite info to remote tasks
   *   Lookup local callsite info in local hashes
   *   Use Reduce to get task statistics at collectorRank
   */
  for (i = 0, ci = 0; i < callsite_count; i++, ci++)
  {
    if (memPi.rank == memPi.collectorRank)
    {
      task_stats = av[ci];
      while (task_stats->cumulativeAlloc == 0)
      {
        ci++;
        task_stats = av[ci];
      }
    }
    else
      task_stats = &cs_buf;

    /*  Broadcast current call site to all tasks   */
    PMPI_Bcast (task_stats, sizeof (callsite_stats_t),
                MPI_CHAR, memPi.collectorRank, memPi.comm);


    /*  Search for task local entry for the current call site   */
    task_stats->rank = memPi.rank;
    if (h_search
        (memPi.task_callsite_stats, task_stats,
         (void **) &task_lookup) == NULL)
    {
      task_lookup = &cs_buf;
      task_lookup->count = 0;
      task_lookup->cumulativeTime = 0;
      task_lookup->cumulativeTimeSquared = 0;
      task_lookup->maxDur = 0;
      task_lookup->minDur = DBL_MAX;
      task_lookup->cumulativeAlloc = 0;
      task_lookup->cumulativeFree = 0;
      task_lookup->maxAlloc = 0;
      task_lookup->minAlloc = DBL_MAX;
      task_lookup->maxFree = 0;
      task_lookup->minFree = DBL_MAX;
      task_lookup->arbitraryMessageCount = 0;
      task_lookup->rank = memPi.rank;
    }
    tot_tasks = 0;
    task_flag = task_lookup->cumulativeAlloc > 0 ? 1 : 0;

    /*  Get minimum aggregate sent and rank for this call site   */
    if (task_lookup->cumulativeAlloc > 0)
      local_min_sent.val = task_lookup->cumulativeAlloc;
    else
      local_min_sent.val = DBL_MAX;
    local_min_sent.rank = memPi.rank;
    PMPI_Reduce (&local_min_sent, &min_sent, 1, MPI_DOUBLE_INT, MPI_MINLOC,
                 memPi.collectorRank, memPi.comm);

    /*  Get maximum aggregate sent and rank for this call site   */
    local_max_sent.val = task_lookup->cumulativeAlloc;
    local_max_sent.rank = memPi.rank;
    PMPI_Reduce (&local_max_sent, &max_sent, 1, MPI_DOUBLE_INT, MPI_MAXLOC,
                 memPi.collectorRank, memPi.comm);


    /*  Get sum of aggregate sent for all tasks for this call site   */
    PMPI_Reduce (&(task_lookup->cumulativeAlloc), &tot_sent, 1,
                 MPI_DOUBLE, MPI_SUM, memPi.collectorRank, memPi.comm);


    /*  Get the number of tasks with non-zero values for this call site   */
    PMPI_Reduce (&task_flag, &tot_tasks, 1, MPI_LONG_LONG, MPI_SUM,
                 memPi.collectorRank, memPi.comm);

    /*  Print summary statistics for this call site  */
    if (memPi.rank == memPi.collectorRank)
    {
      fprintf (fp,
               memP_Report_Formats[MEMP_CALLSITE_MESS_CONCISE_FMT]
               [memPi.reportFormat], memP_func_name (0, av[ci]->op),
               av[ci]->csid, tot_tasks, max_sent.val, tot_sent / tot_tasks,
               min_sent.val, max_sent.rank, min_sent.rank);
    }
  }

  if (memPi.rank == memPi.collectorRank)
    memPi_free (av);
}


void
memPi_coll_print_concise_callsite_io_info (FILE * fp)
{
  int ci, i, ac, callsite_count;
  char buf[256];
  callsite_stats_t **av;
  callsite_stats_t *task_stats, *task_lookup;
  callsite_stats_t cs_buf;
  double tot_io;
  long long task_flag, tot_tasks;
  struct
  {
    double val;
    int rank;
  } min_io, max_io, local_min_io, local_max_io;

#ifndef HAVE_MPI_IO
  return;
#endif

  if (memPi.rank == memPi.collectorRank)
  {
    /*  Sort aggregate callsite stats by descending cumulative io   */
    h_gather_data (memPi.global_callsite_stats_agg, &ac, (void ***) &av);
    qsort (av, ac, sizeof (void *), callsite_sort_by_cumulative_io);
    for (i = 0, callsite_count = 0; i < ac; i++)
    {
      if (av[i]->cumulativeFree > 0)
        callsite_count++;
    }
    if (callsite_count > 0)
    {

      snprintf (buf, 256,
                "Callsite Free statistics (all callsites, bytes): %d",
                callsite_count);
      print_section_heading (fp, buf);
      fprintf (fp, "%-17s %4s %7s %9s %9s %9s %6s %6s\n", "Name", "Site",
               "Tasks", "Max", "Mean", "Min", "MaxRnk", "MinRnk");
    }
  }

  PMPI_Bcast (&callsite_count, 1, MPI_INT, memPi.collectorRank, memPi.comm);

  /* For each global callsite with cumulativeFree > 0:
   *   Broadcast callsite info to remote tasks
   *   Lookup local callsite info in local hashes
   *   Use Reduce to get task statistics at collectorRank
   */
  for (i = 0, ci = 0; i < callsite_count; i++, ci++)
  {
    if (memPi.rank == memPi.collectorRank)
    {
      /*  Find next call site with IO activity   */
      task_stats = av[ci];
      while (task_stats->cumulativeFree == 0)
      {
        ci++;
        task_stats = av[ci];
      }
    }
    else
      task_stats = &cs_buf;

    /*  Broadcast current call site to all tasks   */
    PMPI_Bcast (task_stats, sizeof (callsite_stats_t),
                MPI_CHAR, memPi.collectorRank, memPi.comm);


    /*  Search for task local entry for the current call site   */
    task_stats->rank = memPi.rank;
    if (h_search
        (memPi.task_callsite_stats, task_stats,
         (void **) &task_lookup) == NULL)
    {
      task_lookup = &cs_buf;
      task_lookup->count = 0;
      task_lookup->cumulativeTime = 0;
      task_lookup->cumulativeTimeSquared = 0;
      task_lookup->maxDur = 0;
      task_lookup->minDur = DBL_MAX;
      task_lookup->cumulativeAlloc = 0;
      task_lookup->cumulativeFree = 0;
      task_lookup->maxAlloc = 0;
      task_lookup->minAlloc = DBL_MAX;
      task_lookup->maxFree = 0;
      task_lookup->minFree = DBL_MAX;
      task_lookup->arbitraryMessageCount = 0;
      task_lookup->rank = memPi.rank;
    }
    tot_tasks = 0;
    task_flag = task_lookup->cumulativeFree > 0 ? 1 : 0;

    /*  Get minimum aggregate io and rank for this call site   */
    if (task_lookup->cumulativeFree > 0)
      local_min_io.val = task_lookup->cumulativeFree;
    else
      local_min_io.val = DBL_MAX;
    local_min_io.rank = memPi.rank;
    PMPI_Reduce (&local_min_io, &min_io, 1, MPI_DOUBLE_INT, MPI_MINLOC,
                 memPi.collectorRank, memPi.comm);

    /*  Get maximum aggregate io and rank for this call site   */
    local_max_io.val = task_lookup->cumulativeFree;
    local_max_io.rank = memPi.rank;
    PMPI_Reduce (&local_max_io, &max_io, 1, MPI_DOUBLE_INT, MPI_MAXLOC,
                 memPi.collectorRank, memPi.comm);


    /*  Get sum of aggregate io for all tasks for this call site   */
    PMPI_Reduce (&(task_lookup->cumulativeFree), &tot_io, 1, MPI_DOUBLE,
                 MPI_SUM, memPi.collectorRank, memPi.comm);


    /*  Get the number of tasks with non-zero values for this call site   */
    PMPI_Reduce (&task_flag, &tot_tasks, 1, MPI_LONG_LONG, MPI_SUM,
                 memPi.collectorRank, memPi.comm);

    /*  Print summary statistics for this call site  */
    if (memPi.rank == memPi.collectorRank)
    {
      fprintf (fp,
               memP_Report_Formats[MEMP_CALLSITE_MESS_CONCISE_FMT]
               [memPi.reportFormat], memP_func_name (0, av[ci]->op),
               av[ci]->csid, tot_tasks, max_io.val, tot_io / tot_tasks,
               min_io.val, max_io.rank, min_io.rank);
    }
  }

  if (memPi.rank == memPi.collectorRank)
    memPi_free (av);
}


void
memPi_coll_print_all_callsite_sent_info (FILE * fp)
{
  int i, j, ac;
  char buf[256];
  callsite_stats_t **av;
  callsite_stats_t *task_stats, *task_lookup, *task_data;
  callsite_stats_t cs_buf;
  long long sCount = 0;
  double sMin = DBL_MAX;
  double sMax = 0;
  double sCumulative = 0;
  double tot_data_sent = 0;
  int malloc_check = 1;

  PMPI_Bcast (&memPi.global_mpi_sent_count, 1, MPI_LONG_LONG,
              memPi.collectorRank, memPi.comm);
  if (memPi.global_mpi_sent_count > 0)
  {
    /* Gather global callsite information at collectorRank and print header */
    if (memPi.rank == memPi.collectorRank)
    {
      h_gather_data (memPi.global_callsite_stats_agg, &ac, (void ***) &av);
      qsort (av, ac, sizeof (void *), callsite_sort_by_name_id_rank);
      task_data = memPi_malloc (sizeof (callsite_stats_t) * memPi.size);

      if (task_data == NULL)
      {
        memPi_msg_warn ("Failed to allocate space for task volume data\n");
        malloc_check = 0;
        memPi_free (av);
      }
      else
      {
        sprintf (buf,
                 "Callsite Allocation statistics (all, allocated bytes)");
        print_section_heading (fp, buf);
        fprintf (fp, "%-17s %4s %4s %7s %9s %9s %9s %9s\n", "Name",
                 "Site", "Rank", "Count", "Max", "Mean", "Min", "Sum");
      }
    }

    /*  Check whether collector malloc succeeded.   */
    PMPI_Bcast (&malloc_check, 1, MPI_INT, memPi.collectorRank, memPi.comm);
    if (malloc_check == 0)
      return;


    PMPI_Bcast (&ac, 1, MPI_INT, memPi.collectorRank, memPi.comm);

    /* For each global callsite:
     *   Broadcast callsite info to remote tasks
     *   Lookup local callsite info in local hashes
     *   Gather all task info at collectorRank
     */
    for (i = 0; i < ac; i++)
    {
      if (memPi.rank == memPi.collectorRank)
        task_stats = av[i];
      else
        task_stats = &cs_buf;

      tot_data_sent = task_stats->cumulativeAlloc;
      PMPI_Bcast (&tot_data_sent, 1, MPI_DOUBLE, memPi.collectorRank,
                  memPi.comm);

      if (tot_data_sent > 0)
      {
        PMPI_Bcast (task_stats, sizeof (callsite_stats_t),
                    MPI_CHAR, memPi.collectorRank, memPi.comm);

        task_stats->rank = memPi.rank;

        if (h_search
            (memPi.task_callsite_stats, task_stats,
             (void **) &task_lookup) == NULL)
        {
          task_lookup = &cs_buf;
          task_lookup->count = 0;
          task_lookup->cumulativeTime = 0;
          task_lookup->cumulativeTimeSquared = 0;
          task_lookup->maxDur = 0;
          task_lookup->minDur = 0;
          task_lookup->cumulativeAlloc = 0;
          task_lookup->cumulativeFree = 0;
          task_lookup->maxAlloc = 0;
          task_lookup->minAlloc = 0;
          task_lookup->maxFree = 0;
          task_lookup->minFree = 0;
          task_lookup->arbitraryMessageCount = 0;
          task_lookup->op = 0;
        }

        PMPI_Gather (task_lookup, sizeof (callsite_stats_t),
                     MPI_CHAR, task_data,
                     sizeof (callsite_stats_t), MPI_CHAR,
                     memPi.collectorRank, memPi.comm);

        if (memPi.rank == memPi.collectorRank)
        {
          sCount = 0;
          sMax = 0;
          sMin = DBL_MAX;
          sCumulative = 0;

          for (j = 0; j < memPi.size; j++)
          {
            if (task_data[j].cumulativeAlloc > 0)
            {
              sCount += task_data[j].count;
              sCumulative += task_data[j].cumulativeAlloc;
              sMax = max (task_data[j].maxAlloc, sMax);
              sMin = min (task_data[j].minAlloc, sMin);

              fprintf (fp,
                       memP_Report_Formats
                       [MEMP_CALLSITE_MESS_RANK_FMT][memPi.reportFormat],
                       memP_func_name (0, av[i]->op),
                       av[i]->csid, task_data[j].rank,
                       task_data[j].count,
                       task_data[j].maxAlloc,
                       task_data[j].cumulativeAlloc /
                       task_data[j].count,
                       task_data[j].minAlloc, task_data[j].cumulativeAlloc);
            }

          }
          if (sCumulative > 0)
          {
            fprintf (fp,
                     memP_Report_Formats
                     [MEMP_CALLSITE_MESS_SUMMARY_FMT][memPi.reportFormat],
                     memP_func_name (0, av[i]->op),
                     av[i]->csid,
                     "*", sCount, sMax, sCumulative / sCount, sMin,
                     sCumulative);
          }
          fprintf (fp, "\n");
        }
      }
    }

    if (memPi.rank == memPi.collectorRank)
    {
      memPi_free (av);
      memPi_free (task_data);
    }
  }
}

void
memPi_coll_print_all_callsite_io_info (FILE * fp)
{
  int i, j, ac;
  char buf[256];
  callsite_stats_t **av;
  callsite_stats_t *task_stats, *task_lookup, *task_data;
  callsite_stats_t cs_buf;
  long long sCount = 0;
  double sMin = DBL_MAX;
  double sMax = 0;
  double sCumulative = 0;
  double tot_data_sent = 0;
  int malloc_check = 1;

#ifndef HAVE_MPI_IO
  return;
#endif

  PMPI_Bcast (&memPi.global_mpi_io, 1, MPI_DOUBLE, memPi.collectorRank,
              memPi.comm);

  if (memPi.global_mpi_io > 0)
  {
    /* Gather global callsite information at collectorRank and print header */
    if (memPi.rank == memPi.collectorRank)
    {
      h_gather_data (memPi.global_callsite_stats_agg, &ac, (void ***) &av);
      qsort (av, ac, sizeof (void *), callsite_sort_by_name_id_rank);
      task_data = memPi_malloc (sizeof (callsite_stats_t) * memPi.size);

      if (task_data == NULL)
      {
        memPi_msg_warn ("Failed to allocate space for task free data\n");
        malloc_check = 0;
        memPi_free (av);
      }
      else
      {
        sprintf (buf, "Callsite Free statistics (all, bytes)");
        print_section_heading (fp, buf);
        fprintf (fp, "%-17s %4s %4s %7s %9s %9s %9s %9s\n", "Name",
                 "Site", "Rank", "Count", "Max", "Mean", "Min", "Sum");
      }
    }

    /*  Check whether collector malloc succeeded.   */
    PMPI_Bcast (&malloc_check, 1, MPI_INT, memPi.collectorRank, memPi.comm);
    if (malloc_check == 0)
      return;

    PMPI_Bcast (&ac, 1, MPI_INT, memPi.collectorRank, memPi.comm);

    /* For each global callsite:
     *   Broadcast callsite info to remote tasks
     *   Lookup local callsite info in local hashes
     *   Gather all task info at collectorRank
     */
    for (i = 0; i < ac; i++)
    {
      if (memPi.rank == memPi.collectorRank)
        task_stats = av[i];
      else
        task_stats = &cs_buf;

      tot_data_sent = task_stats->cumulativeFree;
      PMPI_Bcast (&tot_data_sent, 1, MPI_DOUBLE, memPi.collectorRank,
                  memPi.comm);

      if (tot_data_sent > 0)
      {
        PMPI_Bcast (task_stats, sizeof (callsite_stats_t),
                    MPI_CHAR, memPi.collectorRank, memPi.comm);

        task_stats->rank = memPi.rank;

        if (h_search
            (memPi.task_callsite_stats, task_stats,
             (void **) &task_lookup) == NULL)
        {
          task_lookup = &cs_buf;
          task_lookup->count = 0;
          task_lookup->cumulativeTime = 0;
          task_lookup->cumulativeTimeSquared = 0;
          task_lookup->maxDur = 0;
          task_lookup->minDur = 0;
          task_lookup->cumulativeAlloc = 0;
          task_lookup->cumulativeFree = 0;
          task_lookup->maxAlloc = 0;
          task_lookup->minAlloc = 0;
          task_lookup->maxFree = 0;
          task_lookup->minFree = 0;
          task_lookup->arbitraryMessageCount = 0;
        }

        PMPI_Gather (task_lookup, sizeof (callsite_stats_t),
                     MPI_CHAR, task_data,
                     sizeof (callsite_stats_t), MPI_CHAR,
                     memPi.collectorRank, memPi.comm);

        if (memPi.rank == memPi.collectorRank)
        {
          sCount = 0;
          sMax = 0;
          sMin = DBL_MAX;
          sCumulative = 0;

          for (j = 0; j < memPi.size; j++)
          {
            if (task_data[j].cumulativeFree > 0)
            {
              sCount += task_data[j].count;
              sCumulative += task_data[j].cumulativeFree;
              sMax = max (task_data[j].maxFree, sMax);
              sMin = min (task_data[j].minFree, sMin);

              fprintf (fp,
                       memP_Report_Formats
                       [MEMP_CALLSITE_IO_RANK_FMT][memPi.reportFormat],
                       memP_func_name (0, task_data[j].op),
                       av[i]->csid,
                       task_data[j].rank, task_data[j].count,
                       task_data[j].maxFree,
                       task_data[j].cumulativeFree /
                       task_data[j].count, task_data[j].minFree,
                       task_data[j].cumulativeFree);
            }
          }
          if (sCumulative > 0)
          {
            fprintf (fp,
                     memP_Report_Formats
                     [MEMP_CALLSITE_IO_SUMMARY_FMT][memPi.reportFormat],
                     memP_func_name (0, task_data[j - 1].op),
                     av[i]->csid,
                     "*", sCount, sMax, sCumulative / sCount, sMin,
                     sCumulative);
          }
          fprintf (fp, "\n");
        }
      }
    }

    if (memPi.rank == memPi.collectorRank)
    {
      memPi_free (av);
      memPi_free (task_data);
    }
  }
}
#endif

static int
task_info_sort_by_hwm (const void *p1, const void *p2)
{
  memPi_task_info_t *tip_1 = *(memPi_task_info_t **) p1;
  memPi_task_info_t *tip_2 = *(memPi_task_info_t **) p2;

  /*fprintf (stderr, "Comparing %f:%d and %f:%d\n", tip_1->taskHWM, tip_1->rank,
     tip_2->taskHWM, tip_2->rank); */

  return (int) (tip_2->taskHWM - tip_1->taskHWM);
}

void
memPi_profile_print_text_statistics (FILE * fp, memPi_task_info_t ** tips)
{
  int i, dlw = 26;
  double sum = 0, mean = 0, stddev = 0;

  print_section_heading (fp, "Heap HWM Statistics");

  /* Report Max HWM */
  fprintf (fp, "%-*s : %.0f\n", dlw, "Max", tips[0]->taskHWM);

  /*  Report Median  */
  if (memPi.size % 2 == 0)
  {
    double median_mean =
      (tips[memPi.size / 2]->taskHWM +
       tips[(memPi.size / 2) - 1]->taskHWM) / 2;
    fprintf (fp, "%-*s : %.0f\n", dlw, "Median", median_mean);
  }
  else
    fprintf (fp, "%-*s : %.0f\n", dlw, "Median",
             tips[(memPi.size - 1) / 2]->taskHWM);

  /*  Calculate sum for mean  */
  for (i = 0; i < memPi.size; i++)
    sum += tips[i]->taskHWM;

  //fprintf(stderr, "sum is %f, memPi.size is %d\n", sum, memPi.size);
  mean = sum / (double) memPi.size;
  fprintf (fp, "%-*s : %.0f\n", dlw, "Mean", mean);

  fprintf (fp, "%-*s : %.0f\n", dlw, "Min", tips[memPi.size - 1]->taskHWM);

  /*  Calculate stddev  */
  for (i = 0; i < memPi.size; i++)
    stddev += (tips[i]->taskHWM - mean) * (tips[i]->taskHWM - mean);

  if (stddev != 0)
    stddev = sqrt (stddev / (double) memPi.size);

  fprintf (fp, "%-*s : %.0f\n", dlw, "Stddev", stddev);
  fprintf (fp, "%-*s : %f\n", dlw, "Coefficient of variation", stddev / mean);
}


void
memPi_profile_print_text_minimal (FILE * fp)
{
  struct
  {
    double data;
    int rank;
  } sendbuf, maxHWM, minHWM;
  double totHWM;

  /* No information has been previously gathered.  We will obtain
     task data using collectives.
   */

  /*  TODO: Reduce or Allreduce?  To we need to coordinate info from max task?
   */
  sendbuf.data = memPi.taskHWM;
  sendbuf.rank = memPi.rank;

  PMPI_Reduce (&sendbuf, &maxHWM, 1, MPI_DOUBLE_INT, MPI_MAXLOC,
               memPi.collectorRank, MPI_COMM_WORLD);
  PMPI_Reduce (&sendbuf, &minHWM, 1, MPI_DOUBLE_INT, MPI_MINLOC,
               memPi.collectorRank, MPI_COMM_WORLD);
  PMPI_Reduce (&memPi.taskHWM, &totHWM, 1, MPI_DOUBLE, MPI_SUM,
               memPi.collectorRank, MPI_COMM_WORLD);

  if (memPi.rank == memPi.collectorRank)
  {
    /* Print report header */
    memPi_print_report_header (fp);

    fprintf (fp, "Max  Heap HWM : %.0f (Rank %d)\n", maxHWM.data,
             maxHWM.rank);
    fprintf (fp, "Mean Heap HWM : %.0f\n", totHWM / memPi.size);
    fprintf (fp, "Min  Heap HWM : %.0f (Rank %d)\n", minHWM.data,
             minHWM.rank);
  }
}


void
memPi_profile_print_text_concise (FILE * fp)
{
  int i;
  callsite_src_id_cache_entry_t **av;
  int ac;
  memPi_task_info_t **tips = NULL;
  char shbuf[128];
  int rank_width = 10,
    hwm_width = 18, stack_width = 10, sum_width = 18, avail_width = 18;

  if (memPi.rank == memPi.collectorRank)
  {
    /* Print report header */
    memPi_print_report_header (fp);

    snprintf (shbuf, 128,
              "Greatest Heap High Water Mark (top %d, descending, bytes)",
              memPi.size >
              memPi.reportHWMprintCount ? memPi.
              reportHWMprintCount : memPi.size);

    print_section_heading (fp, shbuf);

    /* Get callsite source information that can be accessed by call site id */
    h_gather_data (callsite_src_id_cache, &ac, (void ***) &av);
    qsort (av, ac, sizeof (void *), callsite_src_id_cache_sort_by_id);

    /* Create an array to access task info sorted by descending HWM */
    tips =
      (memPi_task_info_t **) memPi_malloc (sizeof (memPi_task_info_t *) *
                                           memPi.size);
    for (i = 0; i < memPi.size; i++)
    {
      tips[i] = &(memPi.global_task_info[i]);
    }
    qsort (tips, memPi.size, sizeof (memPi_task_info_t *),
           task_info_sort_by_hwm);

#if 0
    for (i = 0; i < memPi.size; i++)
    {
      fprintf (fp, "Sorted HWM is %12.0f for task %d\n", tips[i]->taskHWM,
               tips[i]->rank);
    }
#endif

    fprintf (fp, "%*s%*s%*s%*s",
             rank_width, "Rank",
             hwm_width, "Heap HWM", stack_width, "Stack", sum_width, "Sum");
#ifdef DO_GET_AVAIL
    fprintf (fp, "%*s\n", avail_width, "MinAvail@HWM");
#else
    fprintf (fp, "\n");
#endif

    //  Report on all tasks
    if (memPi.taskHWMrank == -1)
    {
      for (i = 0; i < memPi.size && i < memPi.reportHWMprintCount; i++)
      {
        fprintf (fp,
                 "%*d%*.0f%*.0f%*.0f",
                 rank_width, tips[i]->rank,
                 hwm_width, tips[i]->taskHWM,
                 stack_width, tips[i]->taskHWMstackUse,
                 sum_width, tips[i]->taskHWM + tips[i]->taskHWMstackUse);
#ifdef DO_GET_AVAIL
        fprintf (fp, "%*.0f\n", avail_width, tips[i]->taskHWMheapAvail);
#else
        fprintf (fp, "\n");
#endif
      }
    }
    else                        // Report on 1 task
    {
      for (i = 0; i < memPi.size; i++)
      {
        if (tips[i]->rank == memPi.taskHWMrank)
          fprintf (fp,
                   "Rank %6d HWM heap is %12.0f, stack at HWM is %12.0f\n",
                   tips[i]->rank, tips[i]->taskHWM, tips[i]->taskHWMstackUse);
      }
    }

    fprintf (fp, "\n");

    memPi_profile_print_text_statistics (fp, tips);

    memPi_free (tips);
  }
}


extern void memPi_profile_print_xml_verbose (FILE * fp);
void
memPi_profile_print (FILE * fp, int report_style)
{
  if (report_style == memPi_style_text_minimal)
    memPi_profile_print_text_minimal (fp);

  if (report_style == memPi_style_text_concise)
    memPi_profile_print_text_concise (fp);

  else if (report_style == memPi_style_xml_verbose)
    memPi_profile_print_xml_verbose (fp);
  else
  {
    if (memPi.collectorRank == memPi.rank)
    {
      assert (fp);

      memPi_print_report_header (fp);
    }

#if 0
    if (report_style == memPi_style_verbose)
      memPi_profile_print_verbose (fp);
    else if (report_style == memPi_style_concise)
      memPi_profile_print_concise (fp);
#endif

    if (memPi.collectorRank == memPi.rank)
      print_section_heading (fp, "End of Report");
  }
}

#if 0
void
memPi_profile_print_concise (FILE * fp)
{
  if (memPi.collectorRank == memPi.rank)
  {
    fprintf (fp, "\n");
    memPi_print_concise_task_info (fp);

    memPi_print_top_time_sites (fp);
    memPi_print_top_sent_sites (fp);
    memPi_print_top_io_sites (fp);

    memPi_print_callsites (fp);

    if (memPi.collective_report == 0)
    {
      if (memPi.print_callsite_detail)
      {
        /*memPi_print_concise_callsite_time_info (fp); */
        memPi_print_concise_callsite_sent_info (fp);
        memPi_print_concise_callsite_io_info (fp);
      }
    }
  }
  if (memPi.collective_report == 1)
  {
    /*memPi_coll_print_concise_callsite_time_info (fp); */
    memPi_coll_print_concise_callsite_sent_info (fp);
    memPi_coll_print_concise_callsite_io_info (fp);
  }
}



void
memPi_profile_print_verbose (FILE * fp)
{
  if (memPi.collectorRank == memPi.rank)
  {
    memPi_print_task_assignment (fp);
    fprintf (fp, "\n");

    memPi_print_verbose_task_info (fp);
  }
  print_alloc_stats (fp);
  if (memPi.collectorRank == memPi.rank)
  {

    memPi_print_callsites (fp);

    memPi_print_top_time_sites (fp);
    memPi_print_top_sent_sites (fp);
    memPi_print_top_io_sites (fp);
  }

  if (memPi.print_callsite_detail)
  {
    if (memPi.collective_report == 1)
    {
      memPi_msg_debug0 ("Using collective process reporting routines\n");
      memPi_msg_debug0
        ("MEMORY : collective reporting memory allocation :        %13ld\n",
         sizeof (callsite_stats_t) * memPi.size);

      /*memPi_coll_print_all_callsite_time_info (fp); */
      memPi_coll_print_all_callsite_sent_info (fp);
      memPi_coll_print_all_callsite_io_info (fp);
    }
    else
    {


      if (memPi.collectorRank == memPi.rank)
      {
        memPi_msg_debug
          ("Using standard process reporting routines aggregating data at process rank %d\n",
           memPi.collectorRank);

        /*memPi_print_all_callsite_time_info (fp); */
        memPi_print_all_callsite_sent_info (fp);
        memPi_print_all_callsite_io_info (fp);
      }
    }
  }

}
#endif


/* 

<license>

Copyright (c) 2010, Lawrence Livermore National Security, LLC.
Produced at the Lawrence Livermore National Laboratory
Written by Christopher Chambreau <chcham@llnl.gov>.
CODE-OCEC-10-010
All rights reserved.

This file is part of memP. For details, see http://memp.sourceforge.net .
Please also read this link – Additional BSD Notice.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  o  Redistributions of source code must retain the above copyright notice, 
     this list of conditions and the disclaimer below.

  o  Redistributions in binary form must reproduce the above copyright notice, 
     this list of conditions and the disclaimer (as noted below) in the 
     documentation and/or other materials provided with the distribution.

  o  Neither the name of the LLNS/LLNL nor the names of its contributors may be
     used to endorse or promote products derived from this software without 
     specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL LAWRENCE LIVERMORE NATIONAL
SECURITY, LLC, THE U.S. DEPARTMENT OF ENERGY OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.


Additional BSD Notice

1. This notice is required to be provided under our contract with the U.S.
Department of Energy (DOE). This work was produced at Lawrence Livermore
National Laboratory under Contract No. DE-AC52-07NA27344 with the DOE.

2. Neither the United States Government nor Lawrence Livermore National
Security, LLC nor any of their employees, makes any warranty, express or
implied, or assumes any liability or responsibility for the accuracy,
completeness, or usefulness of any information, apparatus, product, or process
disclosed, or represents that its use would not infringe privately-owned rights.

3. Also, reference herein to any specific commercial products, process, or
services by trade name, trademark, manufacturer or otherwise does not
necessarily constitute or imply its endorsement, recommendation, or favoring by
the United States Government or Lawrence Livermore National Security, LLC.
The views and opinions of authors expressed herein do not necessarily state or
reflect those of the United States Government or Lawrence Livermore National
Security, LLC, and shall not be used for advertising or product endorsement
purposes.

</license>

*/


/* eof */
