/* -*- C -*- 

   memP Heap Profiler ( http://memp.sourceforge.net/ )

   Please see COPYRIGHT AND LICENSE information at the end of this file.

   ----- 

   util.c -- misc util functions

 */

#ifndef lint
static char *svnid = "$Id: util.c 394 2007-04-16 18:17:26Z chcham $";
#endif

#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>

#include "memPi.h"
#ifdef HAVE_LIBDL
#define __USE_GNU
#include <dlfcn.h>
#endif

#include "symbols.h"

static int argc = 0;
static char **argv = NULL;


char *
GetBaseAppName (char *rawName)
{
  char *cp;

  if (rawName == NULL)
    return strdup ("Unknown");

  /* delete directorys and make a copy, if nec */
  if ((cp = rindex (rawName, '/')) == NULL)
  {
    return rawName;
  }
  else
  {
    return cp + 1;              /* skip over the '/' */
  }
}


void
memPi_getenv ()
{                               /* NEED TO ADD SANITY CHECKS HERE */
  char *cp = NULL;
  char *ep = NULL;
  extern int memPi_debug;

  memPi.outputDir = ".";

  ep = getenv ("MEMP");
  memPi.envStr = (ep ? strdup (ep) : 0);
  optind = 1;                   /* reset to avoid conflicts if getopt already called */
  if (ep != NULL)
  {
    int c;
    extern char *optarg;
    int ac = 0;
    char *av[64];
    char *sep = " \t,";

    av[0] = "GETOPT ERROR ";
    for (cp = strtok (ep, sep), ac = 1; (ac < 64) && (NULL != cp); ac++)
    {
      av[ac] = cp;
      /*memPi_msg("av[%d] = %s\n", ac, av[ac]); */
      cp = strtok (NULL, sep);
    }

    av[ac] = NULL;

    for (; ((c = getopt (ac, av, "ae:ng:f:b:s:k:top:xdh:i:j:u:")) != EOF);)
    {
      switch (c)
      {
      case 'f':
        memPi.outputDir = optarg;
        if (memPi.rank == 0)
          memPi_msg ("Set the output directory to [%s].\n", memPi.outputDir);
        break;

      case 'g':
        memPi_debug = atoi (optarg);
        if (memPi.rank == 0)
          memPi_msg ("Set debug mode level to [%d].\n", memPi_debug);
        break;

      case 's':
        {
          int defaultSize = memPi.tableSize;
          memPi.tableSize = atoi (optarg);
          if (memPi.tableSize < 2)
          {
            if (memPi.rank == 0)
              memPi_msg_warn
                ("-s tablesize invalid %d. Using default.\n",
                 memPi.tableSize);
            memPi.tableSize = defaultSize;
          }
          if (memPi.tableSize < 128)
          {
            if (memPi.rank == 0)
              memPi_msg_warn
                ("tablesize small %d. Consider making it larger w/ -s.\n",
                 memPi.tableSize);
          }
          if (memPi.rank == 0)
            memPi_msg
              ("Set the callsite table size to [%d].\n", memPi.tableSize);
        }
        break;

      case 'k':
        {
          int defaultStackDepth = memPi.stackDepth;
          memPi.stackDepth = atoi (optarg);
          if (memPi.stackDepth < 1)
          {
            if (memPi.rank == 0)
              memPi_msg_warn
                ("-k stack depth invalid %d. Using default.\n",
                 memPi.stackDepth);
            memPi.stackDepth = defaultStackDepth;
          }
/*		if (memPi.stackDepth > 2)
		  {
		    if (memPi.rank == 0)
		      memPi_msg_warn
			("stackdepth large %d. Consider making it smaller.\n",
			 memPi.stackDepth);
		  }*/
          if (memPi.stackDepth > MEMP_CALLSITE_STACK_DEPTH_MAX)
          {
            if (memPi.rank == 0)
              memPi_msg_warn
                ("stackdepth too large %d. Making it smaller %d.\n",
                 memPi.stackDepth);
            memPi.stackDepth = MEMP_CALLSITE_STACK_DEPTH_MAX;
          }
          if (memPi.rank == 0)
            memPi_msg
              ("Set the callsite stack traceback depth to [%d].\n",
               memPi.stackDepth);
        }
        break;

      case 'o':
        {
          if (memPi.rank == 0)
            memPi_msg_warn
              ("Disabling MEMP at Init. Code must use Pcontrol to enable.\n");
          memPi.enabled = 0;
          memPi.enabledCount = 0;
        }
        break;

      case 'n':
        memPi.baseNames = 1;
        break;

      case 't':
        memPi.traceAllStacks = 1;
        break;

      case 'e':
        if (optarg != NULL)
        {
          memPi.appFullName = strdup (optarg);
          memPi.av[0] = strdup (optarg);
          memPi.appName = strdup (GetBaseAppName (memPi.appFullName));
          memPi_msg_debug ("Set memPi.appFullName to %s\n",
                           memPi.appFullName);
        }
        break;
      case 'x':
        memPi.report_style = memPi_style_xml_verbose;
        break;
      case 'd':
        /*  Activate printing of call site detail   */
        memPi.print_callsite_detail = 1;
        break;

      case 'h':
        memPi.taskHWMreportThreshold = strtod (optarg, NULL);
        if (memPi.rank == 0)
          memPi_msg
            ("Set the task HWM report threshold to [%3.0lf].\n",
             memPi.taskHWMreportThreshold);
        break;
      case 'i':
        memPi.taskHWMreportMaxCount = atoi (optarg);
        break;
      case 'j':
        memPi.taskHWMrank = atoi (optarg);
        break;
      case 'a':
        memPi.abortAtThreshold = 1;
        break;
      case 'p':
        memPi.reportHWMprintCount = atoi (optarg);
        if (memPi.reportHWMprintCount == -1)
          memPi.reportHWMprintCount = memPi.size;
        if (memPi.rank == 0)
          memPi_msg ("Set report print count to [%d].\n",
                     memPi.reportHWMprintCount);

        break;

      case 'u':
        memPi.skipFrames = atoi (optarg);
        if (memPi.rank == 0)
          memPi_msg
            ("Set the number of stack frames to skip before recording to [%d].\n",
             memPi.skipFrames);
        break;

      case 'z':
      case 'b':
      case 'q':
      case 'w':
      case 'y':
      case '?':
        if (optopt == 'g')
        {
          memPi_debug = 1;
          if (memPi.rank == 0)
            memPi_msg ("Set debug mode level to [%d].\n", memPi_debug);
        }
        else if (memPi.rank == 0)
        {
          memPi_msg_warn
            ("Invalid option or missing argument encountered.\n", c);
        }
        break;
      default:
        if (memPi.rank == 0)
          memPi_msg_warn ("Option flag (-%c) not recognized. Ignored.\n", c);
        break;
      }
    }
  }

  optind = 1;                   /* reset to avoid conflicts if getopt called again */
}


#ifdef Linux
char *
getProcExeLink ()
{
  int pid, exelen, insize = 256;
  char *inbuf = NULL, file[256];

  pid = getpid ();
  snprintf (file, 256, "/proc/%d/exe", pid);
  inbuf = memPi_malloc (insize);
  if (inbuf == NULL)
  {
    memPi_abort ("unable to allocate space for full executable path.\n");
  }

  exelen = readlink (file, inbuf, 256);
  if (exelen == -1)
  {
    if (errno != ENOENT)
    {
      while (exelen == -1 && errno == ENAMETOOLONG)
      {
        insize += 256;
        inbuf = memPi_realloc (inbuf, insize);
        exelen = readlink (file, inbuf, insize);
      }
      inbuf[exelen] = '\0';
      return inbuf;
    }
    else
      memPi_free (inbuf);
  }
  else
  {
    inbuf[exelen] = '\0';
    return inbuf;
  }
  return NULL;
}

#define MEMP_MAX_ARG_STRING_SIZE 4096
void
getProcCmdLine (int *ac, char **av)
{
  int i = 0, pid;
  char *inbuf = NULL, file[256];
  FILE *infile;
  char *arg_ptr;

  *ac = 0;
  *av = NULL;

  pid = getpid ();
  snprintf (file, 256, "/proc/%d/cmdline", pid);
  infile = fopen (file, "r");

  if (infile != NULL)
  {
    while (!feof (infile))
    {
      inbuf = memPi_malloc (MEMP_MAX_ARG_STRING_SIZE);
      if (inbuf == NULL)
        memPi_abort ("Failed to allocate space for command-line agruments");

      if (fread (inbuf, 1, MEMP_MAX_ARG_STRING_SIZE, infile) > 0)
      {
        arg_ptr = inbuf;
        while (*arg_ptr != '\0')
        {
          av[i] = strdup (arg_ptr);
          arg_ptr += strlen (av[i]) + 1;
          i++;
        }
      }
    }
    *ac = i;

    memPi_free (inbuf);
    fclose (infile);
  }
}

#endif

void
memPi_copy_args (int *ac, char **av, int av_len)
{
  assert (ac != NULL);
  assert (av != NULL);

#if defined(AIX)
  {
    /*  Works for C/C++ and Fortran  */
    extern int p_xargc;
    extern char **p_xargv;
    argc = p_xargc;
    argv = p_xargv;
  }
#elif defined(GNU_Fortran)
  {
    extern int f__xargc;
    extern char **f__xargv;
    argc = f__xargc;
    argv = f__xargv;
  }
#elif defined(OSF1)
  {
    /*  Works for C/C++ and Fortran  */
    extern int __Argc;
    extern char **__Argv;
    argc = __Argc;
    argv = __Argv;
  }
#elif defined(PGI)
  {
    extern int __argc_save;
    extern char **__argv_save;
    argc = __argc_save;
    argv = __argv_save;
  }
#endif

  memPi_copy_given_args (ac, av, av_len, argc, argv);
}


void
memPi_copy_given_args (int *ac, char **av, int av_len, int argc, char **argv)
{
  int i;

  assert (ac != NULL);
  assert (av != NULL);

#if defined(USE_GETARG)

#define EXECUTABLE_LEN 2048
  {
    char buf[EXECUTABLE_LEN];
    int len, memPi_argc, oi;
    extern void F77_MEMPI_GET_FORTRAN_ARGC (int *);

    F77_MEMPI_GET_FORTRAN_ARGC (&memPi_argc);
    memPi_argc++;
    *ac = memPi_argc;

    for (i = 0, oi = 0; i < memPi_argc; i++)
    {
      int buf_len = EXECUTABLE_LEN;

      extern void F77_MEMPI_GET_FORTRAN_ARG (int *, int *, char *, int *,
                                             int);
      F77_MEMPI_GET_FORTRAN_ARG (&i, &buf_len, buf, &len, EXECUTABLE_LEN);

      /* With BG/P XL Fortran, argv[0] may be /lib/ld.so.1.
       * If so, exclude it.
       */

      buf[len < EXECUTABLE_LEN ? len : EXECUTABLE_LEN - 1] = 0;
      if ( NULL == strstr(buf, "/ld.so") )
        {
          av[oi] = strdup (buf); oi++;
        }
    }
  }

#else /* USE_GETARG */
  *ac = argc;

  for (i = 0; (i < *ac) && (i < av_len); i++)
    av[i] = strdup (argv[i]);
#endif
}


char *
memP_format_address (void *pval, char *addr_buf)
{
  static int get_sys_info = 0;
  static size_t ptr_hex_chars = 0;
  static char hex_prefix[3] = "";
  char test_buf[8] = "";

  if (get_sys_info == 0)
  {
    ptr_hex_chars = sizeof (char *) * 2;
    snprintf (test_buf, 8, "%p", (void *) 0x1);

    if (strcmp (test_buf, "0x1") != 0)
      strcpy (hex_prefix, "0x");

    get_sys_info = 1;
  }

  sprintf (addr_buf, "%s%p", hex_prefix, pval);

  return addr_buf;
}


int
memPi_verify_unanimous (int inval)
{
  int retval;
  PMPI_Allreduce (&inval, &retval, 1, MPI_INT, MPI_MIN, memPi.comm);
  return retval;
}


char *
memP_get_executable_name ()
{
#ifdef Linux
  return getProcExeLink ();
#else
  int ac;
  char *av[1];

  memPi_copy_args (&ac, av, 1);
  if (av[0] != NULL)
    return av[0];

  return NULL;
#endif
}


void
memPi_print_stack (void **pc_array)
{
  static int init = 0;
  int i;
  char *func = NULL, *infile = NULL;
  char addr_buf[24];

  memPi_msg_debug ("Rank %4d : memP Stacktrace\n", memPi.rank);

#define PRINT_ADDRS_ONLY 1
#if PRINT_ADDRS_ONLY == 1
  for (i = 0;
       i < MEMP_CALLSITE_STACK_DEPTH_MAX && pc_array[i] != NULL
       && (func == NULL || strcmp (func, "main")); i++)
  {
#if 0
    memPi_msg_debug ("      %3d : %s\n", i,
                     memP_format_address (pc_array[i], addr_buf));
#else
    fprintf (stderr, "      %3d : %s\n", i,
             memP_format_address (pc_array[i], addr_buf));
#endif
  }
#else
  if (init == 0)
  {
    /*  Get executable info and bfd handle   */
    infile = memP_get_executable_name ();

    if (infile == NULL || memPi_open_executable (infile) == 0)
    {
      memPi_abort ("print_stack failed to open %s\n", infile);
      exit (1);
    }
    else
      memPi_msg_debug ("opened executable file %s\n", infile);

    init = 1;
  }


  //for (i = 0; i < MAX_STACK && pc_array[i] != NULL && i < frame_count;
  for (i = 0;
       i < MEMP_CALLSITE_STACK_DEPTH_MAX && pc_array[i] != NULL
       && (func == NULL || strcmp (func, "main")); i++)
  {
#if ! defined(DISABLE_BFD) || defined(USE_LIBDWARF)
    int line_no;
    char *file = NULL;

    if (memP_find_src_loc (pc_array[i], &file, &line_no, &func) == 0)
    {
      memPi_msg ("for %s, file: %-10s line:%-2d func:%s\n",
                 memP_format_address (pc_array[i], addr_buf), file,
                 line_no, func);
      memPi_free (file);
      memPi_free (func);
    }
    else
      memPi_msg ("lookup failed for %s\n",
                 memP_format_address (pc_array[i], addr_buf));
#else
    memPi_msg ("  %s\n", memP_format_address (pc_array[i], addr_buf));

#endif
  }
#endif
}


/* 

<license>

Copyright (c) 2010, Lawrence Livermore National Security, LLC.
Produced at the Lawrence Livermore National Laboratory
Written by Christopher Chambreau <chcham@llnl.gov>.
CODE-OCEC-10-010
All rights reserved.

This file is part of memP. For details, see http://memp.sourceforge.net .
Please also read this link – Additional BSD Notice.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  o  Redistributions of source code must retain the above copyright notice, 
     this list of conditions and the disclaimer below.

  o  Redistributions in binary form must reproduce the above copyright notice, 
     this list of conditions and the disclaimer (as noted below) in the 
     documentation and/or other materials provided with the distribution.

  o  Neither the name of the LLNS/LLNL nor the names of its contributors may be
     used to endorse or promote products derived from this software without 
     specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL LAWRENCE LIVERMORE NATIONAL
SECURITY, LLC, THE U.S. DEPARTMENT OF ENERGY OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.


Additional BSD Notice

1. This notice is required to be provided under our contract with the U.S.
Department of Energy (DOE). This work was produced at Lawrence Livermore
National Laboratory under Contract No. DE-AC52-07NA27344 with the DOE.

2. Neither the United States Government nor Lawrence Livermore National
Security, LLC nor any of their employees, makes any warranty, express or
implied, or assumes any liability or responsibility for the accuracy,
completeness, or usefulness of any information, apparatus, product, or process
disclosed, or represents that its use would not infringe privately-owned rights.

3. Also, reference herein to any specific commercial products, process, or
services by trade name, trademark, manufacturer or otherwise does not
necessarily constitute or imply its endorsement, recommendation, or favoring by
the United States Government or Lawrence Livermore National Security, LLC.
The views and opinions of authors expressed herein do not necessarily state or
reflect those of the United States Government or Lawrence Livermore National
Security, LLC, and shall not be used for advertising or product endorsement
purposes.

</license>

*/


/* eof */
